(*
 * Copyright © 1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

(* $Id: abstraction.ml.m4,v 1.5 2005/05/18 21:14:03 rupak Exp $

   *
   * This file is part of the BLAST Project.
*)


(**
  * This module implements the abstraction interface connecting the model
  * checker to the concrete transition system.  It defines a Region module
  * based on Predicate Abstraction and BDDs, the symbolic pre and post,
  * and routines for refinment used by the model checker.
*) 


open BlastArch

module Symbol = Ast.Symbol
module Constant = Ast.Constant
module Expression = Ast.Expression
module Predicate = Ast.Predicate
module Counter = Ast.Counter
module Stats = Bstats


		 
module Make_C_Abstraction =
  functor(C_SD : SYSTEM_DESCRIPTION with module Command = BlastCSystemDescr.C_Command) ->
struct

  module PredBdd =  PredBdd.Make_PredBdd(C_SD) 
  module PredTable = PredTable.Make_PredTable(PredBdd)
  module Absutil = Absutil.Make_Absutil(PredTable)

  open PredTable
  open Absutil

  let arg_subst _ = failwith "Not implememted"

  let check_races = ref None (* this will now be an lval *)

		      
  let assumed_invariants = ref []


  let maxStackLen = ref 0 
  let getCurrentPredIndex () = PredTable.getCurrentIndex ()


				 

  let dump_abs () = PredTable.dump_abs ()
  let is_fun_refined fn = PredTable.is_fun_refined fn 
		      


  (*********************************************************************************************)
  (*********************************************************************************************)
  (* The operation, DataLattice and Region modules  *)
				 include(abs-region.ml)
  (*********************************************************************************************)
  (*********************************************************************************************)
				 
 				 

  (**********************************************************************************************)
  (**********************************************************************************************)
  (* The environment automaton model in the TAR algorithm *)
  (**********************************************************************************************)
  (**********************************************************************************************)
  type opKind = ProgramOperation of Operation.t | AutomatonOperation of (int * int) 

  module EnvAutomaton =
  struct
    
    type onode_label =
	{
	  oid : int;
	  oadr : Region.abstract_data_region;
	  oregion : Region.t
	}
	  
    (* Invariants: Node 0 is always the stuck node *)      
	  
    type anode_label =
	{ id   : int ; (* the id of the abstract automaton node *)
	  root : int ; (* the id of the root node in the cluster *)
	  mutable adr  : Region.abstract_data_region ; (* the abstract data region (is this on globals?) *)
          mutable atomic : bool ;
	  mutable task : bool;
	  mutable event : bool;
	  mutable modifies : Expression.lval list ; (* the list of globals that are modified by this transition *)

	}	
	  
    type aedge_label = 
	Thread of Expression.lval list              (* Solid edge *) 
      | Env                 (* Dotted edge *)
	  
    type oautomaton_node = (onode_label, opKind) BiDirectionalLabeledGraph.node
	
    type automaton_node = ((* node label *) anode_label, (* edge label *) aedge_label) BiDirectionalLabeledGraph.node 
	
    type  ('a, 'b) automaton = {
      
      autom_root : automaton_node ;
      
      
      number_of_states : int ;
      
      id_to_node_ht : (int, automaton_node) Hashtbl.t ;
      
      o_auto_root : oautomaton_node;
      
      cover_set : int Region.coverers;
      (* other maps *)
      o_map : (int, oautomaton_node) Hashtbl.t;  (* auto_map *)
      o_to_a_map : (int, int) Hashtbl.t;         (* auto_shrink_map *)
      a_to_o_map : (int, int list) Hashtbl.t;    (* shrink_elts_map *)
      
     } 

    let getNodeFromId autom i = (* raises Not_found *)
      Hashtbl.find autom.id_to_node_ht i

    let getAdrFromId autom i = (* raises Not_found *)
      let nd = getNodeFromId autom i in
	(BiDirectionalLabeledGraph.get_node_label nd).adr

    let getIdFromNode autom nd = 
      (BiDirectionalLabeledGraph.get_node_label nd).id

    let getModifiesFromNode autom nd = 
      (BiDirectionalLabeledGraph.get_node_label nd).modifies

    let getModifiesFromId autom id = 
      let nd = getNodeFromId autom id in
	(BiDirectionalLabeledGraph.get_node_label nd).modifies

    let getModifiesOnEdge autom id = 
      let nd = getNodeFromId autom id in
      List.fold_left (fun sofar e -> match (BiDirectionalLabeledGraph.get_edge_label e) with 
	Thread lv -> Misc.union lv sofar | Env -> sofar)
	[] (BiDirectionalLabeledGraph.get_out_edges nd)
	
    let getAdrFromNode autom nd = 
      (BiDirectionalLabeledGraph.get_node_label nd).adr

    let getEnvEdges autom i =
      let thisnode = getNodeFromId autom i in
      let outgoing = BiDirectionalLabeledGraph.get_out_edges thisnode in
      List.filter (fun e -> match (BiDirectionalLabeledGraph.get_edge_label e) with 
	Env -> true | Thread _ -> false) outgoing

    let getThreadEdges autom i =
      let thisnode = getNodeFromId autom i in
      let outgoing = BiDirectionalLabeledGraph.get_out_edges thisnode in
      List.filter 
	(fun e -> match (BiDirectionalLabeledGraph.get_edge_label e) with 
	  Env -> false | Thread _ -> true) outgoing
	  



  end


    (**********************************************************************************************)
    (**********************************************************************************************)
    (* End of The environment automaton model in the TAR algorithm *)
    (**********************************************************************************************)
    (**********************************************************************************************)




    (*******************************************************************************************)
    (* for m4 *)
    (* GUI functions *)
    include(abs-gui.ml)
    (* PRE/POST CONDITIONS + lvalue_maps *)
    include(abs-wpsp.ml)
    (*******************************************************************************************)			  

  (* Abstract pre operator. Keep track of predicates mentioned 
     in reg.preds
  *)
  (* Type: Region.abstract_data_region -> C_SD.Command.t -> Region.abstract_data_region *)
  let abstract_data_pre abst_data_reg cmd edge post_location phi_table =
    Message.msg_string Message.Debug "In abstract_data_pre -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg ;
    Message.msg_printer Message.Debug C_SD.Command.print cmd ;

    (* Implementation of the H function from DasDillPark-CAV99 *)
    let rec _H (bdd, p) predlist =
      (* optimization: check if (abst_data_reg, p) is satisfiable? *)
      if (PredTable.isSatisfiable (bdd, p)) then
	begin
	  match predlist with
	      [] -> if (PredTable.isSatisfiable (bdd, p)) then PredTable.bddOne else PredTable.bddZero
	    | a:: rest ->
		let pr = PredTable.getPred a in
		let cr = concrete_data_post { Region.pred = pr ; } cmd edge phi_table in
		let wppos = cr.Region.pred in
		let wpneg = Predicate.negate wppos in
		  CaddieBdd.bddOr 
		    (CaddieBdd.bddAnd (PredTable.getPredIndexBdd a) (_H (bdd, wppos :: p) rest))
		    (CaddieBdd.bddAnd (CaddieBdd.bddNot (PredTable.getPredIndexBdd a)) (_H (bdd, wpneg :: p) rest))
	end
      else
	PredTable.bddZero
    in
    let pre_bdd = 
      if Options.getValueOfBool "bddpost" then
        let src_loc = C_SD.location_coords (C_SD.get_source edge) in
        let dst_loc = C_SD.location_coords post_location in
        let trel_bdd = PredTable.get_transrel_bdd (src_loc,dst_loc) in
          PredTable.bdd_preimage abst_data_reg.Region.bdd trel_bdd
      else
        _H (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds
    in
      { Region.lelement = DataLattice.pre abst_data_reg.Region.lelement cmd ;
        Region.bdd = pre_bdd ;
	Region.preds = abst_data_reg.Region.preds ; }

  let query_fn bdd (qpred:Predicate.predicate)  =
    let trueval = not (PredTable.isSatisfiable (bdd, [ Predicate.Not qpred ]))
    in let falseval = not (PredTable.isSatisfiable (bdd, [ qpred]))
    in
      if (trueval = falseval) then Ast.Dontknow
      else if trueval then  Ast.True
      else Ast.False

  let rec sane_pred_list pred_list =
    match pred_list with
      | [] -> []
      | head::tail ->
	  if (head = Ast.Predicate.True) then sane_pred_list tail
	  else [head] @ (sane_pred_list tail)

	      
  let pred_to_assume_cmd _pred =
    {C_SD.Command.code = C_SD.Command.Pred _pred;
     C_SD.Command.read = [];
     C_SD.Command.written = []}

      
  let rec get_bdd_for_pred_list lpredlist __H_cmd tmpbdd regpreds =
    match lpredlist with
      | [] -> tmpbdd
      | hd::tail ->
	  let newtmpbdd = __H_cmd (pred_to_assume_cmd hd) (tmpbdd, []) regpreds
	  in
	  get_bdd_for_pred_list tail __H_cmd newtmpbdd regpreds
	    
  let abstract_data_post_das_dill abst_data_reg cmd edge post_location phi_table =
    Message.msg_string Message.Debug "In abstract_data_post -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg ;
    Message.msg_printer Message.Debug C_SD.Command.print cmd ;
    
    (* Implementation of the H function from DasDillPark-CAV99 *)
    let rec _H_cmd _cmd (bdd,p) predlist =
      Message.msg_string Message.Debug ("pred = "^(match p with [] -> "" | _ -> Predicate.toString (List.hd p)));
      (* optimization: check if (abst_data_reg, p) is satisfiable? *)
      if (PredTable.isSatisfiable (bdd,p)) then
	begin
	  match predlist with
	      [] -> if (PredTable.isSatisfiable (bdd,p)) then PredTable.bddOne else PredTable.bddZero
	    | a:: rest ->
		let pr = PredTable.getPred a in
		let cr = concrete_data_pre false false { Region.pred = pr ; } _cmd edge phi_table in
		let wppos = cr.Region.pred in
		let ncr = concrete_data_pre false false { Region.pred = Predicate.negate pr ; } _cmd edge phi_table in
		let wpneg = ncr.Region.pred
		in
		  CaddieBdd.bddOr 
		    (CaddieBdd.bddAnd (PredTable.getPredIndexBdd a) (_H_cmd _cmd (bdd, wppos :: p) rest))
		    (CaddieBdd.bddAnd (CaddieBdd.bddNot (PredTable.getPredIndexBdd a)) (_H_cmd _cmd (bdd, wpneg :: p) rest))
	end
      else
	PredTable.bddZero
    in
    let _H = _H_cmd cmd in
      
      match cmd.C_SD.Command.code with 
	  C_SD.Command.FunctionCall fcallExp ->
	    let (fname,_subs,_) = (C_SD.deconstructFunCall fcallExp) in 
	    let asgn_block =
	      { cmd with C_SD.Command.code = C_SD.Command.Block (List.map (fun (x,e) -> C_SD.Command.Expr( Expression.Assignment (Expression.Assign,Expression.Symbol(x), e))) _subs) }
	    in let (_lpredlist, rlelement) = DataLattice.post abst_data_reg.Region.lelement edge (query_fn abst_data_reg.Region.bdd)
	    in let lpredlist = sane_pred_list _lpredlist
	    in 
	      if (List.length lpredlist) = 0 then
		{ 
		  Region.lelement = rlelement;
		  Region.preds = abst_data_reg.Region.preds;
		  Region.bdd =   _H_cmd asgn_block (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds;
		}
	      else
		let tmppreds = abst_data_reg.Region.preds
		in let tmpbdd = _H_cmd asgn_block (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds
		in let tmpbdd2 = get_bdd_for_pred_list lpredlist _H_cmd tmpbdd abst_data_reg.Region.preds
		in 
		{ 
		  Region.lelement = rlelement;
		  Region.preds = tmppreds;
		  Region.bdd = tmpbdd2 ;
		}
		  
	(*let subs = List.map Misc.swap _subs in
	  let rec _make_copy_list lst c_lst pBdd = 
	  match lst with
	  [] -> (c_lst,pBdd)
	  | h::t -> 
	  let pred_h = PredTable.getPred h in
	  let cpred_h = funCallSubstitute subs pred_h in
	  if (cpred_h = pred_h) 
	  then (_make_copy_list t c_lst pBdd)  (* no copy made *)
	  else (
	  let cpred_i = addPred cpred_h in
	  let nextPBdd = CaddieBdd.bddAnd (CaddieBdd.bddBiimp
	  (PredTable.getPredIndexBdd h)
	  (PredTable.getPredIndexBdd
	  cpred_i)) pBdd in
	  _make_copy_list t (cpred_i::c_lst) nextPBdd) (* copy made *)
	  in 
	  let (subsPreds, newBdd) = _make_copy_list abst_data_reg.Region.preds [] abst_data_reg.Region.bdd
	  in
	  {
	  Region.lelement = DataLattice.... ;
	  Region.bdd = newBdd ;
	  Region.preds = abst_data_reg.Region.preds} *)
	(* What was just done for the function call ... *) 
	(* Push stuff onto the stack: ... in the rest of post 
	   1. edge, on return, the location will be
	   the other end of the edge ...
	   2. if edge was a y = f(x) then post on return reflects this *)
	(* Go through list of predicates and compute "copies" *)
	(* Compute BDD of reg ^_i (pred_i = copy_pred_i) *)
	(* Start location = start location of function *)
	|	C_SD.Command.Block l  -> 
		  if ( not (List.exists (function x -> match x with C_SD.Command.Return e -> true | _ -> false) l))
		  then 
		    let _ = Absutil.reset_ret_pointers () in
		    let (_lpredlist, rlelement) = DataLattice.post abst_data_reg.Region.lelement edge (query_fn abst_data_reg.Region.bdd)
		    in let lpredlist = sane_pred_list _lpredlist
		    in 
		      if (List.length lpredlist) = 0 then
			{ 
			  Region.lelement = rlelement;
			  Region.preds = abst_data_reg.Region.preds;
			  Region.bdd =   _H (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds;
			}
		      else
			let tmppreds = abst_data_reg.Region.preds
			in let tmpbdd = _H (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds
			in let tmpbdd2 = get_bdd_for_pred_list lpredlist _H_cmd tmpbdd abst_data_reg.Region.preds
			in 
			  { 
			    Region.lelement = rlelement;
			    Region.preds = tmppreds;
			    Region.bdd = tmpbdd2
			  }
		  else 
		    let (targ, fname, retloc ,subs) = lastTarget () in
		      (* Jhala: the t.o.s. is passed using last_loc_pointer  reg.Region.stack *) 
		    let _ = Absutil.set_ret_pointers targ subs (C_SD.get_location_fname retloc) in
		    let (newPreds,delPreds) = Absutil.removeFunPreds fname abst_data_reg.Region.preds in
		    let _ = Message.msg_string Message.Debug "Return: Calling _H" in
		    let newBdd = (_H  (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds) in
		    let _ = Message.msg_string Message.Debug "Return: Back from _H" in
		    let _predToVar i =  -1 in (* JHALA: To be done ... ??? *)
		    let ex_newBdd = newBdd in
		    let _ = Absutil.reset_ret_pointers () in
		    let	(_lpredlist, rlelement) = DataLattice.post abst_data_reg.Region.lelement edge (query_fn abst_data_reg.Region.bdd)
		    in let lpredlist = sane_pred_list _lpredlist
		    in 
		      if (List.length lpredlist) = 0 then
			{ 
			  Region.preds = newPreds@delPreds;
			  Region.lelement = rlelement;
			  Region.bdd =  ex_newBdd;
			}
		      else
			let tmppreds = newPreds@delPreds
			in let tmpbdd = ex_newBdd (*_H (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds*)
			in let tmpbdd2 = get_bdd_for_pred_list lpredlist _H_cmd tmpbdd (newPreds@delPreds)
			in 
			  { 
			    Region.lelement = rlelement;
			    Region.preds = tmppreds;
			    Region.bdd = tmpbdd2
			  }
	| _ ->
	    let (_lpredlist, rlelement) = DataLattice.post abst_data_reg.Region.lelement edge (query_fn abst_data_reg.Region.bdd)
	    in let lpredlist = sane_pred_list _lpredlist
	    in 
	      if (List.length lpredlist) = 0 then
		{ 
		  Region.bdd = _H (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds ;
		  Region.lelement = rlelement;
		  Region.preds = abst_data_reg.Region.preds ;
		}
	      else
		let tmppreds = abst_data_reg.Region.preds
		in let tmpbdd = _H (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds
		in let tmpbdd2 = get_bdd_for_pred_list lpredlist _H_cmd tmpbdd abst_data_reg.Region.preds
		in 
		  { 
		    Region.lelement = rlelement;
		    Region.preds = tmppreds;
		    Region.bdd = tmpbdd2
		  }
	(*	    
  in
	      { 
		Region.bdd = _H (abst_data_reg.Region.bdd, []) abst_data_reg.Region.preds ;
		Region.lelement = rlelement;
		Region.preds = abst_data_reg.Region.preds ;
	      }
	*)


  (* -----  JHALA: The new abstract_data_post : SLAM_POST *)
  let assume_pred_ref = ref None 
			  (* We use this to check if a given predicate is unchanged in this post ie if its wp = p 
			     ... which also means <?> wpnp = np this is cleared every time postBdd is called *)

    let unchangedPredTable = Hashtbl.create 37 
    let dontcareArray = Array.create PredTable.maxPossiblePreds false 
    let outOfScopeArray = Array.create PredTable.maxPossiblePreds false 


  let extract_scope var = 
    let names = Misc.chop var "@" in
      try List.nth names 1 with _ -> "__BLAST_GLOBAL" 


  let postBdd bdd _pred_list _post_split_preds =
    Message.msg_string Message.Debug "In postBdd..\n";
    Message.msg_string Message.Debug ("post of assume"^(match !assume_pred_ref with 
      Some(e) -> ("YES:"^(Predicate.toString e)) | _ -> "NO")^"\n");
    (* Message.msg_string Message.Debug ("Bdd = "^(Predicate.toString (PredTable.convertBddToPred bdd)));*)
    Message.msg_string Message.Debug "\nList of predicates = ";

    let pred_list = List.map (fun (a,b,c,d) -> (a,b,c)) _pred_list in
    let post_split_preds = List.map (fun (a,b,c,d) -> (a,b,c)) _post_split_preds in
    let craig_flag = (Options.getValueOfInt "craig" >= 2 ) in
    let craig_flag1 = (Options.getValueOfInt "craig" = 1) in
    let dc_flag = (Options.getValueOfBool "dc") in
    if (Options.getValueOfBool "scope") then 
      begin
	Misc.array_reset outOfScopeArray false;
	List.iter (fun (a,_,_,d) -> Array.set outOfScopeArray a d) _pred_list 
      end;
    let rec loop lst = 
      match lst with 
	[] ->  () 
      | (a,b,c,d)::rest -> 
	  begin
	    ignore (Message.msg_string Message.Debug (Predicate.toString (PredTable.getPred a)));
	    ignore (Message.msg_string Message.Debug (Predicate.toString b));
	    ignore (Message.msg_string Message.Debug (Predicate.toString c)); 
	    ignore (Message.msg_string Message.Debug (string_of_bool d));
	    loop rest
	  end
    in
    Message.msg_apply Message.Debug loop _pred_list;
    Message.msg_string Message.Debug ("Will split on: \n");
    Message.msg_apply Message.Debug loop  _post_split_preds ;

    (* assuming that if p = wp, then np = wpnp ...*)
    let wp_different_filter (p,wp,_) = 
      if craig_flag then true 
      else
	match Options.getValueOfString "post" with
	    "slam" -> true
	  | "slam-dumb" -> not (PredTable.getPred p = wp)
	  | "slam1" -> not ((PredTable.getPred p = wp) && (!assume_pred_ref = None))
	  | "slam2" -> let p_pred = PredTable.getPred p in 
	      (not (p_pred = wp) || 
	       (match !assume_pred_ref with 
		    None -> false
		  |	Some(asm) -> not (Misc.intersection (Absutil.getVarExps_deep p_pred) (Absutil.getVarExps_deep asm) = [])))  
		(* JHALA: Instead of this intersection check I would like to say 
		   if the two preds are textually the same or imply 
		   each other or some such ...*)
	  |  a -> failwith ("Unknown post algorithm "^a)
    in
      (* I dont have a good idea of when an "if" affects/doesnt affect a predicate, 
	 so for now, leave them all in ... there should be some cone-check here ...*)
    let (_changed_pred_list, unchanged_pred_list) = Stats.time "wp_diff_filter" (Misc.filter_cut wp_different_filter) pred_list in
    let cpl_addon = if craig_flag1 then post_split_preds else [] in
    let changed_pred_list = _changed_pred_list @ cpl_addon in

    let _ = Hashtbl.clear unchangedPredTable in
    let _ = List.iter (function x -> Hashtbl.add unchangedPredTable (Misc.fst3 x) true) unchanged_pred_list in
      (* all the unchanged preds are dumped into this table *)
    let sumofcubesBdd = ref PredTable.bddZero in
    let num_cubes = ref 0 in

    let conjunctL = ref [] in
      (* conjunct that the present cube produces post-command c *)
    let _ = Misc.array_reset dontcareArray false in

    (* START TIME HOLE *)
    let _check_Pred_list_simplify assumptions pred_triple_list conjunct_pointer =
      let asm_conj = Predicate.conjoinL assumptions in
      let _ = Message.msg_string Message.Debug ("Assuming pre-state: "^(Predicate.toString asm_conj)) in
      let _ = Stats.time "_assume" (TheoremProver._assume ) asm_conj in 
      let _check_pred (*_asm_conj*) (p,wpp,wpnp) = 
	if ( dontcareArray.(p) && dc_flag) then false 
	  (* this tuple will never be called if I dont care about it if craig is true *)
	else 
	  let eToCheck1 = Predicate.implies asm_conj wpp in
	  let eToCheck2 = Predicate.implies asm_conj wpnp in
	  let _ = Message.msg_string Message.Debug ("Post pred: "^(Predicate.toString (PredTable.getPred p))) in
	  let _ = Message.msg_string Message.Debug ("Post query+: "^(Predicate.toString wpp)) in
	  let _ = Message.msg_string Message.Debug ("Post query-: "^(Predicate.toString wpnp)) in
	  let ans1 = Stats.time "simp-a1:" (PredTable.askTheoremProverContext eToCheck1) wpp in
	  let ans2 = 
            if (Options.getValueOfBool "posp") then false
            else
              Stats.time "simp-a2:" (PredTable.askTheoremProverContext eToCheck2) wpnp 
          in
	  if ans1 then statNonDCQueriesForPost := !statNonDCQueriesForPost + 1;
          if ans2 then statNonDCQueriesForPost := !statNonDCQueriesForPost + 1;
          match ans1 with
	    true ->  
	      if  (ans2 = true) then 
		begin
		  Message.msg_string Message.Debug ("Cube implies both wp(p) and wp(neg p): "^(Predicate.toString (PredTable.getPred p))); 
		  conjunct_pointer := (p,true)::(p,false)::(!conjunct_pointer);
		  true
		end
	      else
		begin 
		  conjunct_pointer := (p,true)::(!conjunct_pointer);
		  true               
		end
	  | false ->  
	      if (ans2 = true) then  
		begin
		  conjunct_pointer := (p,false)::(! conjunct_pointer);
		  true
		end
	      else 
		false
      in
      let _ = Stats.time "_check_pred" (List.map (_check_pred (* asm_conj *))) pred_triple_list in
	statWorstCaseQueriesForPost :=  2*(List.length pred_list) + !statWorstCaseQueriesForPost;
	statActualQueriesForPost := 2*(List.length pred_triple_list) + (!statActualQueriesForPost);
	statNumPost := !statNumPost + 1;
	(match (!assume_pred_ref) with
	     Some _ -> 
	       (statAssumeQueriesForPost := 2*(List.length pred_triple_list) + (!statAssumeQueriesForPost); statNumAssumePost := !statNumAssumePost + 1)
	   | None -> ())
	;
	TheoremProver._pop ();
	true
    in
      (* END TIME HOLE *)
(*RETIRING VAMPYRE CODE
    let _check_Pred_list_vampyre assumptions pred_triple_list conjunct_pointer = 
      let is_cube_consistent  = TheoremProver.assume assumptions in
      let asm_conj = Predicate.conjoinL assumptions in
      let _check_pred assumptions (p,wpp,wpnp) = 
	if ((dontcareArray.(p)) && dc_flag) then false
	else 
	  let eToCheck1 = Predicate.implies asm_conj wpp in
	  let eToCheck2 = Predicate.implies asm_conj wpnp in
	  let ans1 = (PredTable.askTheoremProverContext eToCheck1) wpp in
	  let ans2 = (PredTable.askTheoremProverContext eToCheck2) wpnp in
	    match ans1 with
		true ->  
		  if  (ans2 = true) then 
		    begin
		      Message.msg_string Message.Debug ("Cube implies both wp(p) and wp(neg p): "^(Predicate.toString (PredTable.getPred p))); 
		      conjunct_pointer := (p,true)::(p,false)::(!conjunct_pointer);
		      true
		    end
		  else
                    begin 
		      conjunct_pointer := (p,true)::(!conjunct_pointer);
		      true               
		    end
	      | false -> 
		  if (ans2 = true) then  
		    begin
		      conjunct_pointer := (p,false)::(!conjunct_pointer);
		      true
		    end
		  else 
		    false
      in

	if (is_cube_consistent) then 
	  begin 
	    let _ = List.map (_check_pred asm_conj) pred_triple_list in
	      true
	  end 
	else false
    in
*)
    let _check_Pred_list = 
(*RETIRING VAMPYRE CODE 
     match (Options.getValueOfString "s" ) with
	  "vampyre" -> _check_Pred_list_vampyre
	| _ -> *) _check_Pred_list_simplify
    in
      (* the preds below are the integer identifiers *)
    let rec make_conjunct_bdd conjl = 
      match conjl with
	  [] -> PredTable.bddOne 
	| (p,true)::tl ->  CaddieBdd.bddAnd (PredTable.getPredIndexBdd p) (make_conjunct_bdd tl) 
	| (p,false)::tl -> CaddieBdd.bddAnd (CaddieBdd.bddNot (PredTable.getPredIndexBdd p))  (make_conjunct_bdd tl) 
    in

    let rec refine_cube pre_reg cube_list split_preds = 
      (* Jhala: this next line is very slow and should be optimized *)
      let has_no_value (x,wp,wpn) = List.for_all (fun y -> x <> fst y) cube_list in
      let preds_to_split_on = List.filter (has_no_value) split_preds in 
	match preds_to_split_on with
	    [] -> make_conjunct_bdd cube_list
	  | (h,wp,wpneg)::t -> 
	      begin
		let cube_plus = (h,true)::cube_list in
		let pre_reg_plus = wp::pre_reg in
		let conj_plus_ptr = ref cube_plus in
		let _ = _check_Pred_list pre_reg_plus preds_to_split_on conj_plus_ptr in
		let plus_bdd = refine_cube pre_reg_plus !conj_plus_ptr preds_to_split_on in

		let cube_minus = (h,false)::cube_list in
		let pre_reg_minus = wpneg::pre_reg in
		let conj_minus_ptr = ref cube_minus in
		let _ = _check_Pred_list pre_reg_minus preds_to_split_on conj_minus_ptr in
		let minus_bdd = refine_cube pre_reg_minus !conj_minus_ptr preds_to_split_on in

		  CaddieBdd.bddOr plus_bdd minus_bdd
	      end
    in
    let _check_Cube _pred_list (cube : int array)  =
      (Message.log_string_dbg Message.BddPost " In _check_Cube");
      statReachedCubes := !statReachedCubes + 1;
      conjunctL := [];   
      let _ = 
	if (dc_flag) then Misc.array_reset dontcareArray false;
	if (craig_flag) then 
	  begin
	    Misc.array_reset dontcareArray true; 
	    List.iter (fun (i,_,_) -> Array.set dontcareArray i false) pred_list
	      (* make sure we care about the predicates that we care about *)
	  end
      in
      let cubePostIsEmpty = 
	match (!assume_pred_ref) with 
	  Some(e) -> (PredTable.askTheoremProver 
			(Predicate.implies (Predicate.conjoinL [(convertCubeToPred cube);e]) Predicate.False))
	| None -> false
      in 
	if (cubePostIsEmpty) then ()
	else 
	  begin   
	    let pref = match (!assume_pred_ref) with None -> ref [] | Some(e) -> ref [e] in
	    let _plist = ref _pred_list in
	      begin
		for i = 0 to (Array.length cube - 1) do
		  if (cube.(i)==2) 
		  then 
		    begin
		      try
			if ((dc_flag) && (outOfScopeArray.(i)))  then
			  begin
			    Message.log_string_dbg Message.BddPost
			      ("DC about: "^(Predicate.toString (PredTable.getPred i))); 
			    Array.set dontcareArray i true
			  end
		      with _ -> ()
		    end
		  else 
		    begin
		      let i_pred = PredTable.getPred i in
			(* Assume the invariant that i_pred : Predicate.Atom *)
                      let i_unchanged = (Hashtbl.mem unchangedPredTable i)  
		      in
			if (cube.(i)==1) then
			  begin
			    pref := i_pred::(!pref);
			    if (i_unchanged && not(dontcareArray.(i))) 
			    then conjunctL := (i,true)::(!conjunctL);
			  end
			else
			  begin
			    pref := (Predicate.negate i_pred)::(!pref);
			    if (i_unchanged && not(dontcareArray.(i))) 
			    then conjunctL := (i,false)::(!conjunctL);
			  end
		    end
		done;

		(* Now ask the thmprover if pref -> each fellow in predlist *)
		if (Stats.time "_check_Pred_list" (_check_Pred_list !pref)  changed_pred_list conjunctL) 
		then 
		  begin
		    (* Jhala: I'm being paranoid about the soundness and aliasing. something to assuage me *)
		    if (Options.getValueOfBool "scheck") then
		      begin
			let cube_l = List.map 
			    (fun (i,b) -> if b then (PredTable.getPred i) else Predicate.Not(PredTable.getPred i))
			    !conjunctL 
			in
			if (TheoremProver.queryExp (Predicate.Implies ((Predicate.And cube_l),(Predicate.False))) && (!assume_pred_ref = None)) 
			then 
			  begin
			    Message.log_string_always Message.BddPost 
			      "SCHECK: Post of block is FALSE !!"
			  end
			else 
			  ()
		      end ;
		      
		    (*  this is where we plug in the post_split_cube *)
		    let _ = 
		      if (Options.getValueOfInt "craig" >= 3) 
		      then conjunctL := List.filter (PredTable.keep_pred_parity) !conjunctL 
		    in
		    let splitup_bdd = 
		      if (false (*Options.getValueOfBool "cref"*)) then (refine_cube !pref !conjunctL post_split_preds)
		      else (make_conjunct_bdd !conjunctL) 
		    in
		    sumofcubesBdd := CaddieBdd.bddOr (splitup_bdd) (!sumofcubesBdd);
                    num_cubes := 1 + !num_cubes;
                    
		  end; () 
	      end;
	  end
    in
    begin
      if (CaddieBdd.bddEqual bdd PredTable.bddZero) then
	sumofcubesBdd := PredTable.bddZero
      else 
	begin
	  if (CaddieBdd.bddEqual bdd PredTable.bddOne) 
	  then
	    let _tmpcube = (* build a temporary cube with nothing in it *)
	      Array.make (*size *)(getCurrentIndex ()+2) (* content *)2 
		
	      (* Note: The size is certainly a vast	overestimate, but this
		 is the easiest way. The "+2" removes additional complications relating
		 to index being -1 initially. Note that this is a fake cube -- we should
		 never try to print it or convert it to a predicate!Ugly hack, but 
		 works. So far *)
	    in
	    _check_Cube changed_pred_list _tmpcube
	  else
	    ignore (Stats.time "bddForEachCube-post" (CaddieBdd.bddForeachCube bdd) (_check_Cube changed_pred_list));
	end;
      Message.log_string_norm Message.BddPost
	("Returning from post-bdd: number of cubes = "^(string_of_int !num_cubes));
      !sumofcubesBdd 
    end
      
  (** BDD-based image computation takes input bdd, *)   
  let postBdd_bdd bdd (src_id,dst_id) preserved_preds = 
    (* for now, ignore the preserved_preds *)
    let trel_bdd = PredTable.get_transrel_bdd (src_id,dst_id) in
      PredTable.bdd_postimage bdd trel_bdd
	

  (** Implementation of the H function from DasDillPark-CAV99 *)
  let postBdd_dasdillpark bdd assume_pred_opt pred_triple_list =
    let pre_pred = 
      let bdd_pred = PredTable.convertBddToPred bdd in (* use bdd_to_predicate, after it works! *)
	match assume_pred_opt with
	    Some p' -> Predicate.And [p';bdd_pred]
	  | None -> bdd_pred
    in
    let sat_check p = 
      if (true) then (TheoremProver.interpolant_sat (pre_pred, Predicate.And p) = TheoremProver.Sat)
      else PredTable.isSatisfiable (bdd,p)
    in
    let rec _H p _pred_triple_list =
      Message.log_string_norm Message.BddPost ("_H depth : "^(string_of_int (List.length p)));
      Message.log_string_dbg Message.BddPost ("pred = "^(match p with [] -> "" | _ -> Predicate.toString (List.hd p)));
      if (sat_check p) then
	begin
	  match _pred_triple_list with
	      [] -> PredTable.bddOne 
	    | (a,wppos,wpneg,_):: rest ->
                let bdd_a = (PredTable.getPredIndexBdd a) in
		  CaddieBdd.bddOr 
		    (CaddieBdd.bddAnd bdd_a (_H (wppos :: p) rest))
		    (CaddieBdd.bddAnd (CaddieBdd.bddNot bdd_a) (_H (wpneg :: p) rest))
	end
      else
	PredTable.bddZero
    in
      Message.log_string_norm Message.BddPost "begin ddp _H";
      let rv = _H [] pred_triple_list in
	Message.log_string_norm Message.BddPost "end ddp _H";
	rv


	  
  let postBdd_dasdillpark_inc (assert_fn,pop_fn,issat_fn,reset_fn)
    bdd assume_pred_opt pred_triple_list = 
    let pre_pred = 
      let bdd_pred = PredTable.convertBddToPred bdd in (* use bdd_to_predicate, after it works! *)
	match assume_pred_opt with
	    Some p' -> Predicate.And [p';bdd_pred]
	  | None -> bdd_pred
    in
    let rec _H depth _pred_triple_list =
      Message.log_string_norm Message.BddPost ("_H depth : "^(string_of_int depth));
      if not (issat_fn ()) then
	(Message.log_string_norm Message.BddPost "ddpinc:unsat";PredTable.bddZero)
      else
	begin
          match _pred_triple_list with
              [] -> PredTable.bddOne 
            | (a,wppos,wpneg,_)::rest ->
		let recur p = 
		  assert_fn p;
		  let rv = _H (depth + 1) rest in pop_fn (); rv
		in
		let bdd_pos = recur wppos in
		let bdd_neg = recur wpneg in
		let bdd_a = (PredTable.getPredIndexBdd a) in
		  CaddieBdd.bddOr 
		    (CaddieBdd.bddAnd bdd_a bdd_pos) 
		    (CaddieBdd.bddAnd (CaddieBdd.bddNot bdd_a) bdd_neg)
	end
    in
      Message.log_string_norm Message.BddPost "begin ddp-inc _H";
      assert_fn pre_pred;
      let rv = _H 0 pred_triple_list in
	reset_fn ();
	Message.log_string_norm Message.BddPost "end ddp-inc _H";
	rv


  (** uses incremental foci interface *)
  let postBdd_ddp_inc_foci = 
    postBdd_dasdillpark_inc
      (TheoremProver.foci_inc_push,
       TheoremProver.foci_inc_pop,
       TheoremProver.foci_inc_check_sat, 
       TheoremProver.foci_inc_reset)

      
  (** uses incremental simplify interface *)
  let postBdd_ddp_inc_simplify = 
    let issat_fn () = not (TheoremProver._is_contra ()) in
      postBdd_dasdillpark_inc
	(TheoremProver.block_assert,
	 TheoremProver.block_pop,
	 issat_fn, 
	 TheoremProver.block_reset)

	


	
	
  let wp_table = Hashtbl.create 1009 
  let stackInfo = ref None


  let rec s_get_bdd_for_pred_list lpredlist __H_cmd tmpbdd regpreds =
    match lpredlist with
      | [] -> tmpbdd
      | hd::tail ->
	  let assume_cmd = pred_to_assume_cmd hd in
	  let str = C_SD.Command.to_string assume_cmd(*.C_SD.Command.code*)
	  in
	  assume_pred_ref := Some (hd) ;
	  let newtmpbdd = __H_cmd (pred_to_assume_cmd hd) tmpbdd regpreds
	  in
	  assume_pred_ref := None ;
	  s_get_bdd_for_pred_list tail __H_cmd newtmpbdd regpreds

		    
  let rec abstract_data_post_graf_saidi abst_data_reg cmd edge post_location phi_table =
    Message.log_string_norm Message.BddPost "In abstract_data_post GRAF_SAIDI_POST -- arguments are:" ;
    Message.log_print_norm Message.BddPost Region.print_abstract_data_region abst_data_reg ;
    Message.log_print_norm Message.BddPost C_SD.Command.print cmd ;
    let src_loc = C_SD.location_coords (C_SD.get_source edge) in
    let target_loc = C_SD.location_coords post_location in
    let e_id = (src_loc,target_loc) in
    let _ = 
      Message.log_string_norm Message.BddPost (Printf.sprintf "absdatapost preloc: (%d,%d)" 
					  (fst src_loc)  (snd src_loc));
      Message.log_string_norm Message.BddPost (Printf.sprintf "absdatapost postloc: (%d,%d)" 
						 (fst target_loc)  (snd target_loc)) 
    in
    let predlist =  PredTable.lookup_location_predicates target_loc in   
    let _H_cmd _cmd bdd actual_predlist =
      (* optimization: check if (abst_data_reg, p) is satisfiable? *)
      (* so here is where we shall use stackInfo -- the format is (top of stack,2nd top of stack, call op) *)
      let (tos,tos2,call) = match (!stackInfo) with
	  Some (t,t2,c) -> (t,t2,c) 
	| None -> failwith "scope info NOT supplied !"
      in
      let _ = Message.log_string_dbg Message.BddPost
		(Printf.sprintf "tos=%s , tos2=%s , call=%s" tos tos2 (string_of_bool call)) in
      let out_of_scope pred = 
	if (Options.getValueOfBool "scope") then
	  let _stay_ff x = 
	    let scope_x = (extract_scope x) in
	      ((tos = scope_x) || (call && tos2 = scope_x) || (scope_x = "__BLAST_GLOBAL"))
	  in
	  let predVars = Absutil.getVarExps_deep pred in
	    List.exists (fun x -> not (_stay_ff x)) predVars
	else 
	  true
      in
      let prop_eq_bdd a = 
	if (Options.getValueOfBool "bddpost") then
          (let cb = PredTable.make_eq_bdd [a] in
             PredTable.update_transrel_bdd e_id cb) 
      in
      let make_pred_triple_cached () a =
	try Hashtbl.find wp_table (a,e_id) with
	  Not_found ->
            let rv = 
	      begin
		let pr = PredTable.getPred a in
                match (!assume_pred_ref) with 
		  None ->
		    let cr = concrete_data_pre false false { Region.pred = pr ; } _cmd edge phi_table in
		    let wppos = cr.Region.pred in
		    let _ = 
		      if (wppos <> pr) then PredTable.update_pred_affected_table e_id a 
		      else (prop_eq_bdd a) 
		    in
		    let ncr = concrete_data_pre false false { Region.pred = Predicate.negate pr ; } _cmd edge phi_table in
		    let wpneg = ncr.Region.pred in
		    (a,wppos,wpneg, (out_of_scope pr))
		| Some(e) -> 
		    begin 
		      prop_eq_bdd a;
		      (a,pr,Predicate.negate pr,(out_of_scope pr))
		    end
	      end
            in (Hashtbl.replace wp_table (a,e_id) rv; rv)
      in
      let post_split_preds = if (Options.getValueOfInt "craig" = 1) then predlist else [] in
      let pred_triple_ac = Stats.time "mk_pred_triple ac" (List.map (make_pred_triple_cached ())) actual_predlist in
      let pred_triple_sp = Stats.time "mk_pred_triple sp" (List.map (make_pred_triple_cached ())) post_split_preds in
	(* options begin -- bddpost,hybrid,dasdillpark *)
	(** this is code for exact predicate abstraction using various dec. procs  
	  if (Options.getValueOfString "post" = "Hsimplify") then
	  Stats.time "postBdd_ddp" (postBdd_ddp_inc_simplify bdd !assume_pred_ref) pred_triple_ac
	  else if (Options.getValueOfString "post" = "Hfoci") then
	  Stats.time "postBdd_ddp" (postBdd_ddp_inc_foci bdd !assume_pred_ref) pred_triple_ac
	  else if (Options.getValueOfString "post" = "H") then
	  let b1 = Stats.time "postBdd_ddp_foci" (postBdd_ddp_inc_foci bdd !assume_pred_ref) pred_triple_ac in
	  let b2 = Stats.time "postBdd_ddp_simp" (postBdd_ddp_inc_simplify bdd !assume_pred_ref) pred_triple_ac
	  in
	  if not (CaddieBdd.bddEqual b1 b2) then
          begin
          Message.msg_string Message.Error "foci,simplify disagree!";
          Message.msg_string Message.Normal "foci says:";
          Message.msg_string Message.Normal (Predicate.toString(PredTable.convertBddToPred b1));
          Message.msg_string Message.Normal "simplify says:";
          Message.msg_string Message.Normal (Predicate.toString(PredTable.convertBddToPred b2));
          exit 0
          end
	  else b1
	  else *) 
	if (Options.getValueOfBool "bddpost") then  
	  let _ = Stats.time "mk_pred_triple cp" (List.map (make_pred_triple_cached ())) (PredTable.bdd_support bdd) in
	  let b2 = Stats.time "postBdd-bdd" (postBdd_bdd bdd (src_loc,target_loc)) [] in     
	  let b1 = 
            if (Options.getValueOfBool "hybrid") then
              let bdd_cov = Stats.time "bdd_cube_cov" PredTable.bdd_cube_cover bdd in
		Stats.time "postBdd" (postBdd bdd_cov pred_triple_ac) pred_triple_sp
            else PredTable.bddOne   
	  in
          (CaddieBdd.bddAnd b1 b2) 
	else 
	  let b1 = Stats.time "postBdd" (postBdd bdd pred_triple_ac) pred_triple_sp in
	  b1
    in
    let _H = _H_cmd cmd in
      match cmd.C_SD.Command.code with 
          C_SD.Command.Skip 
	| C_SD.Command.SymbolicHook _
	| C_SD.Command.Block [] 
	| C_SD.Command.Pred (Predicate.True) -> 
            begin
              Message.log_string_dbg Message.BddPost "Treats as skip";
              if (Options.getValueOfBool "bddpost") then
		{abst_data_reg with Region.bdd = _H (abst_data_reg.Region.bdd) predlist}
              else abst_data_reg
            end
	| C_SD.Command.FunctionCall fcallExp ->
	    let (fname,_subs,target) = (C_SD.deconstructFunCall fcallExp) in
	    let _add_sub = 
	      if (C_SD.is_defined fname) then [] 
	      else 	[C_SD.Command.Expr( Expression.Assignment (Expression.Assign, target, Absutil.get_new_unknown fname ))] 
		(* SECURITY: havoc all the globals in predicates for really undef functions "pessim" *)
		(* SECURITY: add an option here -- if its undefined then havoc all imp globals,
		   if its just a skip function then leave it as above "optimistic" *)
	    in
	    let asgn_block =
	      { cmd with C_SD.Command.code = C_SD.Command.Block ((List.map (fun (x,e) -> C_SD.Command.Expr( Expression.Assignment (Expression.Assign,Expression.Symbol(x), e ))) _subs)@_add_sub) }
	    in
	    let (_lpredlist, rlelement) = (let l = DataLattice.post abst_data_reg.Region.lelement edge (query_fn abst_data_reg.Region.bdd) in  l (* DataLattice.post l asgn_block *) )
	    in let lpredlist = sane_pred_list _lpredlist
	    in 
	      if (List.length lpredlist) = 0 then
		{ 
		  Region.preds = abst_data_reg.Region.preds;
		  Region.lelement = rlelement;
		  Region.bdd =  _H_cmd asgn_block (abst_data_reg.Region.bdd) predlist (*abst_data_reg.Region.preds*);
		}
	      else
		let tmppreds = abst_data_reg.Region.preds
		in let tmpbdd = _H_cmd asgn_block (abst_data_reg.Region.bdd) predlist
		in let tmpbdd2 = s_get_bdd_for_pred_list lpredlist _H_cmd  tmpbdd  predlist
		in 
		  { 
		    Region.lelement = rlelement;
		    Region.preds = tmppreds;
		    Region.bdd = tmpbdd2
		  }
	(* What was just done for the function call ... *) 
	(* Push stuff onto the stack: ... in the rest of post 
	   1. edge, on return, the location will be
	   the other end of the edge ...
	   2. if edge was a y = f(x) then post on return reflects this *)
	(* Go through list of predicates and compute "copies" *)
	(* Compute BDD of reg ^_i (pred_i = copy_pred_i) *)
	(* Start location = start location of function *)
        | C_SD.Command.Block l  -> 
	    if ( not (List.exists (function x -> match x with C_SD.Command.Return e -> true | _ -> false) l))
	    then 
	      let _ = Absutil.reset_ret_pointers () in
	      let (_lpredlist, rlelement) = DataLattice.post abst_data_reg.Region.lelement edge (query_fn abst_data_reg.Region.bdd)
	      in let lpredlist = sane_pred_list _lpredlist
	      in 
		if (List.length lpredlist) = 0 then
		  { 
		    Region.preds = abst_data_reg.Region.preds;
	            Region.lelement = rlelement;
		    Region.bdd =   _H  (abst_data_reg.Region.bdd) predlist (*abst_data_reg.Region.preds*);
		  }
		else
		  let tmppreds = abst_data_reg.Region.preds
		  in let tmpbdd = _H  (abst_data_reg.Region.bdd) predlist
		  in let tmpbdd2 = s_get_bdd_for_pred_list lpredlist _H_cmd tmpbdd predlist
		  in 
		    { 
		      Region.lelement = rlelement;
		      Region.preds = tmppreds;
		      Region.bdd = tmpbdd2
		    }
	    else 
	      let (targ, fname, retloc , subs) = lastTarget () in
		(* Jhala: the t.o.s. is passed using last_loc_pointer  reg.Region.stack *) 
	      let _ = Absutil.set_ret_pointers targ subs (C_SD.get_location_fname retloc) in 
	      let (newPreds,delPreds) = Absutil.removeFunPreds fname abst_data_reg.Region.preds in
	      let newBdd = (_H  (abst_data_reg.Region.bdd) predlist (*abst_data_reg.Region.preds*)) in
              let ex_newBdd = newBdd in
		
	      let _ = Absutil.reset_ret_pointers () in
	      let (_lpredlist, rlelement) = DataLattice.post abst_data_reg.Region.lelement edge (query_fn abst_data_reg.Region.bdd)
	      in let lpredlist = sane_pred_list _lpredlist
	      in 
		if (List.length lpredlist) = 0 then
		  { 
		    Region.preds = newPreds@delPreds;
	            Region.lelement = rlelement;
		    Region.bdd =  ex_newBdd;
		  }
		else
		  let tmppreds = newPreds@delPreds
		  in let tmpbdd = ex_newBdd
		  in let tmpbdd2 = s_get_bdd_for_pred_list lpredlist _H_cmd tmpbdd (newPreds@delPreds)
		  in 
		    { 
		      Region.lelement = rlelement;
		      Region.preds = tmppreds;
		      Region.bdd = tmpbdd2
		    }
	| C_SD.Command.Pred e -> 
	    begin
	      assume_pred_ref := Some(e);
	      let (_lpredlist, rlelement) =
		DataLattice.post abst_data_reg.Region.lelement edge (query_fn abst_data_reg.Region.bdd) 
	      in let lpredlist = sane_pred_list _lpredlist
	      in 
	      let retReg = 
		if (List.length lpredlist) = 0 then
		  { 
		    Region.bdd = _H (abst_data_reg.Region.bdd) predlist ;
		    Region.lelement = rlelement;
		    Region.preds = abst_data_reg.Region.preds ;
		  }
		else
		  let tmppreds = abst_data_reg.Region.preds
		  in let tmpbdd = _H  (abst_data_reg.Region.bdd) predlist
		  in let tmpbdd2 = s_get_bdd_for_pred_list lpredlist _H_cmd tmpbdd predlist
		  in 
		  { 
		  Region.lelement = rlelement;
		  Region.preds = tmppreds;
		  Region.bdd = tmpbdd2
		}
	      in  
		assume_pred_ref := None;          
		retReg
	    end
	| C_SD.Command.Phi (sym, flag) -> 
	    let e = phi_command_to_pred sym flag phi_table in
              (* Now treat just like C_SD.Command.Pred e *)
	      if (e = Predicate.True) then abst_data_reg
	      else 
		begin
		  assume_pred_ref := Some(e);
		  let retReg = 
		    { 
		      Region.bdd = _H (abst_data_reg.Region.bdd) predlist (* abst_data_reg.Region.preds*) ;
		      Region.preds = abst_data_reg.Region.preds ;
		      Region.lelement = abst_data_reg.Region.lelement;
		    }
		  in  
		    assume_pred_ref := None;          
		    retReg
		end
	| C_SD.Command.Havoc sym -> 
	    let new_var = 
	      match Absutil.getNewTmpVar (Expression.Symbol "__foo") with 
		  Expression.Chlval (Expression.Symbol x, s) -> Expression.Lval (Expression.Symbol (s^x))
		| _ -> failwith "Abstraction.concrete_data_pre: Havoc: TBD"
	    in
	    let assign = Expression.Assignment(Expression.Assign, sym, new_var) in
	    let stmt = C_SD.Command.Expr assign in
	    let cmd' = {cmd with C_SD.Command.code = C_SD.Command.Block [stmt]} in
	      abstract_data_post_graf_saidi abst_data_reg cmd' edge post_location phi_table
	| C_SD.Command.HavocAll -> failwith "Abstraction.abstract_data_post_graf_Saidi: HavocAll: TBD"       


  (* Starting the definition of PRE *)


  let setWPCapRegion reg =
    let conc_reg = Region.concretize reg in
      wpCapRegion :=
      match conc_reg with 
	  Region.Empty -> Predicate.False
	| Region.Atomic atomic_reg -> 
	    begin
	      match atomic_reg.Region.data with
		  Region.Concrete foo -> foo.Region.pred
		| _ -> failwith ("bad call to setWPCapRegion")
	    end
	| Region.ParallelAtomic p_atomic_reg ->
	    begin
	      match p_atomic_reg.Region.p_data with
		  Region.Concrete foo -> foo.Region.pred
		| _ -> failwith ("bad call to setWPCapRegion")
	    end
	| _ -> failwith ("bad call to setWPCapRegion: Strange region type")

  let resetWPCapRegion () = 
    wpCapRegion := Predicate.True

  (* the symbolic pre function: (approximate) weakest pre-condition *)
  let pre_template subs_flag sat_flag reg op phi_table =
    Message.msg_string Message.Debug "In pre -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print reg ;
    Message.msg_printer Message.Debug Operation.print op ;
    Message.msg_string Message.Debug (if subs_flag then "Substitute \n" else "Explicit (no substitute) \n");
    Message.msg_string Message.Debug (if sat_flag then "sat-filter-on \n" else "sat-filter-off \n");
    match reg with
	Region.Empty ->
	  Region.bot
      | Region.Atomic atomic_reg ->
	  if (((C_SD.location_coords atomic_reg.Region.location) <> C_SD.location_coords (C_SD.get_target op)) 
		&& (not (C_SD.isReturn op)) && (not (C_SD.isFunCall op))) then
	    (* Jhala: This is only if this is not a return op *)
	    begin
	      Message.msg_string Message.Error "WARNING: pre: blast internal error suspected" ;
	      Message.msg_string Message.Error "              operation do not match this region's location" ;
	      failwith "location mismatch";
	      Region.bot
	    end
	  else
	    begin
	      let new_stack =
		if (C_SD.isFunCall op) then
		  begin 
		    match (C_SD.get_command op).C_SD.Command.code with

			C_SD.Command.FunctionCall fcexp -> 
			  let function_is_defined fcexp = 
			    match fcexp with
				Expression.FunctionCall(Expression.Lval (Expression.Symbol(fname)), _) -> 
				  C_SD.is_defined fname
			      | Expression.Assignment (_, _, Expression.FunctionCall (Expression.Lval (Expression.Symbol fname), _)) -> 
				  C_SD.is_defined fname
			      | _ -> 
				  (* Rupak: this case happens if there is a function pointer call.
				     Previously we failed here with:
				     failwith ("Bad function call :"^(Expression.toString fcexp))
				     However we can punt this by assuming that the function pointer call
				     does not call a known function, and continue.

				     Do this only if nofp is set.
				  *)
				  if (Options.getValueOfBool "nofp") then
				    false
				  else
				    failwith ("pre_template: Bad function call (Possible function pointer?):"^(Expression.toString fcexp))
			  in
			    if (not (function_is_defined fcexp)) then
			      atomic_reg.Region.stack
			    else 
			      begin
				let (_fname,_,target) = C_SD.deconstructFunCall fcexp in
				if (_fname = PredTable.__NotImplementedFunctionName) 
				then 
				  atomic_reg.Region.stack
				else   
				  if (C_SD.is_defined _fname ) then 
				    try (Region.CallStack (List.tl (Region.list_of_stack atomic_reg.Region.stack))) 
				    with e -> (print_string (Printexc.to_string e);failwith "Incorrect callstack type !!")
				  else 
				    atomic_reg.Region.stack (* was a skip *)
			      end
		    |    _ -> failwith "This can never happen: isFuncall checked that this was a function call!"
		  end
		else
		  if (C_SD.isReturn op)
		  then
		    begin
		      Message.msg_string Message.Debug "Pre: Hit isReturn\n";
		      (* recall that we put the call points in the stack *)
		      (* we assume that a function call edge is the only edge
			 going to its target location *)
		      Message.msg_string Message.Debug "Ingoing edges : ";
		      List.iter (fun a -> Message.msg_string Message.Debug (Operation.to_string a)) (C_SD.get_ingoing_edges atomic_reg.Region.location);
		      Message.msg_string Message.Debug "\nNow matching...\n";
                      let callLoc = 
			(C_SD.get_source (Misc.list_filter_unique C_SD.isFunCall (C_SD.get_ingoing_edges atomic_reg.Region.location))) in
			last_loc_pointer := Some(callLoc);
		      let (rt,_,retloc ,subs) = lastTarget () in
		      Absutil.set_ret_pointers rt subs (C_SD.get_location_fname retloc);
		      last_loc_pointer := None;
		      ( Region.CallStack (callLoc::(Region.list_of_stack atomic_reg.Region.stack)))
		    end
		  else
		    atomic_reg.Region.stack
	      in
	      let new_location = (C_SD.get_source op)
	      and new_data =
		match atomic_reg.Region.data with
		  Region.Abstract abst_data_reg ->
		    Region.Abstract (abstract_data_pre abst_data_reg (C_SD.get_command op) op atomic_reg.Region.location phi_table)
		| Region.Concrete conc_data_reg ->
		    Region.Concrete (concrete_data_pre subs_flag sat_flag conc_data_reg (C_SD.get_command op) (op) phi_table)
	      in
	      Absutil.reset_ret_pointers (); 
	      Region.Atomic { Region.location = new_location ;
			      Region.data = new_data ;
			      Region.stack = new_stack ; }
		
	    end
      |	Region.ParallelAtomic p_atomic_reg ->
	  Region.bot
	    (* i am here *)
      | Region.Union _ ->
	  failwith "pre: argument is a union: unimplemented"
      | Region.ParallelUnion _ ->
	  failwith "pre: argument is a parallel union: unimplemented"


  (* Jhala: am setting sat flag to false for now ... *)
  let pre (* reg op phi_table *) = pre_template true false (* reg op phi_table *)
  let spre (* reg op phi_table *) = pre_template false false (* reg op phi_table *)


  let abstract_data_post abst_data_reg  cmd edge post_location phi_table =
    (abstract_data_post_graf_saidi abst_data_reg  cmd) edge post_location phi_table 
      
      
  let abstract_predicate pred pred_list = 
    Message.msg_string Message.Debug "abstract_predicate called" ;
    assume_pred_ref := Some(pred);
    let abstract_bdd = postBdd bddOne (List.map (function i -> (i,getPred i,Predicate.negate (getPred i),true)) pred_list) [] 
    in
      assume_pred_ref := None;
      Region.Abstract { Region.lelement = DataLattice.top (* TBD: call DataLattice.abstract ... *) ; 
			Region.bdd = abstract_bdd; Region.preds = pred_list}

  let scope_kill data_region fname = 
    Message.msg_string Message.Debug "Inside scope_kill:\n";
    match data_region with
	Region.Abstract(abst_data_reg) ->
	  begin
	    Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg;
	    let _kill_ff x =  
	      let predVars = Absutil.getVarExps_deep (PredTable.getPred x) in
		List.exists (fun x -> Misc.is_substring x ("@"^fname)) predVars
	    in
	    let new_bdd = Absutil.hg_exists _kill_ff abst_data_reg.Region.bdd in
	    let ret_abst_data_reg  = { Region.bdd = new_bdd; 
	                               Region.lelement = abst_data_reg.Region.lelement;
                                       Region.preds = abst_data_reg.Region.preds} in
	      Message.msg_string Message.Debug "Returning from kill_data_region with :\n";
	      Message.msg_printer Message.Debug Region.print_abstract_data_region ret_abst_data_reg;
	      Region.Abstract(ret_abst_data_reg)
	  end
      | _ -> 
	  begin
	    Message.msg_string Message.Debug "bad call to kill"; 
	    data_region
	  end


  (*************************************************************************************)
  (* Automaton post : 
     1. dotted post along Env edges
     2. solid post along Thread edges
  *)
  (*************************************************************************************)

  (* check if the region contains some stuck state *)
  let exists_stuck reg =
    match reg with
      | Region.Empty -> false
      | Region.ParallelAtomic p_at ->
	  let ctrmap = p_at.Region.e_control in
	    not (Counter.eq ctrmap.( (* STUCK STATE *) 0) (Counter.make_counter 0)) && not (Region.is_empty reg)
      | Region.ParallelUnion punion ->
	  Region.AtomicRegionTable.fold
	    (fun (l,s) datalist flag -> flag || (List.exists (fun (ec,_) -> not (Counter.eq ec.((* stuck state *)0) (Counter.make_counter 0))) datalist)) 
	    punion false
      | _ -> failwith "exists_stuck is only used with (unions of) ParallelAtomic regions"

  let cluster_ctrmaps lmaps = (* cluster ctrmaps clusters the maps according to nonzero entries in the map,
                                 and gives back one map for each cluster which is the maximum pointwise map of all maps in the cluster *)
    let _ = Message.msg_string Message.Normal ("In cluster with: "^(string_of_int (List.length lmaps))^" : maps" ) in
      
    let get_nonzero_indices ctrmap =
      let indices = ref [] in for i = 0 to Array.length ctrmap -1 do if not (Counter.is_zero ctrmap.(i)) then indices := i :: !indices done ; !indices in
    let cluster_ht = Hashtbl.create 31 in (* make a hash table for the clusters *)
      List.iter (fun (thismap, _) -> 
		   let poss = get_nonzero_indices thismap in
		     if Hashtbl.mem cluster_ht poss then Hashtbl.replace cluster_ht poss 
		       (Counter.array_max (Hashtbl.find cluster_ht poss) thismap) 
		     else Hashtbl.add cluster_ht poss thismap) lmaps ;
      let rv = Hashtbl.fold (fun a b thislist -> (b,[])::thislist) cluster_ht [] in
      let _ = Message.msg_string Message.Normal ("Leaving cluster with: "^(string_of_int (List.length rv))^" : maps" ) in
	rv
	  


  (*  This part of the post deals with the updates to the counters as the automaton
      moves along dotted edges.
      
  *)

  let envPost automaton ctrmap ( adr : Region.abstract_data_region ) =
    (* for each location, get the dotted successors to locations 
       that contain adr 
    *)
    let allzeros ctrmap =
      let len = Array.length ctrmap in let i = ref 0 in let flag = ref false in
	while (not !flag) && !i < len do (if ctrmap.(!i) = Counter.make_counter 0 then () else flag := true); incr i done ; not !flag 
    in
    let numnodes = automaton.EnvAutomaton.number_of_states in
    let update_array = Array.make numnodes [] in
      if (allzeros ctrmap) then [ (ctrmap, []) ] else
	begin
	  Message.msg_string Message.Debug "In envPost: generating possible updates" ;
	  (* first generate an array of possible updates *)
	  for i = 0 to ((Array.length ctrmap) - 1) do 
	    if (ctrmap.(i) = (Counter.make_counter 0)) then 
	      ()
		
	    else (* three cases: ctrmap.(i) is 1 or 2 or omega *)
	      begin
		Message.msg_string Message.Debug ("Location "^(string_of_int i)^" is not empty");
		let ith_node = EnvAutomaton.getNodeFromId automaton i in
		let dotted_succs = List.filter (fun e -> match (BiDirectionalLabeledGraph.get_edge_label e) with
						    EnvAutomaton.Env -> true | _ -> false) (BiDirectionalLabeledGraph.get_out_edges ith_node)
		in
		let dotted_succ_nodes_containing_adr = (* myself, if I intersect adr *)
		  (if not (Region.adr_is_empty (Region.adr_cap adr (EnvAutomaton.getAdrFromId automaton i))) then [ ith_node ] else [])
		  @
 		  (* together with all my successors (except myself) that intersect adr *)
		  (List.map (BiDirectionalLabeledGraph.get_target) 
		     (List.filter 
			(fun e -> let target = BiDirectionalLabeledGraph.get_target e in
			   (EnvAutomaton.getIdFromNode automaton target <> i)&& not (Region.adr_is_empty (Region.adr_cap adr (EnvAutomaton.getAdrFromNode automaton target))) 
			) dotted_succs))
		in
		  List.iter (fun nd -> Message.msg_string Message.Debug ("Successor node : "^(string_of_int (EnvAutomaton.getIdFromNode automaton nd))) ) dotted_succ_nodes_containing_adr ; 
		  if dotted_succ_nodes_containing_adr = [] then
		    begin
                      (* stuck! *)
                      Message.msg_string Message.Debug "The automaton gets stuck. Do something!" ;
	              (* go to stuck, which is node 0 *)
		      update_array.(i) <- [(0 (* 0 is the stuck state *), ctrmap.(i), (i,0,ctrmap.(i)) )] :: update_array.(i)
		    end
		  else	  (* update maps *)
		    begin
	              Message.msg_string Message.Debug "In envPost: dotted succs not empty" ;	
                      if List.length dotted_succ_nodes_containing_adr = 1 then  (* this is the common case in the -dsq option *)
			update_array.(i) <- List.map (fun s -> let sid = EnvAutomaton.getIdFromNode automaton s in 
							[(sid, ctrmap.(i), (i,sid,(ctrmap.(i))) )]) dotted_succ_nodes_containing_adr
		      else
			if (Counter.eq ctrmap.(i) (Counter.make_counter 1)) then
			  (* the next states are where this counter goes to one of the next states *)
			  update_array.(i) <- List.map (fun s -> let sid = EnvAutomaton.getIdFromNode automaton s in 
							  [(sid, Counter.make_counter 1, (i,sid,(Counter.make_counter 1)) )]) dotted_succ_nodes_containing_adr
			else if (Counter.eq ctrmap.(i) (Counter.make_counter 2)) then
			  (* the next states are (1) both counters go to one state, or (2) one counter each goes to two states *)
			  begin
			    let pairs = Misc.cartesianProduct [ dotted_succ_nodes_containing_adr ; dotted_succ_nodes_containing_adr ] in
			      update_array.(i) <- List.map 
				(fun s -> let sids = List.map (EnvAutomaton.getIdFromNode automaton) s in 
				   match sids with [a;b] -> if a = b then [(a,Counter.make_counter 2, (i,a,(Counter.make_counter 2)))]
				   else 
				     [(a, Counter.make_counter 1, (i,a,(Counter.make_counter 1)) ) ; 
				      (b, Counter.make_counter 1, (i,b,(Counter.make_counter 1)) )  ]) pairs 
			  end
			else (* ctrmap.(i) = omega *)
			  (* the next states are where the omega many counters go to some next states *)
			  (* we approximate by putting omega fellows at each successor *)
			  begin
			    Message.msg_string Message.Debug "In envPost: case omega " ;	
			    let is_consistent ctrlist =
			      let r = List.fold_left (fun sofar (s, _, _) -> Region.adr_cap (EnvAutomaton.getAdrFromId automaton s) sofar) Region.true_adr ctrlist in
				not (Region.adr_is_empty (r)) in
			      if (List.length dotted_succ_nodes_containing_adr > 2) then failwith "envPost: omega and >2 succ"  else  ();
			      let all_omegas = 
				List.map (fun s -> let sid = EnvAutomaton.getIdFromNode automaton s in
					    (sid, Counter.omega (), (i,sid, Counter.omega ()) )) dotted_succ_nodes_containing_adr in 
				if is_consistent all_omegas then
				  update_array.(i) <- [ all_omegas ]
				else begin Message.msg_string Message.Debug "all omegas inconsistent" ; 
				  update_array.(i) <- List.map (fun s -> let sid = EnvAutomaton.getIdFromNode automaton s in 
								  [(sid, Counter.omega (), (i,sid,(Counter.omega() )) )]) dotted_succ_nodes_containing_adr
				end
			  end
		    end
	      end
	  done ;
	  let reduce_ctrmaps_by_dropping_inconsistent_maps ctrmap_list = 
	    List.filter (fun (ctrmap,_) ->
			   Message.msg_string Message.Debug "In drop_inconsistent. Ctrmap is" ;
			   for i = 0 to Array.length ctrmap - 1 do 
			     if not (Counter.is_zero ctrmap.(i)) then 
			       Printf.printf "ctrmap.(%d) = %s " i (Counter.toString ctrmap.(i)) 
			   done; 
			   Printf.printf "\n";
			   Message.msg_string Message.Debug ("Region is "^(Region.abstract_data_region_to_string adr));
			   let reg = ref adr in
			     for i = 1 to Array.length ctrmap - 1 do 
			       if (Counter.eq ctrmap.(i) (Counter.make_counter 0)) then () else
				 reg := Region.adr_cap (EnvAutomaton.getAdrFromId automaton i) !reg 
			     done ;
			     let a = not (Region.adr_is_empty !reg) in (*if a then Printf.printf "Not Dropping this one!\n" else () ;*) a) ctrmap_list
	  in
	    (* then do a Cartesian product of the update array to get all possible updated maps *)
	  let cp = List.map List.flatten (Misc.cartesianProduct (List.filter (fun s -> not (s = [])) (Array.to_list update_array))) in
	    (* each element in cp is a (counter map, list of transitions), where a transition (i,j,n) says n fellows went from i to j *)
	  let ctrmap_list : (Counter.counter array * ((int*int*Counter.counter) list)) list = List.map (fun m -> 
													  let a = Array.make numnodes (Counter.make_counter 0) in let who_goes_where = ref [] in
													  let rec _work lst =
													    match lst with [] -> (a, !who_goes_where)
													      | (s,ctr, (i,t,n) )::rest -> a.(s) <- Counter.plus a.(s) ctr ; if (i<>t) then (who_goes_where := (i,t,n):: !who_goes_where) ; _work rest  in
													    _work m) cp
	  in
	  let no_inconsistent = (reduce_ctrmaps_by_dropping_inconsistent_maps  ctrmap_list) in
            (* if the model checker is run with omega counters then cluster and drop the who went where info *)
	  let exists_omega (cmap,_) = let flag = ref true in let i = ref 0 in 
            while (!flag && !i< Array.length cmap) do if (Counter.is_omega cmap.(!i)) then flag := false else () ; incr i done ; not !flag in
	    if (exists_omega (List.hd no_inconsistent)) then
	      cluster_ctrmaps no_inconsistent else no_inconsistent 
	end

  let havoc_post datareg lvallist postloc =
    let foo_ht = Hashtbl.create 3 in (* fake hash table *)
    let op = C_SD.fake_edge  in
      List.fold_left (fun reg lval -> 
			let havoc_cmd = { C_SD.Command.read = [] ; C_SD.Command.written = [] ; C_SD.Command.code = C_SD.Command.Havoc lval } in
			  abstract_data_post reg havoc_cmd op postloc foo_ht)  datareg lvallist
	

  (* Solid post for an automaton *)
  let threadPost automaton ctrmap i j datareg postloc =
    (* automaton moves from location i to location j along a solid edge *)
    let inode = BiDirectionalLabeledGraph.get_node_label (EnvAutomaton.getNodeFromId automaton i) and
      jnode = BiDirectionalLabeledGraph.get_node_label (EnvAutomaton.getNodeFromId automaton j) in
      
      (* sanity checks *)
      if (i >= automaton.EnvAutomaton.number_of_states || j >= automaton.EnvAutomaton.number_of_states) then failwith "Out of bound in threadpost" ;
      if (Counter.is_zero ctrmap.(i)) then failwith "Error in threadpost 1" ;
      let ijedges = List.filter (fun e -> (BiDirectionalLabeledGraph.get_node_label (BiDirectionalLabeledGraph.get_target e)).EnvAutomaton.id = j ) 
		      (EnvAutomaton.getThreadEdges automaton i) in
	if ijedges = [] then       
	  failwith "no edge between i and j" ;
	(* it may prove a burden to separate these edges later ... *)
	let edge_havocs = 
	  List.fold_left 
	    (fun sofar thisedge -> 
	       match (BiDirectionalLabeledGraph.get_edge_label thisedge) with 
		   EnvAutomaton.Thread lvl -> Misc.union lvl sofar) [] ijedges 
	in
	  if (Region.adr_is_empty 
		(Region.adr_cap datareg (inode).EnvAutomaton.adr)) 
	  then
	    failwith "region at i does not intersect datareg " ;
	  let _get_modifies automaton ctrmap =
	    let lvlist = ref [] in
	      for i = 0 to Array.length ctrmap -1 do
		if (not (Counter.eq ctrmap.(i) (Counter.make_counter 0))) then
		  lvlist := Misc.union (EnvAutomaton.getModifiesFromId automaton i) !lvlist
	      done;
	      !lvlist
	  in

	  (* ok now do the update *)
	  (* we havoc all variables that can be modified by the environment by looking at the modifies lists of all populated states *)
	  let havoced_data = havoc_post datareg (Misc.union edge_havocs (_get_modifies automaton ctrmap)) postloc in
	  let newdatareg = Region.adr_cap havoced_data jnode.EnvAutomaton.adr in 
	    (*
	      Message.msg_string Message.Normal ("Havoced post is " );
	      Message.msg_string Message.Normal (Region.abstract_data_region_to_string havoced_data) ;
	      Message.msg_string Message.Normal ("new data reg is " );
	      Message.msg_string Message.Normal (Region.abstract_data_region_to_string newdatareg) ;
	    *)
	    (* update ctrmap by moving one guy from i to j *)
	    ctrmap.(i) <- Counter.minus ctrmap.(i) (Counter.make_counter 1) ;
	    (* now do dotted post
	    *)
	    let updctrmaps : (Counter.counter array * ((int*int*Counter.counter) list)) list = if Region.adr_leq newdatareg Region.false_adr then 
              [ Array.copy ctrmap,[] ]
            else envPost automaton ctrmap newdatareg(* jnode.EnvAutomaton.adr *) in
	      (*
		Note that envPost must be done after i has been incremented, but before
		j has been incremented, in order not to move the guy
		who just moved to j
	      *)
	      (* finally put the guy back in j *)
	      List.map (fun (updctrmap, who_goes_where) -> updctrmap.(j) <- Counter.plus updctrmap.(j) (Counter.make_counter 1)) updctrmaps ;
	      (newdatareg, updctrmaps)

  (*************************************************************************************)
  (* Done Automaton post :                                                             *)
  (*************************************************************************************)

  exception RecursionException
    
  (* the symbolic post function: (approximate) strongest post-condition *)
  (* TODO: if checking races, add havocs *)
  let post reg op phi_table =
    Message.msg_string Message.Debug "In post -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print reg ;
    Message.msg_printer Message.Debug Operation.print op ;
    Hashtbl.replace statVisitedFunctionTable (C_SD.get_location_fname (C_SD.get_target op)) true  ;
    let post_atomic_reg atomic_reg = 
      try
	if not (Region.check_locations_eq atomic_reg.Region.location  (C_SD.get_source op)) then
	  begin
	    Message.msg_string Message.Debug "WARNING: post: blast internal error suspected" ;
	    Message.msg_string Message.Debug "               operation do not match this region's location" ;
	    Message.msg_string Message.Error "WARNING: post: blast internal error suspected" ;
	    Message.msg_string Message.Error "               operation do not match this region's location" ;
	    
	    Region.bot
	  end
	else
	  begin
            try 
	      let callStack = Region.list_of_stack atomic_reg.Region.stack in
	      let tos_fname = 
		if (not (callStack = [])) 
		then (last_loc_pointer := Some (List.hd callStack);C_SD.get_location_fname (C_SD.get_target op)) 
		else ( last_loc_pointer := None; C_SD.__BLAST_DummyFunctionName) 
	      in
	      let tos_2_fname = 
		if (List.length callStack >= 1) then (C_SD.get_location_fname (List.hd callStack)) else C_SD.__BLAST_DummyFunctionName 
	      in
	      let _ = Message.msg_string Message.Debug (Printf.sprintf "tos_fname: %s, tos_2_fname: %s" tos_fname tos_2_fname) in
		(* So last_loc_pointer points to the top of the stack at all times or to errorloc Rif the stack is empty *)
	      let (new_location, new_stack) = 
                if (C_SD.isFunCall op) then
                  begin
                    match (C_SD.get_command op).C_SD.Command.code with
                        C_SD.Command.FunctionCall fcexp -> 
                          
                          let (_fname,_,target) = C_SD.deconstructFunCall fcexp in
	                  let _ = stackInfo := Some (_fname,tos_fname,true) in
                            if (_fname = PredTable.__NotImplementedFunctionName) 
                            then 
			      ((C_SD.get_target op), atomic_reg.Region.stack)
                            else   
			      begin
                                if ( C_SD.is_defined _fname) then
				  begin
				    let src = C_SD.get_source op in
				      if List.mem src callStack then 
                                        raise RecursionException
				      else
					begin
					  if (List.length callStack = !maxStackLen) then 
                                            (maxStackLen := !maxStackLen + 1;);
				          let newStack = src:: callStack in
				            ((C_SD.lookup_entry_location _fname),
				             Region.CallStack newStack)
					end
				  end
                                else 
                                  begin
                                    (* Do we need to do this? It seems abstract_data_post 
                                       havocs the return target for an unknown function, so doing
				       this here may be unnecessary! -- Rupak *)
				    (* JHALA:You are right -- let me just assume that a HAVOC will happen *)
				    Message.msg_string Message.Normal ("WARNING: Undeclared Important Function"^_fname);
				    ((C_SD.get_target op),atomic_reg.Region.stack)
				      
                                  end
			      end
                      | _ -> failwith "what the (post) !"
			  
                  end
                else 
                  try 
		    if (C_SD.isReturn op) then 
		      begin
			let _ = stackInfo := Some (tos_fname,tos_2_fname,true) in
			let (ret, _, nextLoc,subs) = lastTarget () in
			  Absutil.set_ret_pointers ret subs (C_SD.get_location_fname nextLoc);
			  let newStack =
			    List.tl callStack
			  in
			    (nextLoc,(Region.CallStack newStack) ) 
		      end
		    else 
		      begin
			let _ = stackInfo := Some (tos_fname,tos_2_fname,false) in
			  (C_SD.get_target op,atomic_reg.Region.stack) 
		      end
                  with Failure ("tl") -> 
		    begin
		      let _ = stackInfo := Some (tos_fname,tos_2_fname,false) in
			(C_SD.get_target op, atomic_reg.Region.stack ) 
		    end
	      in 
	      let get_new_data () () =
		let _new_data =
		  try 
		    match atomic_reg.Region.data with
			Region.Abstract abst_data_reg ->
			  Region.Abstract (abstract_data_post abst_data_reg (C_SD.get_command op) op new_location phi_table)
		      | Region.Concrete conc_data_reg ->
			  Region.Concrete 
			    (concrete_data_post conc_data_reg (C_SD.get_command op) op phi_table)
		  with C_SD.NoSuchFunctionException f -> failwith "Foo"
		in
		  (* Adding scoping hack *)
		  (* Adding kill *)
		  match ((Options.getValueOfBool "scope") && (C_SD.isReturn op) (*&& false*) ) with
		      true  -> 
			begin
			  (* Message.msg_printer Message.Debug C_SD.print_location (C_SD.get_source op);
			     Message.msg_printer Message.Debug C_SD.print_location new_location; *)
			  let (_,arg,_,_) = lastTarget () in
			    Stats.time "scope_kill" (scope_kill _new_data) arg
			end
		    | false -> _new_data
	      in
                (* set retTarget back to junk *)
	      let new_data = Stats.time "get_new_data" (get_new_data ()) () in
	      let _ = Absutil.reset_ret_pointers () in
		begin
		  Region.Atomic { Region.location = new_location ;
				  Region.data = new_data ;
				  Region.stack = new_stack ; }
		end
 	    with
		Failure("Empty Call Stack") -> Region.bot
	  end
      with 
          C_SD.NoSuchFunctionException f -> (* treat it as a skip *)
            Message.msg_string Message.Minor ("Unexpected NoSuchFunctionException "^f);
            (* THIS SHOULD BE COMMENTED OUT IF UNDEFINED FUNCTIONS ARE TO BE AVOIDED *)
            (* failwith ("Unexpected NoSuchFunctionException "^f); *)
            Region.Atomic
              { Region.location = C_SD.get_target op;
                Region.data = atomic_reg.Region.data;
                Region.stack = atomic_reg.Region.stack;
	      } 
              (* COMMENT OUT UPTO HERE *)
	      
        | RecursionException ->
            if (Options.getValueOfBool "enablerecursion") then
              begin
                (* treat this as a skip *)
		Message.msg_string Message.Normal "WARNING: Skipping recursive function call";
		Message.msg_string Message.Normal (Operation.to_string op) ; 
		Message.msg_string Message.Normal "For region " ;
		Message.msg_printer Message.Normal Region.print reg ; 

                Region.Atomic {
                  Region.location = C_SD.get_target op ;
                  Region.data = atomic_reg.Region.data ;
                  Region.stack = atomic_reg.Region.stack ;
                }
              end
            else raise RecursionException
    in
      (* Message.msg_string Message.Normal "checking region type"; *)
      match reg with	
	  Region.Empty -> Region.bot
	| Region.Atomic atr -> post_atomic_reg atr
        | Region.AtomicTI atrTI ->
            begin
              let prev_loc = atrTI.Region.ti_location_prev in
              let atr = 
                {Region.location = atrTI.Region.ti_location;
                 Region.data = atrTI.Region.ti_data;
                 Region.stack = atrTI.Region.ti_stack}
              in 
		match (post_atomic_reg atr) with
                    Region.Atomic atr' ->
                      Region.AtomicTI
                      {Region.ti_location_prev = prev_loc;
                       Region.ti_location = atr'.Region.location;
                       Region.ti_stack = atr'.Region.stack;
                       Region.ti_data = atr'.Region.data}
		  | Region.Empty -> Region.Empty
		  | _ -> failwith "bad return from post_atomic_reg" 
            end
	| Region.ParallelAtomic p_atomic_reg ->
            failwith "post: argument is parallel atomic: unimplemented"
	| Region.Union _ ->
            failwith "post: argument is a union: unimplemented"
	| Region.ParallelUnion _ ->
            failwith "post: argument is a parallel union: unimplemented"







  
  (** Tar post : post for automaton as well as thread *)
  let tarpost reg op automaton =
    (* JHALA: adding this function to maintain the invariant mentioned below *)

    let ctrmap_conjoin adr ctrmap =
      let conj_bdd = ref adr.Region.bdd in
      let conj_i i i_ctr =
	if Counter.is_zero i_ctr then ()
	else conj_bdd := 
	  CaddieBdd.bddAnd !conj_bdd
	    (( EnvAutomaton.getAdrFromId automaton i).Region.bdd)
      in
	Array.iteri conj_i ctrmap;
	{ Region.bdd = !conj_bdd;
          Region.lelement = DataLattice.top ;
	  Region.preds = []
	}
    in
      (* RJ: END *)
    let phi_table = Hashtbl.create 1 in
      Message.msg_string Message.Debug "In tarpost ; region is" ;
      Message.msg_printer Message.Debug Region.print reg ;
      (match op with 
	   ProgramOperation opn->
	     Message.msg_printer Message.Debug Operation.print opn ;
	     Hashtbl.replace statVisitedFunctionTable (C_SD.get_location_fname (C_SD.get_target opn)) true  
	 | AutomatonOperation (i,j) ->
	     Message.msg_string Message.Debug (Printf.sprintf "Automaton movement (%d, %d)" i j);
      )	 ;     
      match op with
	  ProgramOperation pmove ->
	    begin
              match reg with
		  Region.Empty -> [ (Region.Empty, []) ]
		| Region.ParallelAtomic p_atomic_reg ->
		    begin
		      try
			let newreg = Region.Atomic { Region.location = p_atomic_reg.Region.p_control.Region.p_location ;
						     Region.stack = p_atomic_reg.Region.p_control.Region.p_stack ;
						     Region.data = p_atomic_reg.Region.p_data ;
						   } in
			let nextreg = post newreg pmove phi_table in (* do the program move *)
			let datareg = (extract_abs_data_reg nextreg) in 
			  (* Notice that datareg is over locals and globals, and the automaton adr is over globals.
			     But we do not need to explicitly quantify out the locals from datareg since
			     we shall simply intersect with the adr and check emptiness.
			     This uses the following theorem:
			     R_1(\vec{x},\vec{y}) \cap R_2(\vec{x}) \not = \emptyset 
			     iff
			     \exists \vec{y}. R_1(\vec{x},\vec{y}) \cap R_2(\vec{x}) \not = \emptyset
			  *)
			let newctrmaps = if Region.adr_leq datareg Region.false_adr then [ (Array.copy p_atomic_reg.Region.e_control,[])]
			else envPost automaton (Array.copy p_atomic_reg.Region.e_control) datareg in (* update counter map *)
			  (*
			    Message.msg_string Message.Debug "Returned from envPost in tarpost (program move); printing new maps ";
			    List.iter (fun (nm, _) -> for i = 0 to Array.length nm -1 do 
			    Message.msg_string Message.Debug (Printf.sprintf "(%d, %s)" i (Counter.toString nm.(i))) done ) newctrmaps; 
			    Message.msg_string Message.Debug "Done printing envmaps" ;
			  *)
			  (* RJ: CRUCIAL INVARIANT !
			     in any region r in the reachability tree: "adr.r <= auto.i" if ctrmap.r.i <> 0 
			     we must conjoin the env regions which are afterall invariants as it were,
			     with the datareg -- it is the state of the system that the envautomaton is 
			     guaranteeing... the feasibility of newctrmap ensures the conjunction is satisfiable
			  *)
			let cmc = ctrmap_conjoin datareg in
			  List.map 
			    (fun (newctrmap,who_goes_where) ->
			       let (loc, stk) = Region.getPcStack nextreg in
			       let new_control = { Region.p_location = loc ;  
						   Region.p_stack = stk ; } in
			       let new_adr = cmc newctrmap in
		       		 (Region.ParallelAtomic 
				    { Region.p_control = new_control; 
				      Region.e_control =  newctrmap;
				      Region.p_data = Region.Abstract new_adr;}
				    , who_goes_where))
			    newctrmaps 
		      with Region.Empty_region -> [ (Region.bot, []) ]
		    end
		| _ -> failwith "tarpost expects a parallel atomic region"
            end
	| AutomatonOperation (i,j) ->
	    begin
	      match reg with
		  Region.Empty -> [ (Region.Empty, []) ]
		| Region.ParallelAtomic p_atomic_reg ->
		    (* update the automaton *)
		    (* recall that threadPost updates the dotted transitions of the automaton as well *)
		    let (postloc,_) = Region.getPcStack reg in
		    let (newdatareg, newctrmaps) = threadPost automaton (Array.copy p_atomic_reg.Region.e_control) i j 
						     (extract_abs_data_reg reg ) postloc in
		    let cmc = ctrmap_conjoin newdatareg in
		      List.map (fun (newctrmap,who_goes_where) -> (Region.ParallelAtomic 
								     { Region.p_control =  p_atomic_reg.Region.p_control;
								       Region.e_control =  newctrmap;
								       Region.p_data = Region.Abstract (cmc newctrmap); 
								       (* RJ: doing the env constraining here as well *)
								     } , who_goes_where) ) newctrmaps
		| Region.Atomic ar -> (* basically compute a thread post without the environment moves *)
		    let (ind, jnd) = (BiDirectionalLabeledGraph.get_node_label (EnvAutomaton.getNodeFromId automaton i),
				      BiDirectionalLabeledGraph.get_node_label (EnvAutomaton.getNodeFromId automaton j)) 
		    in
		      if (i >= automaton.EnvAutomaton.number_of_states || j >= automaton.EnvAutomaton.number_of_states) then failwith "Out of bound in tarpost" ;
		      let ijedges = List.filter (fun e -> (BiDirectionalLabeledGraph.get_node_label (BiDirectionalLabeledGraph.get_target e)).EnvAutomaton.id = j ) 
				      (EnvAutomaton.getThreadEdges automaton i) in
			if ijedges = [] then       
			  failwith "no edge between i and j" ;
			(* it may prove a burden to separate these edges later ... *)
			let all_havocs = 
			  Misc.union (EnvAutomaton.getModifiesFromId automaton i)
			    (List.fold_left 
			       (fun sofar thisedge -> 
				  match (BiDirectionalLabeledGraph.get_edge_label thisedge) with 
				      EnvAutomaton.Thread lvl -> Misc.union lvl sofar) [] ijedges) 
			in
			let datareg = extract_abs_data_reg reg in
			let datareg' = (Region.adr_cap datareg (ind).EnvAutomaton.adr) in
			  if (Region.adr_is_empty datareg') then ([ (Region.bot, []) ]) 
			  else
			    begin
	      		      (* ok now do the update *)
			      (* we havoc all variables that can be modified by the environment 
				 by looking at the modifies lists of all populated states *)
			      let havoced_data = havoc_post datareg' (all_havocs) ar.Region.location in
			      let newdatareg = Region.adr_cap havoced_data jnd.EnvAutomaton.adr in 
			      [ ( Region.Atomic { Region.location = ar.Region.location ; 
						  Region.stack = ar.Region.stack ; 
						  Region.data = Region.Abstract newdatareg; }, 
				 []) ]
			    end
		| _ -> failwith "tarpost expects parallel atomic region"
	    end
	      

  (***************************************************************************************************)



  let statsMaxNumberOfPredsActive = ref 0
				      (* Keeps track of the number of predicates active
					 at any point, discounting scope
				      *)

  (* do_focus : get new predicates to add to the current set *)
  exception NoNewPredicatesException



  (* this is a highly stateful function! *)
  let last_no_of_preds = ref 0
  let predicate_fixpoint_reached () = 
    if (!last_no_of_preds = List.length !PredTable.tableOfPreds) then 
      begin
	(* printTableOfPreds (); *)
	true
      end
    else
      begin
	last_no_of_preds := List.length !PredTable.tableOfPreds;
	false
      end


  let do_focus abst_data_reg empty_conc_data_reg =
    Message.msg_string Message.Debug "In do_focus -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print_abstract_data_region abst_data_reg ;
    Message.msg_printer Message.Debug Region.print_concrete_data_region empty_conc_data_reg ;
    let emptyPred = empty_conc_data_reg.Region.pred in
    let allPreds = try extractUsefulPredicates emptyPred with Failure "simplify" -> raise NoNewPredicatesException in
    let oldnum = getCurrentIndex () in
    let added_preds = List.map addPred allPreds in
    let newnum = getCurrentIndex () in
      if (oldnum = newnum) then
	begin
	  Message.msg_string Message.Debug "Restarting..." ;
	  Options.setValueOfBool "localrestart" true ;
	end
      else
	begin
	  Message.msg_string Message.Debug "At least one predicate added ..." ;
	  Options.setValueOfBool "localrestart" false ;
	end ;

      let newpredlist = Misc.compact (Misc.difference (added_preds) (-1 :: abst_data_reg.Region.preds)) 
			  (* This is not correct ... as there may be a descendant along the error path 
			     that does not have a certain pred that is in this list which the 
			     ancestor ie the current data region has... if you are doing PATH that is ...*)
      in
	if (newpredlist = []) then (Message.msg_string Message.Error "No new predicates ! Unexpected ...";
				    printTableOfPreds () ;
				    if (Options.getValueOfInt "predH" < 6) then
				      raise NoNewPredicatesException;
				   );
	begin  
	  (* RJ: leave the next line as NORMAL -- reqd. for demo mode in gui *)
	  (* let _ =  Message.msg_string Message.Normal ("NEW PREDICATES FOUND \n") in *)
	  let _listOfAllActivePreds = abst_data_reg.Region.preds @ newpredlist in
	    ignore (if (List.length _listOfAllActivePreds) > !statsMaxNumberOfPredsActive then 
		      begin
			statsMaxNumberOfPredsActive := List.length _listOfAllActivePreds;
			Message.msg_string Message.Debug ("Size of pred set: "^(string_of_int (List.length !PredTable.tableOfPreds)) );
			Message.msg_string Message.Debug "Active Predicates"; 
			List.iter (function i -> Message.msg_string Message.Debug  (Predicate.toString (PredTable.getPred i)))  
			  _listOfAllActivePreds
		      end);
	    { Region.bdd = abst_data_reg.Region.bdd ;
              Region.lelement = abst_data_reg.Region.lelement ;
	      Region.preds = _listOfAllActivePreds ; }
	end


  let replace_data_region reg new_data_reg =
    match reg with 
	Region.Atomic _reg ->
	  Region.Atomic 
	  { Region.location = _reg.Region.location;
	    Region.data = new_data_reg;
	    Region.stack = _reg.Region.stack;
	  }
      | _ -> failwith ("bad call to replace_data_region")


  (* ******************* ERROR TRACE FILTER ****************************** *)

  (* the following procedure takes a trace and spits out that subtrace of it which is used to 
     get a contradiction -- it is to help in "manual" counterexample analysis *)

  let  trace_filter region_list op_list = 
    (* the above is the unsatisfiable list of regions and operations -- arranged in "past --> present" order*)
    let reg_list_r = List.rev region_list in
    let op_list_r = List.rev op_list in
    let process_op_reg (phi , map) (op, reg) = () in
      ()

  (* ******************* find_all_preds ********************************** *)
  (* this function will look at ALL subtraces of a given trace and find all predicates coming from it *)





  let quantify_and_extract_abs_data_reg reg =
    let _ = Message.msg_string Message.Debug "In qae:" in

    let quant_fn = fun i -> predicate_mentions_locals (getPred i) in
    let abst_data_reg = extract_abs_data_reg reg in
    let __l = ref [] in
    let _get_v a = 
      for i = 0 to Array.length a - 1 do if a.(i) = 2 then () else __l := i::!__l ; done in
    let _preds_in_bdd = let _ = CaddieBdd.bddForeachCube (CaddieBdd.bddSupport abst_data_reg.Region.bdd) (_get_v) in !__l in
    let _preds_to_quantify = List.filter quant_fn _preds_in_bdd in
    (* BDDFIX:let _varset = CaddieBdd.fromList bddManager _preds_to_quantify in 
    let new_bdd = CaddieBdd.exists _varset abst_data_reg.Region.bdd in *)
    let new_bdd = bdd_exist_quantify (abst_data_reg.Region.bdd) _preds_to_quantify in
      {Region.bdd = new_bdd;
       Region.lelement = abst_data_reg.Region.lelement ;
       Region.preds = []
      }


  let insert_data_reg reg dr = 
    match reg with
	Region.Atomic atomic_reg ->
	  begin
	    Region.Atomic 
	      { Region.data = dr;
		Region.location = atomic_reg.Region.location;
		Region.stack = atomic_reg.Region.stack;
	      }
	  end
      | _ -> failwith "bad call to insert_data_reg"


  (* this is called only in "block" mode ... and with noforget *)
  let focus_stamp reg =
    let abst_data_reg = extract_abs_data_reg reg in
    let cur = getCurrentIndex () in
    let present_prec = 
      try List.hd abst_data_reg.Region.preds 
      with _ -> -1 
    in
    let new_reg = 
      insert_data_reg reg 
	(Region.Abstract 
	   { Region.bdd = abst_data_reg.Region.bdd;
             Region.lelement = abst_data_reg.Region.lelement;
	     Region.preds = [cur];
	   })
    in
    if (present_prec = cur && not ((Options.getValueOfInt "craig" >= 2))) then
      begin
	dump_abs ();
	raise NoNewPredicatesException;
      end
    else
      new_reg




  (* Cheap hack: go through all the predicates along oplist and add them to the list of predicates. *)
  (* again oplist is in increasing order past --> present *)
  (* r1 -> o1 -> r2 -> o2 -> ... -> on -> rn *)

  let find_all_predicates_cheap oplist =
    let rec _more lst accum =
      match lst with
	  [] -> accum
	| a :: rest ->
            begin
              match (C_SD.get_command a).C_SD.Command.code with
		  C_SD.Command.Pred e ->
                    if e = Predicate.True then _more rest accum
                    else 
                      begin
			let ee = match e with 
			  Predicate.Not e' -> e' | Predicate.Atom _ -> e 
			| _ -> failwith "nonatomic predicate on path!" 
			in
			_more rest (ee :: accum)
                      end
		| _ -> _more rest accum
            end
    in
    let all_preds_in_path = _more oplist [] in
    let newpredlist = Misc.compact 
			(Misc.difference 
                           (List.map addPred all_preds_in_path) 
                           (-1 :: ( PredTable.get_all_pred_indexes ()))) in 
    let number = List.length newpredlist in
      Message.msg_string Message.Debug ("find_all_predicates_cheap found "^
					(string_of_int number)^" predicates.") ;
      number


  let find_all_predicates reglist oplist = 
    let list_of_predicates = ref [] in
    let oplist_r = List.rev oplist in
    let reglist_r = List.rev reglist in

    let rec _proc_sub seed ol = 
      (* take wp of seed wrt (ol.head,rl.head) -- query for unsat if unsat take preds... quit and break out ..
	 else if sat, then take region and recur on it with ol.tail, rl.tail *)
      match ol with
	  op::t -> 
	    begin 
	      let emptytable = Hashtbl.create 2 in
	      let pre_reg = pre seed op emptytable in
		if (Region.is_empty pre_reg) then 
		  begin
		    let allpreds = 
		      try extractUsefulPredicates (extract_pred pre_reg) 
		      with Failure "simplify"-> raise NoNewPredicatesException 
		    in
		    Message.msg_string Message.Debug ("Learning new predicates: ");
		    List.iter (fun x -> list_of_predicates := x::(!list_of_predicates); 
		      Message.msg_string Message.Debug ((Predicate.toString x)^" , ")) allpreds;
		    Message.msg_string Message.Debug ("From region: "^(Predicate.toString (extract_pred pre_reg)))
			(* do we or do we not recur on pre_reg ? as of now, NO *)
		  end
		else 
		  _proc_sub pre_reg t
	    end
	| [] -> ()
    in

    let rec process_subtrace ol_r rl_r = 
      match rl_r with
	  h::h1::t ->
	    let seed_reg = (replace_data_region h (Region.Concrete {Region.pred = Predicate.True})) in
	      _proc_sub seed_reg ol_r; 
	      process_subtrace (List.tl ol_r) (h1::t)
	| _ -> ()
    in
      (* this is the point where the main function is called ! *)
      (*  The stuff in this needs to be worked on ... its *very* slow ... 
	  let _ = process_subtrace oplist_r reglist_r in
	  let newpredlist = Misc.compact (Misc.difference (List.map addPred !list_of_predicates) (-1 :: ( List.map Misc.fst3 !PredTable.tableOfPreds))) in
	  Message.msg_string Message.Normal "find_all_predicates found: \n";
	  List.iter (fun x -> Message.msg_string Message.Normal (Predicate.toString x)) !list_of_predicates;
	  Message.msg_string Message.Normal ("# new preds = "^(string_of_int (List.length newpredlist)));
	  if (newpredlist = [] ) then failwith ("find_all_predicates also fails to get new predicates !!") *)

    if find_all_predicates_cheap oplist == 0 then
      begin
	let _ = process_subtrace oplist_r reglist_r in
	let newpredlist = Misc.compact 
	    (Misc.difference (List.map addPred !list_of_predicates) (-1 :: (PredTable.get_all_pred_indexes () ))) 
	in
	Message.msg_string Message.Debug "find_all_predicates found: \n";
	List.iter (fun x -> Message.msg_string Message.Debug (Predicate.toString x)) !list_of_predicates;
	Message.msg_string Message.Debug ("# new preds = "^(string_of_int (List.length newpredlist)));
	if (newpredlist = [] ) then failwith ("find_all_predicates also fails to get new predicates !!")
      end




	  
  (*******************************************************************************************************) 
	  (* this is for m4 *)
	  include(abs-boolean.ml)
	  include(abs-include.ml)
	  include(abs-focimc.ml)

  (*******************************************************************************************************) 
	  


  (*******************************************************************************************************) 
  (****this function takes the various arrays and a list of useful blocks and gets and adds new preds ****)
  (*******************************************************************************************************) 
  (*******************************************************************************************************)


  let parallel_interpolant_refine 
    extract_block conj_block_array reg_array post_reg_array
    tid_list  tid_array
    (block_l : int list) = 
    
    Message.msg_string Message.Debug "Inside parallel interpolant refine";
    (* assert (not (block_l = [])); *)
    let ord_block_l = List.sort compare block_l in
    let _ = Message.msg_string Message.Normal ("Block set: "^(Misc.strList (List.map string_of_int ord_block_l)) )
    in
    let start = List.hd ord_block_l in
    let stop = List.hd (List.rev ord_block_l) in
    let cons_l = List.map extract_block  ord_block_l in
    let start_cons = 
      if check_unsat (Predicate.And cons_l) then Predicate.True
      else conj_block_array.(start) 
    in	
      
    let start_tid = tid_array.(start) in
      (* for debugging purposes, we shall dump the interesting blocks *)
    let _ = 
      begin
	Message.msg_string Message.Normal (Printf.sprintf "num blocks = %d" (List.length block_l));
	Message.msg_string Message.Normal (Printf.sprintf "start cons = \n %s" (Predicate.toString start_cons));
	Message.msg_string Message.Normal ("Constraints");
	List.iter (fun p -> Message.msg_string Message.Debug (Predicate.toString p) ) cons_l
      end
    in
      (* RJ: on the ken-wishlist: let proof = getFociProof (start_cons::cons_l) in *)
      (* new method. Lets not add these global things everywhere.... only:
	 1. add the preds from the tids interpolant
	 2. add them from: min (the min block in interesting blocks / tid
	 3. add them to: ...
	 let k = max interesting block.
	 case: k \not \in tid -> till the max block in tid <= k
      *)

    let cut tid =
      let (this_l,env_l) = List.partition (fun i -> tid_array.(i) = tid) block_l in
	if not (this_l <> [] && env_l <> []) then
	  Message.msg_string Message.Normal "All cons in one tid"
	else
	  let (s_c,e_c) = if tid = start_tid then ([start_cons],[]) else ([],[start_cons]) 
	  in
	  let cons_l_tid = s_c @ (List.map extract_block this_l) 
	  and cons_l_env = e_c @ List.map extract_block env_l
	  in
	  let foci_preds = 
	    (* JHALA: temporary patch for segfaults! *)
	    match (check_unsat (Predicate.And cons_l_tid)) || (check_unsat (Predicate.And cons_l_env)) with
		true -> []
	      | false -> extract_foci_preds (cons_l_tid,cons_l_env) 
	  in 
	  let _ = 
	    Message.msg_string Message.Normal "extraction done";
	    List.iter (fun i -> Message.msg_string Message.Normal (string_of_int i)) foci_preds
	  in 
	  let _ = 
	    if (Options.getValueOfInt "craig" >= 2) then 
	      begin
		(* now this maneuver where one only adds the preds to the relevant location *)
  		let ftid tid i = if tid_array.(i) = tid then Some (i) else None in 
		let (tid_min) =  (Misc.get_first (ftid tid) ord_block_l) in
		  match (tid_min) with
		      (None) -> ()  (* this tid does not contribute to unsat *)
		    | (Some this_tid_start) ->
			begin
			  for i = this_tid_start to stop do
			    if (tid_array.(i) = tid) then
			      begin  
				Message.msg_string Message.Debug ("Accessing i= "^(string_of_int i));
				
				let pre_loc = C_SD.location_coords (location_coords_of_region (reg_array.(i))) in
				let post_loc = C_SD.location_coords (location_coords_of_region (post_reg_array.(i))) in
				  PredTable.add_local_preds pre_loc foci_preds;
				  PredTable.add_local_preds post_loc foci_preds
			      end
			  done;
			end
	      end
	  in
	  let _ =
	    if (Options.getValueOfInt "craig" < 2 || Options.getValueOfBool "gpred") then
	      begin
		let to_add = 
		  List.filter
		    (fun p_i -> not ( List.mem p_i !PredTable.globally_useful_preds)) foci_preds 
		in 
		  Message.msg_string Message.Debug "Adding to gpredtable "; 
		  List.iter 
		    (fun p_i -> 
		       PredTable.block_a_t_flag := true; 
		       Message.msg_string Message.Debug (string_of_int p_i);
		       PredTable.globally_useful_preds := p_i :: !PredTable.globally_useful_preds) 
		    to_add  
	      end
	  in ()

    in
      List.iter cut tid_list




  (********************************* THE E-TAR refinement procedures *******************************)
  (******************************* Begin the e-tar refinement procedures ***************************)
  (* copying a bunch of functions from the lazymodelchecker file to here *)


  (* sq_loc tells you the position at some time and sq_edge, the last edge traversed to get to that squished loc
     the option is None when the move is a stutter *)

  type sq_edge_t = {sq_loc : int ; sq_edge : (int * int * EnvAutomaton.aedge_label) option }
      (* an invariant about the above -- if sq_edge is not None, then the snd is the same as sq_loc *)

  let get_shrink_node_id sn = 
    (BiDirectionalLabeledGraph.get_node_label sn).EnvAutomaton.id

  let concretize_shrink_node env_auto sn_id = 
    (*let sn_id = get_shrink_node_id sn in *)
    let onode_id_list = 
      try Hashtbl.find env_auto.EnvAutomaton.a_to_o_map sn_id 
      with Not_found -> failwith "bad env_auto -- cannot concretize"
    in
      (* List.map (Hashtbl.find env_auto.EnvAutomaton.o_map) *) onode_id_list
      (* for now, return only the ids, not the actual nodes *)

  let race_lval () = 
    match !check_races with
	None -> failwith "No race variable!!"
      | Some x -> x 

  let type_of_edge e_label =  
    let gvars = 
      (race_lval ())::(!PredTable.globally_important_lvals) 
    in
      match e_label with
	  ProgramOperation op -> 
	    let this_lvals = (snd (C_SD.reads_and_writes_at_edge op)) in
	      EnvAutomaton.Thread 
		(List.filter (fun l -> List.mem l this_lvals) gvars)
	| AutomatonOperation _ -> EnvAutomaton.Env
	    

  let get_auto_node_id an = 
    (BiDirectionalLabeledGraph.get_node_label an).EnvAutomaton.oid


  let abstract_auto_node env_auto an_id = 
    try
      Hashtbl.find env_auto.EnvAutomaton.o_to_a_map an_id
    with Not_found -> failwith "bad o_to_a_map !"


  (* I shall assume this to be the number of threads ... *)
  let num_threads = ref 1 


  (* a couple of type transducers *)

  let reggify (l) =
    let make_tuple e =
      let (n,op,n') = BiDirectionalLabeledGraph.deconstruct_edge e in
	match op with
	    ProgramOperation op' -> 
	      let n_label = BiDirectionalLabeledGraph.get_node_label n in
	      let n'_label = BiDirectionalLabeledGraph.get_node_label n' in
		Some (n_label.EnvAutomaton.oregion, op', n'_label.EnvAutomaton.oregion)
	  | _ -> None
    in
      Misc.map_partial make_tuple l

  (*************************************************************************)

  (* this next function will return "ct_l" list of sequences of ops for each thread *) 

  let env_consistency_check thread_posn_array env_map = 
    let loc_table = Hashtbl.create 17 in
    let proc_ctr = ref 0 in
    let check_eq loc num = 
      proc_ctr := !proc_ctr + num;
      env_map loc = num 
    in
    let gather i {sq_loc = loc} =
      if i <> 0 then
	if Hashtbl.mem loc_table loc 
	then
	  let present_count = Hashtbl.find loc_table loc in
	    Hashtbl.replace loc_table loc (present_count + 1)
	else
	  Hashtbl.replace loc_table loc 1
    in
      Array.iteri gather thread_posn_array;
      try
	let rv = Misc.hashtbl_forall check_eq loc_table in
	  rv && (!proc_ctr = !num_threads)
	    
      with e -> failwith "env_consistency raises exception" 
	
  (* this below returns an array such that tau.i.j tells you where process j was at time i,
     as well as the last operation it took to get there ... *)

  let compute_tau env_aut trace target_list = 
    (* trace will be used to build ct 
       first, we build trace', the sequence of automaton states *)
    Message.msg_string Message.Normal "Building Tau";
    let deconstruct_p_trace trace = 
      let extract_env_map reg = 
	let env_arr = 
	  match reg with
	      Region.ParallelAtomic pat -> pat.Region.e_control
	    | _ -> failwith "bad call to extract_env_map"
	in
	  (* RJ: Warning -- arrays are updated in place so this function could do funky things if one isn't careful ...
	     ie. if somehow the underlying array has changed for some reason ... *)
	  (fun i -> try Counter.to_int (env_arr.(i)) with _ -> failwith "bad counter map in decon_pt")
      in
      let extract_edge_map (opk,eml) = 
	let et = Hashtbl.create 37 in
	let _ = List.iter 
		  (fun (j1,j2,c) -> Hashtbl.replace et (j1,j2,EnvAutomaton.Env) 
		     (Counter.to_int c))
		  eml
	in
	let _ =
	  match opk with
	      ProgramOperation _ -> ()
	    | AutomatonOperation (j1,j2) -> 
		Hashtbl.replace et (j1,j2,EnvAutomaton.Thread []) 1 (* dont look at the actual
								       lvals modified *)
	in
	  et
      in
	(List.map extract_env_map (fst trace), List.map extract_edge_map (snd trace))
    in

    let (eloc_map_l, eedge_map_l ) = deconstruct_p_trace trace in 

    let trace_length = List.length eedge_map_l in 
      (* RJ: Is this the num of ops or regs which is the former + 1 ? 
	 -- say ops -- then its perfect -- so assume that trace_length is the number of ops *)
    let shrink_root = env_aut.EnvAutomaton.autom_root in
    let shrink_root_id = get_shrink_node_id shrink_root in
      
    let init_loc = {sq_loc = shrink_root_id; sq_edge = None} in
    let tau = Array.make_matrix (trace_length + 1) (!num_threads) init_loc in
      (* building target information *)
    let tids_at_shrink_loc l = 
      Misc.array_filter (fun a -> a.sq_loc = l) tau.(trace_length) 
    in

      
    (* this is the array that we shall fill up. tau.i.j contains the squished state that thread j is in 
       at time i. That squished state is of course, merely an int *)
      

    (* RJ: can do a sanity check here that in the first elt of the trace, the region has all the
       env threads at the start position *)
    let _ =  assert (env_consistency_check (tau.(0)) (List.hd eloc_map_l)) in
    let proc_step i (loc_map,edge_map) =
      
      Message.msg_string Message.Normal ("Proc step "^(string_of_int i));
      Array.iteri (fun i a -> Printf.printf "Tid: %d  SQ_loc: %d \n" i a.sq_loc) tau.(i-1);
      let edge_print (j1,j2,e) num = 
	let s =
	  match e with
	      EnvAutomaton.Thread [] -> "thread"
	    | _ -> "Env"
	in      
      	  Printf.printf "Edge (%d, %d, %s) : %d \n" j1 j2 s num
      in 
	Hashtbl.iter edge_print edge_map;
	let p_table = Hashtbl.create 37 in
	let _ = assert (env_consistency_check (tau.(i-1)) (loc_map)) in
	  Array.iteri (fun j loc -> Misc.hashtbl_update p_table (loc.sq_loc) j) tau.(i-1);
	  
	  let mop_up loc plist = 
	    List.iter (fun j -> tau.(i).(j) <- {sq_loc = loc; sq_edge = None}) plist
	  in
	  let proc_edge (j1,j2,etype) num_proc_using_edge = 
	    (* let num_proc_using_edge = try Hashtbl.find edge_map (j1,j2,etype) with Not_found -> 0 in *)
	    let proc_at_source = try Hashtbl.find p_table j1 with Not_found -> [] in
	    let (go,stay) = 
	      try
		Misc.cut_list_at_k proc_at_source num_proc_using_edge 
	      with _ -> failwith "Not enough threads at squished state !"
	    in
	      (* after the step i we are at the ith point in the trace -- the trace starts at t=0, steps are 1,2,3 ... *)
	      (* so we shall just say the threads that "go"ed end up at j2 at the i'th step *)
	      List.iter (fun tid -> tau.(i).(tid) <- {sq_loc = j2; sq_edge = Some (j1,j2,etype)}) go;
	      Hashtbl.replace p_table j1 stay 
		(* only these are left, the rest have been assigned already *)
	  in
	    
	    Hashtbl.iter proc_edge edge_map;
	    (* the processes that are still at a location are presumably stuttering ... 
	       so we put that into their info like so ...*)
	    Hashtbl.iter mop_up p_table;
	    (i+1) (* for fold_left purposes *)
    in
    let lmem_l = List.combine (List.tl eloc_map_l) (eedge_map_l) in
    let _ =   List.fold_left proc_step 1  lmem_l in
      (* now the tau is loaded *)
    let target_map = 
      match target_list with
	  [] -> -1
	| [h] -> 
	    let tasl = (tids_at_shrink_loc (abstract_auto_node env_aut (get_auto_node_id h))) in
	      if tasl = [] then failwith "no tid at loc -- error in tau-generation! "
	      else 
		let i = List.hd tasl in
		  i
	| _ -> failwith "bad target_list"
    in
      (tau,target_map)
	(* the concrete trace for each thread is a list of (c_edge * region),
	   where type c_edge = Move of Opkind | Stutter
	   and region is, well, region -- though with the envmaps projected 
	   away, since the whole thing is constructed as a path through the orig auto
	   which is a projection of the reachability tree 
	*) 

  type visited_type = In | Out
      
  let sq_loc_to_string sq_l = 
    Printf.sprintf "[ sq_l = %d ]" sq_l

  let sq_edge_to_string sq_e = 
    match sq_e with 
	None -> ""
      | Some (j1,j2,e) ->
	  let s = 
	    match e with
		EnvAutomaton.Thread _ -> "thread" 
	      | _ -> "env" 
	  in
	    Printf.sprintf "Edge (%d, %d, %s)" j1 j2 s

  let  sq_to_string sq = 
    (sq_loc_to_string sq.sq_loc)^"  "^(sq_edge_to_string sq.sq_edge)

  let print_shrink_path a_ =
    Message.msg_string Message.Normal "Shrink_path:";
    let pr sq = 
      Message.msg_string Message.Normal (sq_to_string sq)
    in List.iter pr a_

  let oauto_node_to_string onode = 
    string_of_int ((BiDirectionalLabeledGraph.get_node_label onode).EnvAutomaton.oid)

  let oauto_edge_to_string oedge = 
    let (s,_,t) = BiDirectionalLabeledGraph.deconstruct_edge oedge in
      Printf.sprintf "%s --> %s" (oauto_node_to_string s) (oauto_node_to_string t)

  let findpath env_auto a_ start target = 
    let abstract_auto_node n = abstract_auto_node env_auto (get_auto_node_id n) in
    let target_id = match target with Some targ -> get_auto_node_id targ | None -> -1 in
    let stamped_table = Hashtbl.create 1009 in (* this is the stamped table *)
      (* here a is a squished node and s is a concrete node *)
    let local_out_edges s = 
      let a_id = abstract_auto_node s in
	List.filter 
	  (fun e' -> abstract_auto_node ( (BiDirectionalLabeledGraph.get_target e')) = a_id)
	  (BiDirectionalLabeledGraph.get_out_edges s)
    in
    let local_successors s = 
      List.map BiDirectionalLabeledGraph.get_target (local_out_edges s)
    in

    let local_connect = BiDirectionalLabeledGraph.local_connect local_out_edges in
      

    let get_succ etyp a_id s = 
      let is_solid etyp = 
	match etyp with
	    EnvAutomaton.Thread _ -> true
	  | _ -> false
      in
	(* here etyp is either Thread [] or Env ... -- lval mod info is NOT there ..*)
	List.filter 
	  (fun e -> 
	     (abstract_auto_node (BiDirectionalLabeledGraph.get_target e) = a_id )	   && 
	     (is_solid (type_of_edge 
			  (BiDirectionalLabeledGraph.get_edge_label e))    
	      = (is_solid etyp)))
	  (BiDirectionalLabeledGraph.get_out_edges s)
    in
    let frontier s a =
      let a_id = a.sq_loc in
      let fnodes = BiDirectionalLabeledGraph.collect_component local_successors [s] in
	(* now take only the fellows that have an edge leaving the component *)
      let is_frontier_node s' = 
	(List.exists (fun s'' -> abstract_auto_node s'' <> a_id) 
	   (BiDirectionalLabeledGraph.get_children s'))
	|| (get_auto_node_id s' = target_id)
      in (* the final target is always a frontier -- should you wish to come there and wait *)
      let fnodes = List.filter is_frontier_node fnodes in
      let forced_local_connect s n = 
	match local_connect s n with
	    Some el -> el
	  | None -> failwith "forced_local_connect fails! bad source-target ?"
      in
	List.map (fun n -> ((forced_local_connect s n),n)) fnodes

    in
    let pathlen = List.length a_ in

    let rec _findpath a_ i s = 
      let s_id = get_auto_node_id s in 
	(* here s is an auto node *)
	if Hashtbl.mem stamped_table (s_id,i,In) then None (* no path from here ... *)
	else 
	  begin
	    Hashtbl.replace stamped_table (s_id,i,In) true;
	    match a_ with
		[] -> failwith "THIS SHOULD NOT HAPPEN -- bad a_"
	      | [a] -> 
		  begin
		    match target with
			Some targ ->
			  begin
			    Message.msg_string Message.Normal ("findpath--lasttarget: "^(string_of_int ((BiDirectionalLabeledGraph.get_node_label targ).EnvAutomaton.oid) ));
			    match local_connect s targ with
				None -> None
			      | Some el -> 
				  begin
				    let ps = Misc.strList (List.map (oauto_edge_to_string) el) in
				      Message.msg_string Message.Normal ("connectpath: "^ps^":");
				      Some [el]
				  end
			  end
		      | None -> 
			  begin
			    Message.msg_string Message.Normal "findpath--no last target";
			    Some [[]] (* It needs to do nothing *)
			  end
		  end
	      | a::a'::_ -> 
		  let _ = assert (a.sq_loc = 0 || (Hashtbl.find env_auto.EnvAutomaton.o_to_a_map s_id) = a.sq_loc) in
		  let continue_path (squiggle, n') = 
		    let n'_id = get_auto_node_id n' in
		      if Hashtbl.mem stamped_table (n'_id,i,Out) then None
		      else
			begin
			  Hashtbl.replace stamped_table (n'_id,i,Out) true;
			  let rest_a_ = List.tl a_ in
			    match a'.sq_edge with
				None -> (* a stutter edge *) 
				  begin
				    match _findpath (rest_a_) (i+1) n' with
					None -> None
				      | Some t' -> Some (squiggle::t')
				  end
			      | Some (j1,j2,ety) ->
				  begin
				    let e_l = get_succ ety j2 n' in
				    let targs = List.map BiDirectionalLabeledGraph.get_target e_l in
				      match Misc.get_first (_findpath rest_a_ (i+1)) targs with
					  None -> None
					    
					| Some [] -> failwith "assert: this cannot happen in continue path"
					| Some [[]] ->
					    begin
					      (* this happens if:
						 1. a' is the last fellow in the sq_trace AND
						 2a. there is no target 
						 || 2b. there is a target and there is an edge in e_l 
						 whose succ is target *)
					      if (List.length rest_a_ <> 1)
					      then 
						begin
						  print_shrink_path rest_a_;
						  failwith "I was wrong. RJ." 
						end

					      else
						let e' = 
						  match target with
						      Some targ ->
							begin
							  try
							    List.find (fun e -> (BiDirectionalLabeledGraph.get_target e = targ)) e_l with _ -> failwith "I was really wrong. RJ"
					    		end
						    | None -> 
							begin
							  try List.hd e_l
							  with _ -> failwith "but but ..."
							end
						in
						  Some ((squiggle@[e'])::[[]])
					    end

					| Some (t') ->  
					    let e' = 
					      match Misc.hd_ll t' with
						  Some e'' ->
						    begin
						      try
							List.find 
							  (fun e -> BiDirectionalLabeledGraph.get_target e = BiDirectionalLabeledGraph.get_source e'') 
							  e_l 
							  (* as an abstract edge was taken -- it means you have to 
							     find a successor here *)
						      with Not_found -> failwith "list_find_fails!"
						    end
						| None -> List.hd e_l 
					    in
					      Some ((squiggle@[e'])::t')
				  end
			end
		  in
		    if a'.sq_loc = 0 then 
		      (* i.e. the next state is the stuck state -- we've done i-1 moves so we need n-(i-1) more *)
		      begin
			let rec bl k = 
			  if k = 0 then [] else []::(bl (k-1))
			in
			  Some (bl (pathlen - i + 1))
		      end
		    else
		      let outl = frontier s a in
			Misc.get_first continue_path outl
	  end
    in
      _findpath a_ 1 start
	
  let get_thread_trace trace = 
    let (r_l,e_l) = trace in
    let reg_pairs_l = Misc.make_list_pairs (List.map Region.project_par_atomic r_l) in
      List.map2 (fun (r1,r2) (opk,_) -> 
		   match opk with
		       ProgramOperation op -> (0,[(r1,op,r2)])
		     | _ -> (0,[])
		)
	reg_pairs_l e_l
	

  let auto_node_to_string {EnvAutomaton.oid = i ; EnvAutomaton.oadr = adr} =
    Printf.sprintf "%d $$ %s" i (Region.abstract_data_region_to_string adr)

  let auto_edge_to_string op =
    match op with
	ProgramOperation op -> Operation.to_string op
      | AutomatonOperation (i,j) -> (Printf.sprintf "AutomatonMove(%d,%d) " i j)





  let dumpauto env_auto j a_ = 
    Message.msg_string Message.Normal "Dumpauto: Printing Automaton:";
    let init_node = env_auto.EnvAutomaton.o_auto_root in
    let _  =  BiDirectionalLabeledGraph.output_graph_dot 
		auto_node_to_string 
		auto_edge_to_string 
		(Options.getValueOfString "auto") [init_node] in
    let _ = Message.msg_string Message.Normal "Sh_el_info:" in
    let il_to_s il = Misc.strList (List.map string_of_int il) in
      Hashtbl.iter (fun i elts_l -> Message.msg_string Message.Normal
		      (Printf.sprintf "%d : %s" i (il_to_s elts_l))) 
	env_auto.EnvAutomaton.a_to_o_map

	

  let build_g_trace env_auto trace target_list = 
    let time_list = 0::(Misc.make_list (List.length (snd trace))) in
    let oroot = env_auto.EnvAutomaton.o_auto_root in
      
    let target_map i tid = 
      if tid <> i then None 
      else
	if target_list = [] then None
	else Some (List.hd target_list) 
    in
      
    let get_ethread_ct tau target_tid j =
      let a_ = try List.map (fun i -> tau.(i).(j)) time_list with _ -> failwith "offbyone?" in
	match (findpath env_auto a_ oroot (target_map target_tid j)) with
	    Some l -> 
	      let ps = Misc.strList (List.map (fun p -> Misc.strList (List.map oauto_edge_to_string p)) l )  in
	      let _ = Message.msg_string Message.Normal ("Path: tid "^(string_of_int j)^" : "^ps) in
		List.map (fun a -> (j+1,(reggify a))) l (* to deal with the start by zero issue *)
		  (* RJ: I leave b as a list as well -- a real edge will be a list of length 1,
		     but for stuttering you put an empty list -- I'm getting tired of them options *)
	  | None -> 
	      begin
		dumpauto env_auto j a_ ;
		failwith ("Cannot find witness trace for thread "^(string_of_int j))
	      end
    in
    let (tau,target_tid) = compute_tau env_auto trace target_list in
    let ct0 = get_thread_trace trace in 
    let ct_env = List.map (get_ethread_ct tau target_tid) (0::(Misc.make_list (!num_threads -1))) 
    in (* now ct0 may have length 1 less than each list in ct_env as the env can do one last squiggle ... so make kcombine robust! *)
    let serial_trace = Misc.kcombine (ct0::ct_env) in (*   (int * (reg,op,reg) list) list *)
      List.flatten (List.map (fun (i,l) -> List.map (fun x -> (i,x)) l) serial_trace) 
	
  (* RJ:  what does each "edge" have above ? Since we have to homogenize it ... make an edge a (reg,op,reg) ... 
     where by "reg" I mean an Atomic region *)



  (********************************)

  let parallel_block_analyze_trace env_auto trace target_list number_of_threads =
    let _ = Message.msg_string Message.Normal "In par_block_analyze_trace" in
    let _ = num_threads := Counter.to_int number_of_threads in (* should be a finite number *)
    let _ = (PredTable.block_a_t_flag := false) in 
      (* to check if any new info has been added *)
    let serial_trace  = build_g_trace env_auto trace target_list in 
      (* the above is now an (int * (r,op,r')) list *)  
      (* Phase 1: make the giant arrays corresponding to the whole trace               *)
      (* here "g" stands for "giant" or globbed *)
    let len_op_l = List.length serial_trace in
    let g_op_l = List.map (fun (_,(_,op,_)) -> op) serial_trace in
    let g_reg_l = List.map (fun (_,(r,_,_)) -> r) serial_trace in
    let g_post_reg_l = List.map (fun (_,(_,_,r)) -> r) serial_trace in
    let g_tid_l = List.map (fun (tid,_) -> tid) serial_trace in 
    let print_strace (tid,(_,op,_)) =
      Message.msg_string Message.Normal ("Thread"^(string_of_int tid));
      Message.msg_string Message.Normal (Operation.to_string op);
    in
    let _ = List.iter print_strace serial_trace in
      if (Options.getValueOfBool "stop") then failwith "FALSE ERROR CHOKE" ;
      let g_tid_array = Array.of_list g_tid_l in
      let g_op_array = Array.of_list g_op_l in
      let g_reg_array = Array.of_list g_reg_l in
      let g_post_reg_array = Array.of_list g_post_reg_l in
      let g_conj_block_array = Array.create (len_op_l) Predicate.True in
      let g_block_array = Array.create len_op_l Predicate.True in
	(* INV: all the above arrays have length len_op_l *)
      let _ = Message.msg_string Message.Debug "pblock_a_t: building blocks start" in  
      let unsat_flag = ref false in
      let _ = block_reset () in

      let _ = Message.msg_string Message.Normal ("par_trace length = "^(string_of_int len_op_l)) in
      let posn = ref (len_op_l - 1) in 
      let unsat_somewhere = ref false in
      let parallel_cons_folder (tid,(pre_reg,op,post_reg)) () = 
	Message.msg_string Message.Debug ("par_cons_folder !posn = "^(string_of_int !posn));
	Message.msg_string Message.Debug ("par_cons_folder tid = "^(string_of_int tid));
	if (not (!unsat_flag)) then
	  begin
	    let atomic_post_reg = 
	      match post_reg with
		  Region.Atomic atr -> atr
		| _ -> failwith "bad region given to cons_folder!"
	    in
	    let fname = reg_function_name pre_reg in
	    let block_cons = (block_concrete_data_pre fname tid atomic_post_reg (C_SD.get_command op) op) 
	    in
	    let conj_block_cons = pred_lval_map tid (extract_pred pre_reg) in
	    let _ = Message.msg_string Message.Normal ("posn:"^(string_of_int !posn)) in
	    let _ = Message.msg_string Message.Normal ("bcons"^(Predicate.toString block_cons)) in
	    let _ = Message.msg_string Message.Normal ("cbcons"^(Predicate.toString conj_block_cons)) in

	    let is_unsat = 
	      block_assert block_cons;
	      block_assert conj_block_cons;
	      let rv = TheoremProver._is_contra () in
		block_pop ();
		rv
	    in
	      Array.set g_block_array !posn block_cons;
	      Array.set g_conj_block_array !posn conj_block_cons;
	      Array.set g_tid_array !posn tid; (* so we know which thread moved at this timestep *)
	      posn := !posn - 1;
	      if (is_unsat) then
		Message.msg_string Message.Debug 
		  ("Found unsat at level"^(string_of_int (!posn +1)));
	      if (false && is_unsat && not (Options.getValueOfInt "predH" >= 6))
	      then unsat_flag := true;
	      unsat_somewhere := !unsat_somewhere || is_unsat
	  end
      in
	(* let _ = if (Options.getValueOfBool "talias") then load_lv_table g_op_l in *)
	
      let _ = List.fold_right parallel_cons_folder serial_trace ();block_reset ()
      in
	
      let _ = Message.msg_string Message.Debug "pblock_a_t: building blocks end" in
	
      (* Phase 2: Partition giant array into arrays for each thread                    *)
	
      let tid_list = (0::(Misc.make_list !num_threads)) in
      let step_list = 0::(Misc.make_list len_op_l) in  
      let partition_steps = 
	List.map 
	  (fun tid -> Misc.array_filter (fun j -> j = tid) g_tid_array) 
	  tid_list
      in 
	
      let partition_array arr = 
	Misc.array_of_list2 (Misc.array_select2 arr partition_steps) 
      in

      let par_op_array = partition_array g_op_array in
      let par_conj_block_array = partition_array g_conj_block_array in
      let par_block_array = partition_array g_block_array in
      let par_reg_array = partition_array g_reg_array in
	(* Phase 3: For each thread, call get_useful_blocks and then interpolant_refine  *)
	
      let local_refine tid =
	let _ = Message.msg_string Message.Normal ("In local_refine tid = "^(string_of_int tid)) in
	let (_ , interesting_block_sets, suppress_stuff) =
	  get_useful_blocks par_op_array.(tid) par_block_array.(tid) par_conj_block_array.(tid) 
	in 
	let _ = Message.msg_string Message.Normal ("found "^(string_of_int (List.length interesting_block_sets))^" useful block sets") 
	in
	let (suppress_array, 
	     extract_block, 
	     drop_constraint, 
	     restore_constraint, 
	     restore_all_constraints) =  suppress_stuff 
	in
	let _ = restore_all_constraints () in
	  (match (Options.getValueOfInt "craig" <> 0) with
	       true -> Misc.ignore_list_iter (interpolant_refine extract_block par_conj_block_array.(tid) par_reg_array.(tid))
	     | false -> Misc.ignore_list_iter (focus_refine extract_block par_conj_block_array.(tid) par_reg_array.(tid))
	  )  interesting_block_sets
      in
      let _ = print_string "tid_list: "; List.iter print_int tid_list in
      let _ = List.iter local_refine tid_list in
      let _ = Message.msg_string Message.Normal "finished local ref" in
	(* Phase 4: Call get_useful_blocks on the giant array *)
      let (_, interesting_block_sets, suppress_stuff) =
	get_useful_blocks g_op_array g_block_array g_conj_block_array 
      in 
      let _ = Message.msg_string Message.Normal ("Globally found "^(string_of_int (List.length interesting_block_sets))^" useful block sets") 
      in
      let (suppress_array, 
	   extract_block, 
	   drop_constraint, 
	   restore_constraint, 
	   restore_all_constraints) =  suppress_stuff 
      in
	(* Phase 5: Call parallel_interpolant_refine on the result of the above *)
      let _ = Message.msg_string Message.Normal "Now calling par_int_ref" in
      let _ = restore_all_constraints () in
      let _  = List.iter 
		 (parallel_interpolant_refine 
		    extract_block g_conj_block_array g_reg_array g_post_reg_array tid_list g_tid_array) 
		 interesting_block_sets
      in
	if (not !unsat_somewhere) then 
	  begin
	    List.iter print_strace serial_trace;
	    failwith "real error!"
	  end
	else 
	  if (not !PredTable.block_a_t_flag) then failwith "Par: No new predicates found ...!"
	  else ()

  (* TBD:
     1. "target" for each thread
     2. get_thread_trace
     3. the actual predicate extraction from the combined trace 
     4. hookup the max_count with num_threads 
     5. The aliasing thing above search for TBDPLDI *)


  (************************* End the e-tar refinement procedures  *************************)
  (****************************************************************************************)

  (* the refinment operator required by the ABSTRACTION signature *)
  let focus reg_to_refine empty_reg  =
    Message.msg_string Message.Debug "In focus -- arguments are:" ;
    Message.msg_printer Message.Debug Region.print reg_to_refine ;
    Message.msg_printer Message.Debug Region.print empty_reg ;

    match (reg_to_refine, empty_reg) with
        (Region.Atomic abst_atomic_reg, Region.Atomic empty_conc_atomic_reg) ->
          begin
            try 
              match (abst_atomic_reg.Region.data, empty_conc_atomic_reg.Region.data) with
                  (Region.Abstract abst_data_reg, Region.Concrete empty_conc_data_reg) ->
                    Region.Atomic { Region.location = abst_atomic_reg.Region.location ;
                                    Region.data = Region. Abstract (Stats.time "do_focus" (do_focus abst_data_reg) empty_conc_data_reg) ;
                                    Region.stack = abst_atomic_reg.Region.stack ; }
		| _ -> failwith "focus: input regions should be abstract and concrete in this order"
            with NoNewPredicatesException ->
              ( Message.msg_string Message.Debug "In focus -- arguments are:" ;
                Message.msg_printer Message.Debug Region.print reg_to_refine ;
                Message.msg_printer Message.Debug Region.print empty_reg ;
                Message.msg_string Message.Debug "No new predicates found. Aborting." ;
                raise NoNewPredicatesException
              )
          end
      | _ -> begin
          Region.print (Format.std_formatter) reg_to_refine ;
          Region.print (Format.std_formatter) empty_reg ;
          failwith "focus: arguments should be atomic regions"
        end


  (* the following function takes a concrete data region and a list of predicates and returns an 
     abstract region corresponding to it. Cousot would call it alpha *)

  let abstract atomic_conc_reg pred_list =
    let assumption  =
      match atomic_conc_reg.Region.data with
          Region.Concrete cr -> cr.Region.pred
        | _ -> failwith ("bad call to atomic_conc_reg")
    in
      assume_pred_ref := Some(assumption);
      let abstract_bdd = postBdd PredTable.bddOne (List.map (function i -> (i,PredTable.getPred i,Predicate.negate (PredTable.getPred i),true)) pred_list) [] 
      in
        assume_pred_ref := None;
        Region.Abstract { Region.bdd = abstract_bdd; Region.lelement = DataLattice.top; Region.preds = pred_list}


  let intersect_with_abstraction reg1 reg2 =
    (* reg1 : abstract, reg2 : concrete or abstract, returns reg1 cap reg2 : abstract *)
    let reg2' = precise reg2 in
      match (reg1,reg2') with
          (Region.Atomic atr1, Region.Atomic atr2) ->
            begin
	      Message.msg_string Message.Debug ("intersect_with_abstraction:\n" ^ 
						"    reg1 = " ^ Region.to_string reg1 ^
						"    reg2 = " ^ Region.to_string reg2);
              match (atr1.Region.data,atr2.Region.data) with
		  (Region.Abstract abst_1,Region.Concrete conc_2) ->
                    let preds = abst_1.Region.preds in
                    let abst_2 = abstract atr2 preds in
                    let newreg2 = Region.Atomic { Region.location = atr2.Region.location;
                                                  Region.data = abst_2;
                                                  Region.stack = atr2.Region.stack; }
                    in
                      Region.cap reg1 newreg2
		| _ -> failwith "bad call to intersect_with_abstraction"
            end
        | _ -> failwith "bad call to intersect_with_abstraction 2"

  let print_abs_stats () = 
    Message.msg_string Message.Debug ("TP Queries:" ^ (string_of_int !statActualQueriesForPost));
    Message.msg_string Message.Debug ("Reached Cubes:" ^ (string_of_int !statReachedCubes));
    ()


  let _tar_enabled_ops automaton reg =
    Message.msg_string Message.Debug "In tar_enabled_ops" ;
    match reg with
	Region.ParallelAtomic p_atomic_reg ->
	  begin
	    let get_autom_actions automaton ctrmap =
              let succs = ref [] in
		for i = 0 to (Array.length ctrmap -1) do
                  if not (Counter.eq ctrmap.(i) (Counter.make_counter 0)) then
                    let thread_succs = List.map (fun e -> AutomatonOperation (i,EnvAutomaton.getIdFromNode automaton (BiDirectionalLabeledGraph.get_target e))) 
					 (EnvAutomaton.getThreadEdges automaton i) in
                      succs := if (Counter.is_omega ctrmap.(i)) then !succs @ thread_succs else thread_succs @ !succs
                  else () 
		done ;
		!succs
            in
	    let progops = List.map (fun o -> ProgramOperation o) (C_SD.get_outgoing_edges p_atomic_reg.Region.p_control.Region.p_location) in
              if (is_atomic reg (* p_atomic_reg.Region.p_control.Region.p_location*)) then progops 
              else
		begin
		  let get_atomics = 
                    let atomic_ones = ref [] in
                      for i = 0 to Array.length p_atomic_reg.Region.e_control -1 do
			let cnt_i = p_atomic_reg.Region.e_control.(i) in
			  if (not (Counter.eq cnt_i (Counter.make_counter 0)) && 
			      (BiDirectionalLabeledGraph.get_node_label (EnvAutomaton.getNodeFromId automaton i)).EnvAutomaton.atomic = true) then
			    begin
			      atomic_ones := i :: !atomic_ones;
			      if Counter.geq cnt_i (Counter.make_counter 2) then 
				failwith "more than one atomic!"
			    end

			  else ()
                      done ;
                      if List.length !atomic_ones > 1 then failwith "More than one atomic operation in tar_enabled_ops!" ;
                      !atomic_ones
		  in
		    match get_atomics with 
			[] ->
			  let automops = (* for each location i in the automaton with counter value > 0, return 
					    (i,j) where j is a solid neighbor of i *)
			    (get_autom_actions automaton p_atomic_reg.Region.e_control)
			  in
			    progops @ automops
		      | [ a ] -> List.map (fun e -> AutomatonOperation (a,EnvAutomaton.getIdFromNode automaton (BiDirectionalLabeledGraph.get_target e))) 
	                  (EnvAutomaton.getThreadEdges automaton a) 
		      | _ -> failwith "cannot happen"
		end 
	  end
      | _ -> List.map (fun o -> ProgramOperation o) (enabled_ops reg) (* no env I suppose *)

  let exists_atomic automaton reg =
    match reg with
	Region.ParallelAtomic p_atomic_reg ->
	  let env_has_atomic = 
	    let atomic_flag = ref false in
	      for i = 0 to Array.length p_atomic_reg.Region.e_control -1 do
		if (not (Counter.eq p_atomic_reg.Region.e_control.(i) 
			   (Counter.make_counter 0)) && 
		    (BiDirectionalLabeledGraph.get_node_label (EnvAutomaton.getNodeFromId automaton i)).EnvAutomaton.atomic = true) 
		then
		  atomic_flag := true;
              done ;
	      !atomic_flag
	  in
            (C_SD.is_atomic p_atomic_reg.Region.p_control.Region.p_location) || env_has_atomic
      | _ -> false
	  

  let exists_multiple_tasks automaton reg = 
    match reg with
	Region.ParallelAtomic p_atomic_reg -> 
	  begin
	    let task_ctr = 
	      if (is_task reg) 
	      then ref (Counter.make_counter 1) 
	      else 
		ref (Counter.make_counter 0) 
	    in
	      for i = 0 to Array.length p_atomic_reg.Region.e_control -1 do
		if (not (Counter.eq p_atomic_reg.Region.e_control.(i) (Counter.make_counter 0)) && 
		    (BiDirectionalLabeledGraph.get_node_label (EnvAutomaton.getNodeFromId automaton i)).EnvAutomaton.task = true) 
		then
		  task_ctr := Counter.plus !task_ctr p_atomic_reg.Region.e_control.(i)
	      done ;
	      Counter.geq (!task_ctr) (Counter.make_counter 2)
	  end
      | _ -> false

  let exists_event automaton reg = 
    match reg with
	Region.ParallelAtomic p_atomic_reg -> 
	  begin
	    let event_flag = 
	      if (is_event reg) 
	      then ref true
	      else 
		ref false
	    in
	      for i = 0 to Array.length p_atomic_reg.Region.e_control -1 do
		if (not (Counter.eq p_atomic_reg.Region.e_control.(i) (Counter.make_counter 0)) && 
		    (BiDirectionalLabeledGraph.get_node_label (EnvAutomaton.getNodeFromId automaton i)).EnvAutomaton.event = true) 
		then
		  event_flag := true
	      done ;
	      !event_flag
	  end
      | _ -> false

  let op_is_task_op auto reg op = 
    match op with
	ProgramOperation _ -> is_task reg
      | AutomatonOperation (i,_) ->
	  begin
	    match reg with
		Region.ParallelAtomic p_atomic_reg -> 
		  begin
		    let i_node = EnvAutomaton.getNodeFromId auto i in
		      (BiDirectionalLabeledGraph.get_node_label i_node).EnvAutomaton.task = true
		  end
	      | _ -> failwith "this cannot happen in op_is_task_op!"
	  end
	    
  let tar_enabled_ops automaton reg = 
    let rv = _tar_enabled_ops automaton reg in
      if (Options.getValueOfBool "alt" = false || (exists_atomic automaton reg) ) then rv
      else
	(* ASSERT : no one is atomic ... *)
	(* rules: in decreasing priority 
	   1. multiple tasks -> []
	   2. >1 event -> all nothings, events
	   3. True -> everything
	*)
	begin
	  match reg with
	      Region.ParallelAtomic p_atomic_reg ->
		begin
		  if (exists_multiple_tasks automaton reg) 
		  then []
		  else 
		    if (exists_event automaton reg) 
		    then 
		      List.filter (fun op -> not (op_is_task_op automaton reg op)) rv
		    else
		      rv
		end
	    | _ -> rv     
	end

  let initialize_abstraction () =
    PredTable.load_abs ();
    if (Options.getValueOfBool "cf")=false then DataLattice.initialize ()

  (* Saving final abs from tree *)
  let save_region_abs reg = 
    try 
      let l_id = C_SD.location_coords (location_coords_of_region reg) in
      let preds = PredTable.bdd_support (extract_abs_data_reg reg).Region.bdd in
	PredTable.add_local_preds l_id preds
    with _ -> Message.msg_string Message.Normal "Warning: Save failed!"


  (* this function will allow us to create the initial and error regions
     to give to the model checking function *)
  let ac_create_region abs_conc_flag loclist stack_bool seedPreds initialStatePredicate =
    (* when abs_conc_flag is true -- then create an abstract region, else concrete *)
    (* JHALA: The init and error regions have different stacks -- 
       I dont know how to navigate the module hierarchy so as to be able to export 
       this so I will say that true means EveryStack and false means emptyStack *)
    (* add relevant variables to the table *)
    let _ = 
      if (Options.getValueOfBool "skipfun") then 
	match !check_races with
	    Some lv -> 
	      begin
		(* Message.msg_string Message.Normal 
		   ("Adding imp_lval:"^(Expression.lvalToString lv)); *)
		Hashtbl.replace PredTable.all_important_lvals_table lv true;
		PredTable.globally_important_lvals := lv::!PredTable.globally_important_lvals
	      end
	  | _ -> ()
    in
    let atom_list = 
      (Predicate.getAtoms seedPreds) @ (Predicate.getAtoms initialStatePredicate)
    in
    let ind_list = List.map (fun x -> addPred x) atom_list in 
    let pred_indexes = List.filter (fun a -> if (a> -1) then true else false) ind_list in
    let _ = PredTable.globally_useful_preds := pred_indexes in
    let _ = List.iter (fun i -> Hashtbl.add PredTable.pred_parity_table i [true]) pred_indexes 
    in
    let data_region = 
      match abs_conc_flag with
	  true ->
	    let our_bdd =
	      try
		convertPredToBdd initialStatePredicate
	      with Not_found -> failwith "create_region: Some predicate in the initial state not found in the predicate table!"
	    in
	      Region.Abstract { Region.bdd = our_bdd ; Region.lelement = DataLattice.top; Region.preds = pred_indexes ; }
	| false ->
	    Region.Concrete {Region.pred = initialStatePredicate}
    in
      if (Options.getValueOfBool "par") then
	failwith "ParallelAtomic not implemented" 
      else
	Region.Atomic {
	  Region.location = List.hd loclist ;
	  Region.data = data_region;
	  Region.stack = if stack_bool then Region.EveryStack else Region.CallStack [] ;
        }

  let create_region = ac_create_region true

			
  (* initreg is an atomic region with abstract data region *)
  let create_parallel_atomic_region initreg automaton initctr =
    let ctrmap = Array.make (automaton.EnvAutomaton.number_of_states) (Counter.make_counter 0) in
      ctrmap.(EnvAutomaton.getIdFromNode automaton automaton.EnvAutomaton.autom_root) <- initctr ;
      let ir = match initreg with Some i -> i 
        | None -> (* fake initial region for program *) Region.Atomic {Region.location = C_SD.fake_location ; 
								       Region.stack = Region.CallStack [] ; 
								       Region.data =  Region.Abstract Region.true_adr; } 
      in
	match ir with
	    Region.Empty -> failwith "create_parallel_atomic_region called with empty region!"
	  | Region.Atomic atreg ->
	      Region.ParallelAtomic { Region.p_control = { Region.p_location = atreg.Region.location ; Region.p_stack = atreg.Region.stack ; } ;
	                              Region.e_control = ctrmap ;
                                      Region.p_data = atreg.Region.data ; }
	  | _ -> failwith "create_parallel_atomic_region called with strange region"
	      
      
      
  let ti_init_region () = 
    let main_fn = try List.hd (Options.getValueOfStringList "main") with _ ->
      failwith "Empty main fn list!" 
    in
    let abs_data_reg =
      let lv_list = ref [] in
      let f lv = lv_list := lv::!lv_list in
	C_SD.iterate_all_scope_lvals main_fn f;
	let constr = 
          let e_sym_l = List.map (fun lv -> Expression.Lval (Expression.make_symvar main_fn lv)) !lv_list in
          let e_lv_l = List.map (fun lv -> Expression.Lval lv) !lv_list in
            Predicate.equate e_lv_l e_sym_l
	in
	  
	let pred_i_list =  (PredTable.get_all_pred_indexes ()) in
          abstract_predicate constr pred_i_list        
    in     
    let make_reg loc = 
      Region.AtomicTI 
	{ Region.ti_location_prev = loc; 
          Region.ti_location = loc;  
          Region.ti_stack = Region.CallStack [];
          Region.ti_data = abs_data_reg;}
    in
    let all_locations = C_SD.lookup_all_locations main_fn in
      List.map make_reg all_locations
        


  let print_stats fmt () =
    Message.msg_string Message.Normal ("Total number of queries = "^(string_of_int (!statTotalNumberOfQueries))^" ");
    Message.msg_string Message.Normal ("Total number of non-post queries = "^(string_of_int (!statCoveredQueries))^" ");
    Message.msg_string Message.Normal ("Worst-case number of post queries = "^(string_of_int (!statWorstCaseQueriesForPost))^" ");
    Message.msg_string Message.Normal ("Actual number of post queries = "^(string_of_int (!statActualQueriesForPost))^" ");
    Message.msg_string Message.Normal ("NonDC number of post queries = "^(string_of_int (!statNonDCQueriesForPost))^" ");
    Message.msg_string Message.Normal ("Assume post queries = "^(string_of_int (!statAssumeQueriesForPost))^" ");
    Message.msg_string Message.Normal ("Total posts = "^(string_of_int (!statNumPost))^" ");
    Message.msg_string Message.Normal ("Total assume posts = "^(string_of_int (!statNumAssumePost))^" ");

      Message.msg_string Message.Normal ("Total foci queries = "^(string_of_int (!TheoremProver.stats_nb_foci_queries))^" ");
   
    Message.msg_string Message.Normal ("Foci Theory Clauses:");
    (iter_theory_clauses (fun p -> Message.msg_string Message.Normal
			    (Predicate.toString p)));
    Message.msg_string Message.Normal ("List of predicates: ");
    printTableOfPreds ();
    Message.msg_string Message.Normal 
      ("Maximum number of predicates active together (discounting scope) = "^(string_of_int (!statsMaxNumberOfPredsActive))^"\n");
    Message.msg_string Message.Normal ("Functions visited:");
    Hashtbl.iter (fun x y -> print_string (x^" , ")) statVisitedFunctionTable;
    dump_abs (); ()




  (* **************** CONCURRENCY **************** *)

  (* this function will reset all the appropriate global data structures ... *)

  let reset () = 
    Hashtbl.clear theoremProverCache;
    Hashtbl.clear predTbl;
    Hashtbl.clear predTbl_deep; 
    Hashtbl.clear unchangedPredTable; 
    (* [RUPAK] Do not reset the predicates
       PredTable.tableOfPreds := [];
       predIndex := -1;
    *)
    Absutil.reset_ret_pointers ();
    last_loc_pointer := (None);
    assume_pred_ref := None;
    ()

  let exit_abstraction () = PredTable.kill_predbdd ()

  (*  let iter_global_variables = C_SD.iter_global_variables
  *)
  let iter_global_variables = C_SD.iterate_all_lvals



  type access_type = Anone | Aread | Awrite | Aboth
  let accessing op v =
    let reads, writes = C_SD.reads_and_writes_at_edge op in
      match (List.mem v reads, List.mem v writes) with
	| false, false -> Anone
	| false, true -> Awrite
	| true, false -> Aread
	| true, true -> Aboth

  let get_all_writes op =
    snd (C_SD.reads_and_writes_at_edge op)

  let get_all_reads op =
    fst (C_SD.reads_and_writes_at_edge op)

  let reads_and_writes op = C_SD.reads_and_writes_at_edge op 

  let globally_important_lvals () = !PredTable.globally_important_lvals 

  let aliasesOf lval = if can_escape lval then ( lval::(AliasAnalyzer.getAllAliases lval)) else []

  let eq_addr_constraint_region lval1 lval2 = 
    let p = 
      Predicate.Atom 
	(Expression.Binary 
	   (Expression.Eq, 
	    Expression.addressOf (Expression.Lval lval1),
	    Expression.addressOf (Expression.Lval lval2)))
    in
      predicate_to_region p

  (* this should return the list of possible aliases in the whole program, of some lval 
     maybe we should keep a Hashtbl or something and query that every time 
  *)

  let make_multiply_instantiable_region r1 r2 lval1 lval2 = 
    Message.msg_string Message.Debug "In make_multiply instantiable\n\n" ;
    Message.msg_string Message.Debug ("r1 is"^(Region.to_string r1)) ;
    Message.msg_string Message.Debug ("r2 is"^(Region.to_string r2)) ;
    let rename_fun = new_name (Absutil.brand_new_suffix ()) in
    let eq_addr_const_reg = eq_addr_constraint_region lval1 (Expression.alpha_convert_lval rename_fun lval2) in 
      Region.cap 
	(Region.cap !assumed_invariants_region eq_addr_const_reg)
	(Region.cap r1 (Region.rename_vars r2 rename_fun))
	
	
  (* the below function should ONLY be called on two regions known to be such that:
     1. lval_1, lval_2 are such that they are accessed in region r1 r2 respectively
     2. neither of the lvals is a lock.
  *)
  let multiply_instantiable r1 r2 lval1 lval2 =
    let r = make_multiply_instantiable_region r1 r2 lval1 lval2 
    in
      if Region.is_empty r then None else Some r

  let get_modifies automaton reg =
    match reg with
      | Region.Empty -> (([(* program read *)], [(*program write*)]), [])
      | Region.ParallelAtomic p_at ->
	  begin
	    let rec exists_atomic auto ctrmap i =
	      if i = Array.length ctrmap then false 
	      else
		if (not (Counter.eq ctrmap.(i) (Counter.make_counter 0) ) 
		    && (BiDirectionalLabeledGraph.get_node_label (EnvAutomaton.getNodeFromId auto i)).EnvAutomaton.atomic) then
		  true
		else
		  exists_atomic auto ctrmap (i+1)
	    in
            let is_atomic = exists_atomic automaton p_at.Region.e_control 0 in
	    let progreads_and_writes =
	      if (is_atomic || C_SD.is_atomic p_at.Region.p_control.Region.p_location) then ([],[]) 
              else (* if the current location is atomic, then don't consider its reads/writes *)
		List.fold_left
		  (fun (lvr, lvw) edge -> 
		     let (r, w) = C_SD.reads_and_writes_at_edge edge in (Misc.union r lvr, Misc.union w lvw)) ([],[])
		  ((* find all the enabled program moves at this program location *)enabled_ops reg) 
	    in
	    let lvlist = ref [] in
              if is_atomic then () 
              else begin
		for i = 0 to automaton.EnvAutomaton.number_of_states -1 do
		  if (not (Counter.eq p_at.Region.e_control.(i) (Counter.make_counter 0)) && 
		      (not ((BiDirectionalLabeledGraph.get_node_label (EnvAutomaton.getNodeFromId automaton i)).EnvAutomaton.atomic))) then
		    let autom_modifies = 
		      (List.map (fun l -> (i,l))
			 (Misc.union 
			    (EnvAutomaton.getModifiesFromId automaton i)
			    (EnvAutomaton.getModifiesOnEdge automaton i)))
		    in
		      
		      lvlist := Misc.union autom_modifies !lvlist
		done;
	      end ;
	      (progreads_and_writes, !lvlist)
	  end
      | _ -> failwith "get_modifies: requires ParallelAtomic"
	  
	  
	  
	  (* **************** END concurrency **************** *)
	  (************* FOR BLASTPCC ************************)
	  (* for m4 -- include blast-pcc.ml *)
	  include(blast-pcc.ml)

  (************* START test generation  ************************)
  let genTests reg =
    let formals = match (C_SD.lookup_formals (List.hd (Options.getValueOfStringList "main"))) with
        C_SD.Fixed f -> f | C_SD.Variable f -> f
    in
    let is_arg a =
      let _is_arg_lval l = 
        match l with
            Expression.Symbol sym -> List.mem sym formals
          | _ -> false
      in
	match a with Expression.Lval l -> _is_arg_lval l
          | Expression.Chlval (l,s) -> (* _is_arg_lval l *) false
    in
    let is_unknown a =
      let _is_unknown_lval l = 
        match l with
            Expression.Symbol sym ->  Misc.is_prefix "BLAST_UNKNOWN@" sym 
          | _ -> false
      in
	match a with Expression.Lval l -> _is_unknown_lval l
          | Expression.Chlval (l,s) -> _is_unknown_lval l
    in

    let pred = extract_pred reg in
    let sym_to_val_list = TheoremProver.generateTest pred in
      (* sort the unknowns *)
    let unknowns = List.sort (fun (a,b) (c,d) ->
                                match (a,c) with
                                    (Expression.Chlval (l,s1) , Expression.Chlval (l', s2)) -> - (String.compare s1 s2) (* sort in decreasing order, since the
                                                                                                                           ctrex analysis goes backwards *)
                                  | _ -> failwith ("Trouble sorting unknowns! "^(Expression.toString a)^(Expression.toString c))
                             )
	             (List.filter (fun (a,b) -> is_unknown a) sym_to_val_list) in
    let args = List.filter (fun (a,b) -> is_arg a) sym_to_val_list  in
      (List.map (fun a -> let ea = Expression.Lval (Expression.Symbol a) in (ea, try List.assoc ea args with Not_found -> 0.0 (* any value *))) formals  ) @ unknowns 
	

  (* END test generation stuff *)


  type spec = unit
  type spec_result =
    | Yes
    | No
    | Maybe of Ast.Predicate.predicate
  let summary_post _ = failwith "summary_post not implemented"
  let init_spec = summary_post
  let build_spec = summary_post
  let check_spec = summary_post

end

