
  
module Symbol = Ast.Symbol
module Constant = Ast.Constant
module Expression = Ast.Expression
module Predicate = Ast.Predicate
    
module C_SD = BlastCSystemDescr.C_System_Descr

(* debug flag: I want to debug this module separately *)
let debug_bdd_pts_to = ref false 
(** Code relating to initializing the BDD package.
    Defines the manager, true, and false -BDDs.
    *)
let bddManager = CaddieBdd.init 0 64 256 512 
let bddZero = CaddieBdd.bddFalse bddManager
let bddOne = CaddieBdd.bddTrue bddManager
let allocFuns = ["alloc"; "malloc"; "calloc"; "realloc"; "kmalloc" ]  @ [ "fopen"; "fdopen" ] 

    let rec fullsimp_lval lv =
	try
      match lv with
        Expression.Symbol s -> lv
      | Expression.Access(_, e,_) ->
                fullsimp_exp e
      | Expression.Dereference e ->
                fullsimp_exp e
      | Expression.Indexed(e,_) ->
                fullsimp_exp e
	with excn -> failwith ("fullsimp-lval: value "^(Expression.lvalToString lv)^" raises exception:"^(Printexc.to_string excn))
    and fullsimp_exp e =
        match e with
          Expression.Lval l -> fullsimp_lval l
        | Expression.Unary (Expression.Address, e') -> fullsimp_exp e'
(*
        | Expression.Binary(Expression.Plus, e1, e2) 
        | Expression.Binary (Expression.Minus, e1, e2) -> 
	  begin
		try
		let t1 = C_SD.get_type_of_exp e1 in
	        match t1 with
		  Cil.TPtr _ -> fullsimp_exp e1
		| _ 	-> fullsimp_exp e2
		with _ -> Message.msg_string Message.Normal ("get_type raises exception for "^(Expression.toString e)) ;failwith "foo" ;Expression.Symbol "__JUNK"
	  end
	| Expression.Constant _ -> (Expression.Symbol "__JUNK" )  
*)
        | _ -> failwith ("bddptsto[fullsimp_lval]: Strange expression "^(Expression.toString e))




(** This hash table contains the map from locations to their bdd functions *)
let location_to_bdd_map = Hashtbl.create 31

let lookup_location loc =
  (* if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal "In lookup_location" ; *)
  try
    let (a,b) = C_SD.location_coords loc in
    Hashtbl.find location_to_bdd_map (a,b)
  with Not_found ->
    begin
      (* if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal "Location was not found in the hash table" ; *)
      Hashtbl.add location_to_bdd_map (C_SD.location_coords loc) bddZero ;
      bddZero
    end
  | _ -> if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal "lookup fails" ; 
      if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("Length = "^(string_of_int (List.length (Hashtbl.fold 
	(fun key data curlist -> key :: curlist) location_to_bdd_map []))))  ; failwith "Here"

(** This hash table will contain a map from lvalues to their bdd encoding. *)
let lvalue_to_bdd_map = Hashtbl.create 31
    
(** This hash table will contain a map from edge to its allocated memory *)
let edge_to_mem_map = Hashtbl.create 31

(** This hash table will contain a map from function calls to the target locations of their edges *)
let fcn_to_target_map = Hashtbl.create 31
    
(** This hash table will contain a map from function name to the list of lvalues in them *)
let fcn_to_lvalue_map = Hashtbl.create 31
let fcn_to_lvalbdd_map = Hashtbl.create 31

let list_from_hashtbl htbl =
  Hashtbl.fold 
	(function key -> function valu -> 
		function lst -> key :: lst) htbl []

(** Code to generate new temporary variables *)

let bddPTSTONAME = "__bddptsto_mem"

let tmpVarIndex = ref (-1)

let getNextTmpVarIndex () =
  incr tmpVarIndex ; !tmpVarIndex

let resetTmpVarIndex () = tmpVarIndex := -1

let is_malloc_location key = 
  match key with
    Expression.Symbol s -> Misc.is_prefix bddPTSTONAME s
  | _ -> false
(*********************************************)
    
(** worklist contains all locations that changed in last analysis *)

(*************************************************************************)
    
(** this function initialize location_to_bdd_map and worklist I use bfs search *)
let map_nodes_fname f fname  = 
  let module LocationSet =
    Set.Make (struct
      type t = int 
      let compare x y =  compare x  y
    end) 
      in
      let already_visited_source_locations = ref LocationSet.empty in
      let rec map_loc_and_edges loc_list =
	match loc_list with
	  [] -> []
	| loc::t -> 
	    begin
	      let (_,loc_id) = C_SD.location_coords loc in
	      if (LocationSet.mem loc_id !already_visited_source_locations) 
	      then 
		(map_loc_and_edges t)
              else
		begin
		  let _ = 
		    (already_visited_source_locations := LocationSet.add loc_id !already_visited_source_locations) 
		  in
		  let children_h = List.map C_SD.get_target (C_SD.get_outgoing_edges loc) in
		  (* fixed by li. !!! if we write (f loc) :: (...) the last function will be executed first *)
		  let f_list = f loc in
		  f_list::(map_loc_and_edges (t @ children_h))
		end
	    end
      in
      map_loc_and_edges [(C_SD.lookup_entry_location fname)]

let init_bdd_location worklist =
  let node_map node =
    worklist#add_element node ;
    ()
  in
  ignore (List.map (fun fn -> map_nodes_fname (node_map) fn) (C_SD.list_of_functions ()))


let collapse_expression e =
  let rec get_ptr_lval l =
(*
    match l with
      Expression.Dereference ee ->
	begin
	  match (get_ptr ee) with
	    [] -> if !debug_bdd_pts_to then Message.msg_string Message.Normal ("collapse expression: "^(Expression.toString e)) ; []
	  | [a] -> [ Expression.Dereference (Expression.Lval a) ]
	  | x -> if !debug_bdd_pts_to then Message.msg_string Message.Normal "collapse_expression: More than one lval!"; [] 
		(* failwith ("Too many lvals for "^(Expression.toString e)) *)
	end
    | _ ->
*)
	begin
	try
	  match (C_SD.get_type_of_lval l) with
	    Cil.TPtr _ | Cil.TArray _ -> [l]
	  | _ -> []
	with _ -> []
	end
  and get_ptr ex =
    match ex with
      Expression.Lval l -> get_ptr_lval l
    | Expression.Cast (_, e') -> get_ptr e'
    | Expression.Binary (Expression.FieldOffset, e1, e2) ->
	begin
	  match e1 with
	    Expression.Lval l -> get_ptr_lval l
	  | _ -> []
	end
    | Expression.Binary (op, e1, e2) -> 
	begin
	match op with
          Expression.Plus | Expression.Minus ->
          begin
	  let p1 = get_ptr e1
	  and p2 = get_ptr e2 in
	  match (p1,p2) with
	    ([], []) -> []
	  | ([a], []) -> [a]
	  | ([], [b]) -> [b]
	  | _ -> if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("get_ptr: Too many pointers! " ^ (Expression.toString ex)); []
          end
          | _ -> []
	end
    | Expression.Unary (Expression.Address, e1) ->
	[]
    | Expression.Unary (op, e1) ->
	get_ptr e1
    | Expression.Constant _ -> []
    | Expression.Assignment _ 
    | Expression.FunctionCall _ 
    | _ -> failwith ("get_ptr : "^(Expression.toString e))
  in
  let lst = get_ptr e in
  match lst with
    [] -> e
  | [a] -> Expression.Lval a
  | _ -> failwith ("collapse expression returns too many pointers for "^(Expression.toString e))

(* this function simplify the lvalues. Structures that we can handle:
   A[i], p.f, *(&a).
   *(p + i) reduced to *p.
   Only field and index are removed. dereference info is kept. *)
let rec simplify_lval lv = 
  match lv with
    Expression.Symbol s -> lv 			
  | Expression.Access(_, e,_) -> 
	simplify_exp e
  | Expression.Dereference e ->
      (* we still need to keep the dereference info *)
      begin 
	match e with
	  Expression.Unary(Expression.Address, Expression.Lval lval) -> simplify_lval lval (* *(&e) is simplified as e *)
	| Expression.Lval lval -> Expression.Dereference (Expression.Lval (simplify_lval lval))
	| _ -> 
	    begin
	      if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal "Interesting case for dereference" ;
	      if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("Expression is "^(Expression.lvalToString lv)) ;
	      let e_ = (collapse_expression e) 
	      in
	      if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("We return * "^(Expression.toString e_)) ;
	      Expression.Dereference (e_)
	    end
      end
  | Expression.Indexed(e, _) -> 
      begin
	match e with 
	  Expression.Lval e_lv -> Expression.Dereference (Expression.Lval (simplify_lval e_lv))
		(* sth should be wrong here *)
	| _ -> lv
      end		      
 and simplify_exp e =
  match e with
   Expression.Lval l -> simplify_lval l
  | _ -> failwith ("simplify_exp: Not handled " ^ (Expression.toString e))	

(** debug functions *)
let debug_printedge source target = 
  match (C_SD.get_source_position source) with
    Some (filename, row, _) ->
      begin
	Printf.fprintf stdout "source loc at (%s:%d)\n" filename row ;
	match (C_SD.get_source_position target) with
	  Some (filename1, row1, _) ->
	    Printf.fprintf stdout "target loc at (%s:%d)\n" filename1 row1 
	| None -> () 
      end
  | None -> ()


let get_addrofs code =
  let rec get_addrofs_lval l =
    match l with
      Expression.Symbol _ -> []
    | Expression.Dereference e -> get_addrofs_exp e
    | Expression.Access (_, e, _) -> get_addrofs_exp e
    | Expression.Indexed (e1, e2) -> (get_addrofs_exp e1) @ (get_addrofs_exp e2)
  and
   get_addrofs_exp e =
    match e with
      Expression.Lval lv -> get_addrofs_lval lv
    | Expression.Chlval _ -> []
    | Expression.Assignment (_, target, exp) -> (get_addrofs_lval target) @ (get_addrofs_exp exp)
    | Expression.Binary (_, e1, e2) -> (get_addrofs_exp e1) @ (get_addrofs_exp e2)
    | Expression.Cast (_, e1) -> get_addrofs_exp e1
    | Expression.Constant _ -> []
    | Expression.Constructor _ -> []
    | Expression.FunctionCall (f1, flist) -> (get_addrofs_exp f1) @(List.flatten (List.map get_addrofs_exp flist))
    | Expression.Sizeof _ -> []
    | Expression.Unary (Expression.Address, Expression.Lval (Expression.Symbol s)) -> [ Expression.Symbol s ]
    | Expression.Unary (Expression.Address, _) -> failwith "Strange address operator"
    | Expression.Unary (_, e1) -> get_addrofs_exp e1
  in
  let rec get_addrofs_pred p =
    match p with
      Predicate.And plist 
    | Predicate.Or plist ->
	List.flatten (List.map get_addrofs_pred plist)
    | Predicate.Not notp -> get_addrofs_pred notp
    | Predicate.Atom a -> get_addrofs_exp a
    | _ -> []
  in
  let rec get_addrofs_stmt s =
    match s with 
      C_SD.Command.Expr e -> get_addrofs_exp e
    | C_SD.Command.Return _ -> [] (* there are no address-of's in returns *)
  in
  match code with
    C_SD.Command.Skip -> []
  | C_SD.Command.Block stmts -> List.flatten (List.map get_addrofs_stmt stmts)
  | C_SD.Command.Pred p -> get_addrofs_pred p
  | C_SD.Command.FunctionCall (e) -> get_addrofs_exp e
  | C_SD.Command.GuardedFunctionCall (p,e) -> (get_addrofs_pred p) @ (get_addrofs_exp e) 
  | C_SD.Command.Havoc lv -> get_addrofs_lval lv
  | _ -> failwith ("get_addrofs : code not handled") 

(** Go through all edges and put all different lvalues to lval_list. 
    For each memory allocation, we assign a new lvalue to that memory. 
    A[i] is simplied as A
    q.f ( *q).f is simplied as q
    PLUS: we keep the map from each function call to its target locations.
    *)
let fread_lvals = function edge1 -> 
  let cmd = C_SD.get_command edge1 in
  begin
		(* we read all lvalues used in this command and simplify them *)
    let tmp1 = List.append (cmd.C_SD.Command.written) (cmd.C_SD.Command.read) 
    in
    let tmp2 = get_addrofs cmd.C_SD.Command.code in (* get_addrofs gets the lvals whose address has been taken in the code.
					  We need this to catch those variables y which only appear as x= &y
					  in the code. (note that such y's are not read :-) *)
    let result_list = (List.map fullsimp_lval (tmp2 @ tmp1) )
    in
      
		(* we give each memory block a name, according to their locations *)
    match cmd.C_SD.Command.code with 
      C_SD.Command.FunctionCall(e)  -> 
	begin
	  match e with
	    Expression.FunctionCall (fnname, arglist) 
	  | Expression.Cast(_, Expression.FunctionCall (fnname, arglist))
	  | Expression.Assignment (Expression.Assign, _, Expression.FunctionCall (fnname, arglist)) 
	  | Expression.Assignment (Expression.Assign, _, Expression.Cast(_, Expression.FunctionCall (fnname, arglist))) ->
	      begin
		(* we keep a map from function call to their targets *)
		let target = C_SD.get_target edge1 in 
		match fnname with
		  Expression.Lval (Expression.Symbol s) ->
		  begin
		    Hashtbl.add fcn_to_target_map s target;
		    if (List.mem s allocFuns) then 
						(* we apply a new variable for this memory block *)
		      let getNewVar s =
			let i = getNextTmpVarIndex()
			in 
			Expression.Symbol( s^( string_of_int i ))
		      in 
		      let newMem = getNewVar bddPTSTONAME
		      in
		      Hashtbl.add edge_to_mem_map edge1 newMem ;
		      (newMem :: result_list)
		    else result_list 
		  end
		| _ -> result_list
	      end
	  | _ -> failwith "Strange function call"
	end
    | _ -> result_list
  end

let get_all_lvals () =
  resetTmpVarIndex () ;
  let unique_globals = Hashtbl.create 31 in
  let update_list curlist key = 
    if (not (is_malloc_location key)) && C_SD.is_array key then 
    begin
      match key with Expression.Symbol s -> key :: ((Expression.Symbol (s^"#aloc")) :: curlist) 
      | _ -> Message.msg_string Message.Normal ("Strange lvalue : " ^ (Expression.lvalToString key)); failwith "oh"
    end
    else (key :: curlist) 
  in
  let process_single_routine fnname = 
    let tmplevel2_list = C_SD.map_edges_fname fread_lvals fnname in
    let tmp_list = List.flatten tmplevel2_list in
    let get_formals fnname = 
      match (C_SD.lookup_formals fnname) with
	C_SD.Fixed flist -> List.map (fun s -> Expression.Symbol s) flist
      | C_SD.Variable vlist -> List.map (fun s -> Expression.Symbol s) vlist
    in
    let formal_list = get_formals fnname in
    let unique_lvals = Hashtbl.create 31 in


    let rec remove_dup_lvals tmp_list =     
      match tmp_list with
	[] -> ()
      | cur_lval :: rest ->
	  begin
	    if C_SD.is_global cur_lval then
	      begin
		try 
		  let _ = Hashtbl.find unique_globals cur_lval in
		  remove_dup_lvals rest
		with Not_found ->
		  Hashtbl.add unique_globals cur_lval 0;
		  remove_dup_lvals rest
	      end
	    else
	      begin
		try
		  let _ = Hashtbl.find unique_lvals cur_lval in
		  remove_dup_lvals rest
		with Not_found ->
			(* we add this to lval_list and hash table *)
		  Hashtbl.add unique_lvals cur_lval 0;
		  remove_dup_lvals rest
	      end	    
	  end
    in
    remove_dup_lvals (tmp_list @ formal_list );
    (* for each variable of type array, add the additional array location aloc *)
    let result_list = Hashtbl.fold 
      (fun key data curlist -> update_list curlist key)
      unique_lvals []
    in
    Hashtbl.add fcn_to_lvalue_map fnname result_list;
    result_list
  in
  let local_list = List.flatten (List.map process_single_routine (C_SD.list_of_functions ())) 
  in
  let global_list = Hashtbl.fold (fun key data curlist -> update_list curlist key) unique_globals []
  in
  (global_list, local_list, (List.map (fun s -> Expression.Symbol s) (C_SD.list_of_functions ())  ))

(*****************************************************************************)    


(*** code to map variables to bdd variables *)    

(** make source to target/target to source maps we will use later
    they should be global *)
let stvarMap = ref []
let tsvarMap = ref []
let num_bddVar = ref 0 

(** the encoding can be kept an array of bdd variable nodes. initially ^x1^x2...^xn
    Note we need two sets of bdd variables since we represent p->r 
    their indexes are kept in code1_index and code2_index*)
let encode_variables num_bddvars lval_list =   
    let cur_code = Array.make num_bddvars 0 in
    let cur_encode = Array.make num_bddvars bddZero in
    let code1_index = Array.make num_bddvars 0 in
  
  let rec go_through_list l_list =
    match l_list with
      [] -> ()
    | lvalu :: rest ->
                    (* encode this variable *)
                    (* a tutorial for me. how to use pointers in ocaml *)
	let cur_encoding = ref (Array.get cur_encode 0) in
	for i = 1 to num_bddvars - 1 do
	  cur_encoding := CaddieBdd.bddAnd !cur_encoding cur_encode.(i)
	done ;
	(* debug 
	if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal (
	Printf.sprintf "lvalue %s is encoded as \n"  
	(Expression.lvalToString lvalu));
	CaddieBdd.bddPrint !cur_encoding; *)
	Hashtbl.add lvalue_to_bdd_map lvalu !cur_encoding ;	
	            (* so we don't have break in ocaml *)
	let rec update_current_encoding i =
	  if(i>=0) then
            begin
	      cur_code.(i) <- 1 - cur_code.(i);
	      cur_encode.(i) <- CaddieBdd.bddNot(cur_encode.(i));
	      if cur_code.(i)==1 then
		()
	      else update_current_encoding (i-1)
	    end
	in			
	update_current_encoding (num_bddvars - 1) ;
		     (* after that, continue with the rest of the list *)
	go_through_list rest
  in 
  begin 
	(* initialize lvalues encoding *)
    for i = 0 to num_bddvars - 1 do
      cur_code.(i) <- 0;
      let tmpbdd =  CaddieBdd.bddNot (CaddieBdd.bddNewVar bddManager) in
      cur_encode.(i) <- tmpbdd;
      code1_index.(i) <- CaddieBdd.bddIndex tmpbdd
    done;
    go_through_list lval_list;
	
        (* encode the second set of bdd variables, they will be used for q in p->q *)
    let code2_index = Array.make num_bddvars 0 in
    for i =0 to num_bddvars - 1 do
      let tmpbdd = CaddieBdd.bddNewVar bddManager in
      code2_index.(i) <- CaddieBdd.bddIndex tmpbdd
    done;

        (* build up two mapping lists *)
    for i = 0 to num_bddvars -1 do
      stvarMap := (code1_index.(i), code2_index.(i)) :: !stvarMap;
      tsvarMap := (code2_index.(i), code1_index.(i)) :: !tsvarMap
    done
  end

(**********************************************************************************************)

(** point to analysis *)

    (* Find the bdd corresponding to the lvalue lv in the BDD bdd_loc *)
let rec get_bdd_expensive bdd_loc lv =
  match lv with 
    Expression.Symbol _ -> 
      let b = Hashtbl.find lvalue_to_bdd_map lv in
      b
  |	Expression.Dereference e -> 
      let lv_pts_to = CaddieBdd.bddCofactor bdd_loc (get_bdd_e_expensive bdd_loc e)
      in
      if (!debug_bdd_pts_to) then 
	begin
	  Message.msg_string  Message.Normal ("Points to set:" );
	  CaddieBdd.bddPrint lv_pts_to 
	end;
      let b = CaddieBdd.replace lv_pts_to (CaddieBdd.toMap bddManager !stvarMap) in
      if (!debug_bdd_pts_to) then 
	begin
	  Message.msg_string  Message.Normal ("In terms of original vars:" );
	  CaddieBdd.bddPrint b 
	end;
      b
  |	Expression.Access(Expression.Dot, e, _) ->
      get_bdd_e_expensive bdd_loc e
  |	Expression.Indexed (e1, _) ->
      get_bdd_e_expensive bdd_loc e1
and
    get_bdd_e_expensive bdd_loc e =
  match e with
    Expression.Lval l -> get_bdd_expensive bdd_loc l
  |  Expression.Binary (Expression.Plus, e1, e2)
  |  Expression.Binary (Expression.Minus, e1, e2) ->
	begin
(*
	  let c = collapse_expression e in
	  if c = e then get_bdd_e bdd_loc e1 else
	  get_bdd_e bdd_loc (c)
*)
	(* RUPAK: TBD: Unsound *)
	  get_bdd_e_expensive bdd_loc e1
	end
  |	Expression.Unary (Expression.Address, _)
  |	_ -> failwith ("checkAliasFI: cannot handle "^(Expression.toString e))

	

(** This function Or the bdd function at source to the bdd function at target *)
    (* Find the bdd corresponding to the lvalue lv in the BDD bdd_loc *)
let get_bdd bdd_loc lv =
  if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("In get_bdd : lv is "^(Expression.lvalToString lv)) ;
  match lv with 
    Expression.Symbol _ -> 
      let b = Hashtbl.find lvalue_to_bdd_map lv in
      b
  |	Expression.Dereference e -> 
     begin
      if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("Dereference" );
      let bdd_e = match e with
        Expression.Lval (Expression.Symbol s) -> Hashtbl.find lvalue_to_bdd_map (Expression.Symbol s)
      | _ -> get_bdd_e_expensive bdd_loc e
      in
      let lv_pts_to = CaddieBdd.bddCofactor bdd_loc (bdd_e)
      in
      if (!debug_bdd_pts_to) then 
	begin
	  Message.msg_string  Message.Normal ("Points to set:" );
	  CaddieBdd.bddPrint lv_pts_to 
	end;
      let b = CaddieBdd.replace lv_pts_to (CaddieBdd.toMap bddManager !stvarMap) in
      if (!debug_bdd_pts_to) then 
	begin
	  Message.msg_string  Message.Normal ("In terms of original vars:" );
	  CaddieBdd.bddPrint b 
	end;
      b
    end
  |	Expression.Access(Expression.Dot, Expression.Lval (Expression.Symbol s), _) ->
      Hashtbl.find lvalue_to_bdd_map (Expression.Symbol s)
  |	Expression.Access(Expression.Dot, Expression.Lval (Expression.Dereference (Expression.Lval (Expression.Symbol s))), _) ->
      let b = Hashtbl.find lvalue_to_bdd_map (Expression.Symbol s) in
      let lv_pts_to = CaddieBdd.bddCofactor bdd_loc b in
      let b' = CaddieBdd.replace lv_pts_to (CaddieBdd.toMap bddManager !stvarMap) in
      b'
  |	Expression.Access(Expression.Dot, e, _) ->
      get_bdd_e_expensive bdd_loc e
  |	Expression.Indexed (e1, _) ->
      get_bdd_e_expensive bdd_loc e1
	

(** This function Or the bdd function at source to the bdd function at target *)
let inherit_st target new_bdd_at_target worklist =
  try
    (* if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal "Inheriting BDDs" ; *)
    let bdd_at_target = lookup_location target in
    if (CaddieBdd.bddEqual (CaddieBdd.bddImp new_bdd_at_target bdd_at_target) bddOne) then
      ()
    else
      begin
	(* if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal "BDD Changed!" ; *)
	let target_bdd = CaddieBdd.bddOr bdd_at_target new_bdd_at_target in
	Hashtbl.remove location_to_bdd_map (C_SD.location_coords target) ;
	Hashtbl.replace location_to_bdd_map (C_SD.location_coords target) target_bdd ;
	worklist#add_element target;
	()
      end
  with Not_found ->
		(* something is wrong *)
    failwith ( "location not in hash table ")


(** This function is for p= (arbitrary) lval statement. Note bdd_target will possibly be changed in this function *)
let anyp_anyq monotonic bdd_source p q =
    try
      let bdd_pcode = get_bdd bdd_source p in 
      let bdd_qcode = get_bdd bdd_source q in
		(* get rid of p->y *)
      let elim_py = if monotonic then bdd_source else
                    CaddieBdd.bddAnd bdd_source (CaddieBdd.bddNot bdd_pcode) in
		(* get x *)
      let qx = CaddieBdd.bddCofactor elim_py bdd_qcode in
		(* get p->x *)
      let px = CaddieBdd.bddAnd bdd_pcode qx in
      let new_result = CaddieBdd.bddOr elim_py px in
      new_result
    with Not_found ->
      failwith ( " sth not found in any_p=any_q " ^ (Expression.lvalToString p) ^ (Expression.lvalToString q))
    | e -> begin print_string (Printexc.to_string e) ; failwith "p=q aha" end

	
(** This function is for p = &q statement. Note bdd_target will possibly be changed in this function *)
let anyp_addr_anyq monotonic bdd_source p q =
	(* p=&q. Get rid of any p->y. p->q *)
    try 
      let bdd_pcode = get_bdd bdd_source p in
      let bdd_qcode = get_bdd bdd_source q in
		(* get rid of p->y *)
      let elim_py = if monotonic then bdd_source else CaddieBdd.bddAnd bdd_source (CaddieBdd.bddNot bdd_pcode) in
		(* map q to the target set of variables *)
      let qtarget = CaddieBdd.replace bdd_qcode (CaddieBdd.toMap bddManager !stvarMap) in
      let pq = CaddieBdd.bddAnd bdd_pcode qtarget in
      let new_result = CaddieBdd.bddOr elim_py pq in
      new_result
    with Not_found ->
      failwith ( " sth not found in p=&q") 
    | e -> begin print_string (Printexc.to_string e) ; failwith "p = &q aha" end

(** This function is for p = &q statement. Note bdd_target will possibly be changed in this function *)
let p_addrq monotonic bdd_source p q =
	(* p=&q. Get rid of any p->y. p->q *)
    try 
      let bdd_pcode = Hashtbl.find lvalue_to_bdd_map p in
      let bdd_qcode = Hashtbl.find lvalue_to_bdd_map q in
		(* get rid of p->y *)
      let elim_py = if monotonic then bdd_source else CaddieBdd.bddAnd bdd_source (CaddieBdd.bddNot bdd_pcode) in
		(* map q to the target set of variables *)
      let qtarget = CaddieBdd.replace bdd_qcode (CaddieBdd.toMap bddManager !stvarMap) in
      let pq = CaddieBdd.bddAnd bdd_pcode qtarget in
      let new_result = CaddieBdd.bddOr elim_py pq in
      new_result
    with Not_found ->
      failwith ( " sth not found in p=&q") 
    | e -> begin print_string (Printexc.to_string e) ; failwith "p = &q aha" end
	
	
(** This function is for *p = *q statement. Note bdd_target will possibly be changed in this function *)
let derefp_derefq monotonic bdd_source p q =
	(* *p=*q. if q->x, x->r, p->y, then check if y is unique. if so, get rid of y->z. add y->r *)
    try 
      let bdd_pcode = Hashtbl.find lvalue_to_bdd_map p in
      let bdd_qcode = Hashtbl.find lvalue_to_bdd_map q in 
		(* get x *)
      let x = CaddieBdd.bddCofactor bdd_source bdd_qcode in
		(* get r, we need to first map x to the source set of variables *)
      let xsource = CaddieBdd.replace x (CaddieBdd.toMap bddManager !tsvarMap) in
      let r = CaddieBdd.bddCofactor bdd_source xsource in
		(* get y *)
      let bdd_py = CaddieBdd.bddAnd bdd_source bdd_pcode in 
      let y = CaddieBdd.bddCofactor bdd_py bdd_pcode in
      let ysource = CaddieBdd.replace y (CaddieBdd.toMap bddManager !tsvarMap) in
		(* compute y->r *)
      let yr = CaddieBdd.bddAnd ysource r in
      let py_num = if monotonic then 0 else int_of_float(CaddieBdd.bddCountMinterm bddManager bdd_py (2*(!num_bddVar))) in
      if (py_num = 1) then
	begin      
		(* get rid of y->z *)
	  let elim_yz = CaddieBdd.bddAnd bdd_source (CaddieBdd.bddNot ysource) in
	  let new_result = CaddieBdd.bddOr elim_yz yr in
	  new_result
	end
      else
	begin
	  let new_result = CaddieBdd.bddOr bdd_source yr in
	  new_result
	end
    with Not_found ->
      failwith ( " sth not found in *p=*q") 
    | e -> begin print_string (Printexc.to_string e) ; failwith "*p = *q aha" end

(** This function is for *p = &q statement. Note bdd_target will possibly be changed in this function *)
let derefp_addrq monotonic bdd_source p q = 
	(* *p=&q. if p->y, then check if y is unique. if so, get rid of y->z. add y->q *)
    try 
      if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("Now performing assignment *"^(Expression.lvalToString p)^" = &"^(Expression.lvalToString q)) ;
      let bdd_pcode = try Hashtbl.find lvalue_to_bdd_map p with Not_found -> print_string (Expression.lvalToString p) ; raise Not_found in
      let bdd_qcode = try Hashtbl.find lvalue_to_bdd_map q with Not_found -> print_string (Expression.lvalToString q) ; raise Not_found in 
      let bdd_py = CaddieBdd.bddAnd bdd_source bdd_pcode in 
		(* get y *)
      let y = CaddieBdd.bddCofactor bdd_py bdd_pcode in
      let ysource = CaddieBdd.replace y (CaddieBdd.toMap bddManager !tsvarMap) in
		(* map q to target set of variables *)
      let qtarget = CaddieBdd.replace bdd_qcode (CaddieBdd.toMap bddManager !stvarMap) in
		(* compute y->q *)
      let yq = CaddieBdd.bddAnd ysource qtarget in
      let py_num = if monotonic then 0 else int_of_float(CaddieBdd.bddCountMinterm bddManager bdd_py (2*(!num_bddVar))) in
      if (py_num = 1) then
	begin
		(* get rid of y->z *)
	  let elim_yz = CaddieBdd.bddAnd bdd_source (CaddieBdd.bddNot ysource) in
	  let new_result = CaddieBdd.bddOr elim_yz yq in
	  new_result
	end
      else
	begin 
	  let new_result = CaddieBdd.bddOr bdd_source yq in
	  new_result
	end
    with Not_found ->
	  failwith ( " sth not found in *p=&q") 
    | e -> begin print_string (Printexc.to_string e) ; failwith "*p = &q aha" end
	    
(** This function is for p (or *p) = alloc(...). Note bdd_target will possibly be changed in this function *)
let assign_alloc monotonic bdd_source ptarget qalloc = 

    Message.msg_string Message.Debug "In assign_alloc " ;
    Message.msg_string Message.Debug ("ptarget is "^(Expression.lvalToString ptarget)) ;
    
    try 
      let bdd_pcode = get_bdd bdd_source ptarget (* Hashtbl.find lvalue_to_bdd_map (ptarget) *) in
      let bdd_qcode = Hashtbl.find lvalue_to_bdd_map qalloc in
      match (simplify_lval ptarget) with
      Expression.Symbol p  -> 	
      (* p = alloc(..), then get rid of p->x. add p->qalloc *)
	begin
	  if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal "p = alloc ()" ; 
		(* get rid of p->x *)
	  let elim_px =  if monotonic then bdd_source else CaddieBdd.bddAnd bdd_source (CaddieBdd.bddNot bdd_pcode) in 
		(* map qalloc to target variables *)
	  let qtarget =  
	    CaddieBdd.replace 
	      bdd_qcode 
	      (CaddieBdd.toMap bddManager !stvarMap) 
	  
	  in
		(* compute p->qalloc *)
	  let pq = CaddieBdd.bddAnd bdd_pcode qtarget in
	  let new_result = CaddieBdd.bddOr elim_px pq in
	  new_result
	end
      | Expression.Dereference (Expression.Lval (Expression.Symbol p)) -> 
      (* *p=alloc(...), then if p->x, then check if x is unique. if so, get rid of x->r. add x->qalloc *)
	begin
	  if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal "*p = alloc ()" ; 
          CaddieBdd.bddPrint bdd_pcode;
		(* map qalloc to target variables *)
	  let qtarget = CaddieBdd.replace bdd_qcode (CaddieBdd.toMap bddManager !stvarMap) in
          if (!debug_bdd_pts_to) then CaddieBdd.bddPrint qtarget;
(*
		(* get x *)
	  let bdd_px = CaddieBdd.bddAnd bdd_source bdd_pcode in
          CaddieBdd.bddPrint bdd_px;
	  let x = CaddieBdd.bddCofactor bdd_px bdd_pcode in
          CaddieBdd.bddPrint x;
	  let xsource = CaddieBdd.replace x (CaddieBdd.toMap bddManager !tsvarMap) in
          CaddieBdd.bddPrint xsource;
		(* compute x->qalloc *)
	  let xq = CaddieBdd.bddAnd xsource qtarget in
*)
          let xq = CaddieBdd.bddAnd bdd_pcode qtarget in
           if (!debug_bdd_pts_to) then CaddieBdd.bddPrint xq;
	  (* let px_num = if monotonic then 0 else int_of_float(CaddieBdd.bddCountMinterm bddManager bdd_px (2*(!num_bddVar))) in *)
          (*
	  if (px_num = 1) then
	    begin      
		(* get rid of x->r *)
	      let elim_xr = CaddieBdd.bddAnd bdd_source (CaddieBdd.bddNot xsource) in
	      let new_result = CaddieBdd.bddOr elim_xr xq in
	      new_result
	    end
	  else
          *)
	    begin
	      let new_result = CaddieBdd.bddOr bdd_source xq in
	      new_result
	    end
	end
      | _ -> Message.msg_string Message.Normal ("Do not handle this yet. To do." ^ (Expression.lvalToString ptarget))  ; failwith "alloc"
    with Not_found ->
      Message.msg_string Message.Normal ( (" sth not found in alloc "^ (Expression.lvalToString ptarget) ^ (Expression.lvalToString qalloc))  ) ; failwith "alloc'"
    (* | e -> begin print_string (Printexc.to_string e) ; failwith "alloc " end *)
	

let rec do_assignment monotonic bdd ptarget qvalue =
    (* Statement: ptarget = qvalue *)
  
			(* we have  types of assignments: p=q(or q[i]); p=*q; p=&q; *p=q(or q[i]); *p=*q; *p=&x; *)
			(* first match lhs *)
  let p_lv = simplify_lval ptarget in
  if !debug_bdd_pts_to then begin
  	Message.msg_string Message.Normal ("ptarget is "^(Expression.lvalToString ptarget)) ;
  	Message.msg_string Message.Normal ("p_lv is "^(Expression.lvalToString p_lv)) ;
  end ;

  match p_lv with
      (* ptarget *) Expression.Symbol p  ->
	begin
	  match qvalue with
	  |  (* qvalue *) Expression.Lval lv 
	  |  (* qvalue *) (Expression.Cast (_, Expression.Lval lv)) ->
	      begin
		let q_lv = simplify_lval lv in
                    anyp_anyq monotonic bdd p_lv q_lv
	      end
	  | (* qvalue *) Expression.Unary (Expression.Address, e) 
	  | (* qvalue *) Expression.Cast (_, Expression.Unary (Expression.Address, e)) ->
	      begin
		match e with 
		  Expression.Lval lv -> (* case p = &q *)
		    let s_lv = simplify_lval lv in
		    begin
		      match s_lv with
			Expression.Symbol _ ->
				Message.msg_string Message.Normal ("p = & q with q : " ^ (Expression.lvalToString s_lv));
		          p_addrq monotonic bdd p_lv s_lv 
		      | Expression.Dereference e -> (* case p = & ( *e) I have simplified ptarget once, no need to do it again*)
			  do_assignment monotonic bdd p_lv e
                      | _ -> failwith "do_assignment: address_of : Strange expression"
		    end
		| _ -> failwith ("do_assignment: Not handled q = "^(Expression.toString e))
	      end
	| (* qvalue *) Expression.Binary (Expression.FieldOffset, e1, e2) -> (* assume no pointer arithmetic *)
	    begin
	      do_assignment monotonic bdd p_lv e1
	    end
	  | (* qvalue *) Expression.Binary (op, e1, e2) (* assume no pointer arithmetic *)
	  | (* qvalue *) Expression.Cast(_, Expression.Binary (op, e1, e2)) -> 
	      begin
	      if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("qvalue Binary "^(Expression.toString qvalue)) ;
		let simp_e = collapse_expression qvalue in
		if simp_e = qvalue then bdd else 
		begin
		  if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("Doing assignment on "^(Expression.toString simp_e)) ;
		  do_assignment monotonic bdd p_lv simp_e
		end
	      end
	  | (* qvalue *) Expression.Constant _ -> (* do nothing *)
	      bdd
	  | (* qvalue *) Expression.Sizeof _ -> (* do nothing *)
	      bdd
	  | (* qvalue *) Expression.Unary _ -> bdd
	  | _ -> failwith ("do_assignment : do not handle" ^ (Expression.toString qvalue))
	end
  | (* ptarget *) Expression.Dereference (Expression.Lval (Expression.Symbol p)) -> (* LHS is *p *)
      begin
	match qvalue with
	|   Expression.Lval lv 
	|   (Expression.Cast(_, Expression.Lval lv)) ->
	    begin
	      let q_lv = simplify_lval lv in
              anyp_anyq monotonic bdd p_lv q_lv
(*
	      match q_lv with
		Expression.Symbol s -> (* case *p = q *)
		  derefp_q monotonic bdd (Expression.Symbol p) q_lv
	      | Expression.Dereference (Expression.Lval (Expression.Symbol qs)) -> (* case *p = *q *)
		  derefp_derefq monotonic bdd (Expression.Symbol p) (Expression.Symbol qs)
	      | _ -> failwith ("do_assignment: Not handled q = "^(Expression.lvalToString lv))
*)
	    end
	| Expression.Unary (Expression.Address, e) 
	| Expression.Cast(_, Expression.Unary (Expression.Address, e)) ->
	    begin
	      match e with 
		Expression.Lval lv -> (* case p = &q *)
		  let s_lv = simplify_lval lv in
		  begin
		    match s_lv with
		      Expression.Symbol _ ->
		        derefp_addrq monotonic bdd (Expression.Symbol p) s_lv 
		    | Expression.Dereference e -> (* case *p = & ( *e) I have simplified ptarget once, no need to do it again*)
			do_assignment monotonic bdd p_lv e
                    | _ -> failwith "do_assignment: address_of : Strange expression"
		  end
	      | _ -> failwith ("do_assignment: Not handled q = "^(Expression.toString e))
	    end
	| (* qvalue *) Expression.Binary (Expression.FieldOffset, e1, e2) -> (* assume no pointer arithmetic *)
	    begin
	      do_assignment monotonic bdd p_lv e1
	    end
	| (* qvalue *) Expression.Binary (op, e1, e2) -> (* assume no pointer arithmetic *)
	    begin
	      if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("qvalue Binary "^(Expression.toString qvalue)) ;
	      let simp_e = collapse_expression qvalue in
	      if simp_e = qvalue then bdd else 
	      begin
		if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("Doing assignment on "^(Expression.toString simp_e)) ;
		do_assignment monotonic bdd p_lv simp_e
	      end
	    end
	| (* qvalue *) Expression.Constant _ -> (* do nothing *)
	    bdd
	| (* qvalue *) Expression.Sizeof _ -> (* do nothing *)
	    bdd
	| (* qvalue *) Expression.Unary _ -> bdd 
	| _ -> failwith ("do_assignment : do not handle"^(Expression.toString qvalue))
      end
  | (* ptarget *) _ -> 
	begin
	match qvalue with
	|   Expression.Lval lv 
	|   (Expression.Cast(_, Expression.Lval lv)) ->
	    begin
	      let q_lv = simplify_lval lv in
              anyp_anyq monotonic bdd p_lv q_lv
	    end
	| Expression.Unary (Expression.Address, e) 
	| Expression.Cast(_, Expression.Unary (Expression.Address, e)) ->
	    begin
	      match e with 
		Expression.Lval lv -> (* case p = &q *)
			anyp_addr_anyq monotonic bdd p_lv lv
(*
		  let s_lv = simplify_lval lv in
		  begin
		    match s_lv with
		      Expression.Symbol _ ->
		        derefp_addrq monotonic bdd (Expression.Symbol p) s_lv 
		    | Expression.Dereference e -> (* case *p = & ( *e) I have simplified ptarget once, no need to do it again*)
			do_assignment monotonic bdd p_lv e
                    | _ -> failwith "do_assignment: address_of : Strange expression"
		  end
*)
	      | _ -> failwith ("do_assignment: Not handled q = "^(Expression.toString e))
	    end
	| (* qvalue *) Expression.Binary (Expression.FieldOffset, e1, e2) -> (* assume no pointer arithmetic *)
	    begin
	      do_assignment monotonic bdd p_lv e1
	    end
	| (* qvalue *) Expression.Binary (op, e1, e2) -> (* assume no pointer arithmetic *)
	    begin
	      if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("qvalue Binary "^(Expression.toString qvalue)) ;
	      let simp_e = collapse_expression qvalue in
	      if simp_e = qvalue then bdd else 
	      begin
		if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("Doing assignment on "^(Expression.toString simp_e)) ;
		do_assignment monotonic bdd p_lv simp_e
	      end
	    end

	| (* qvalue *) Expression.Constant _ -> (* do nothing *)
	    bdd
	| (* qvalue *) Expression.Sizeof _ -> (* do nothing *)
	    bdd
	| (* qvalue *) Expression.Unary _ -> bdd 
	| _ -> failwith ("do_assignment : do not handle"^(Expression.toString qvalue))
	end 
	

	  
(* The main routine *)
let bdd_pts_to_flow_sensitive () =
  (* IMPROVEMENT: this is wasteful. i go through edges for lval_list and then go through them again for worklist. *)
  let lval_list = let (a, b, c) = get_all_lvals () in a  @ b @ c in
  Message.msg_string Message.Normal "Got all Lvals" ;
  (* how many bdd variables we need to encode all lvalues *)
  let num_bddvars_lval lval_list = int_of_float (ceil ((log (float_of_int (List.length lval_list))) /. (log 2.)))
  in
  let num_bddvars = num_bddvars_lval lval_list in
  Message.msg_string Message.Normal ("Number of BDD variables = "^(string_of_int num_bddvars)) ;
  encode_variables num_bddvars lval_list ;
  num_bddVar := num_bddvars;

  let uninterest_lval_fcn fcn_fnname lvalue_list =
    let rec bdd_fcn lval_bdd lval_list =
      match lval_list with
	[] -> lval_bdd
      | single_lval::rest ->
	  begin
	    try
	      let bdd_single = Hashtbl.find lvalue_to_bdd_map single_lval in
	      bdd_fcn (CaddieBdd.bddOr lval_bdd bdd_single) rest
	    with Not_found -> failwith "variable doesn't exist"
	  end
    in
    let bdd_lval_fcn = CaddieBdd.bddNot(bdd_fcn bddZero lvalue_list) in
    Hashtbl.add fcn_to_lvalbdd_map fcn_fnname bdd_lval_fcn
  in
  Hashtbl.iter uninterest_lval_fcn fcn_to_lvalue_map;
    
  (*
  let opfile = open_out "__bddtest_pst" in
  let printlval lval dummy_bdd = 
    Printf.fprintf opfile "%s\n" (Expression.lvalToString lval)
  in Hashtbl.iter printlval lvalue_to_bdd_map;
  close_out opfile;
    *)
  Message.msg_string Message.Normal (Printf.sprintf "Got all encodes %d" (List.length lval_list));

  let worklist = new Search.bfsiterator in
  (* we initialize all bdd functions to be false *)
  init_bdd_location worklist;

  Message.msg_string Message.Normal "bdd_pts_to_flow_sensitive: Starting iteration.\n";



  (* Process one edge *)
  let process_edge cur_loc bdd_at_cur_loc edge  = 
    let source = C_SD.get_source edge and
	target = C_SD.get_target edge in
    assert (cur_loc = source) ;

    (* debug
    Printf.fprintf debug_of "source %s target %s command %s \n" 
      (C_SD.location_to_string source) (C_SD.location_to_string target) 
      (C_SD.Command.to_string (C_SD.get_command edge)) ;
    *)
    let bdd_at_target = lookup_location target in
    let cmd = C_SD.get_command edge in
    match cmd.C_SD.Command.code with
    | C_SD.Command.Block l ->
	(* this is the interesting case -- *)
	begin
             (* analyze_statement takes a points to bdd and a statement (assignment or return)
		and updates the points to set for the effect of the statement 
		*)
	  let analyze_statement bdd stmt = 
	    match stmt with
	      C_SD.Command.Expr e ->
		begin
		  match e with
		    Expression.Assignment(Expression.Assign, ptarget, qvalue) ->
		      begin
			do_assignment false bdd ptarget qvalue
		      end
		  | _ -> failwith "Analyze_statement: this case not handled"
		end
	    | C_SD.Command.Return _ -> 
		(* we need to propagate the bdd functions to all possible target locations *)
		let fcn_fnname = C_SD.get_location_fname source in
		(* optimize1: we only return the bdd on interested variables *)
		try
		  let bddnot_lval_fcn = Hashtbl.find fcn_to_lvalbdd_map fcn_fnname in
		  let bdd_interest = CaddieBdd.bddAnd bdd bddnot_lval_fcn in
		  try 
		    let target_list = Hashtbl.find_all fcn_to_target_map fcn_fnname in
		    List.iter (function target -> inherit_st target bdd_interest worklist) target_list;
		    (* question: why do we have this bdd return here? *)
		    bdd
		  with Not_found -> failwith "function name unknown"
		with Not_found -> failwith "function name not found"
	  in
	  let new_bdd_at_target = List.fold_left (analyze_statement) bdd_at_cur_loc l in
	  inherit_st target new_bdd_at_target worklist
	end
    | C_SD.Command.FunctionCall fexp ->
	begin
	  match fexp with
	    Expression.FunctionCall (Expression.Lval (Expression.Symbol fname), arglist) 
	  | Expression.Cast (_, Expression.FunctionCall(Expression.Lval (Expression.Symbol fname), arglist)) ->
	      begin
              try
		match (C_SD.lookup_formals fname) with
		  C_SD.Fixed formals ->
		    begin
		      let nameAssocList = 
			try
			  List.combine 
			    (List.map (function s -> Expression.Symbol s) formals)
			    arglist
			with _ -> failwith "Incorrect number of parameters in call"
		      in
		      let entry_loc = C_SD.lookup_entry_location fname in
		      let bdd_at_entry_loc = lookup_location entry_loc
		      in
		      let increase_bdd_at_entry_loc = 
			List.fold_left (function bdd -> function (lv, e) -> do_assignment false bdd lv e) 
			  bdd_at_cur_loc nameAssocList 
		      in
		      let new_bdd_at_entry_loc = CaddieBdd.bddOr increase_bdd_at_entry_loc bdd_at_entry_loc in
		      inherit_st entry_loc new_bdd_at_entry_loc worklist
		    end
		| _ -> (* failwith "Do not handle variable parameters" *)
		inherit_st target bdd_at_cur_loc worklist
		
		(* the following exception deals with functions that are not found *)
              with _ -> inherit_st target bdd_at_cur_loc worklist
	      end
	  | Expression.Assignment (Expression.Assign, retval, 
				   Expression.FunctionCall(Expression.Lval (Expression.Symbol fname), arglist)) 
	  | Expression.Assignment (Expression.Assign, 
				   retval, 
				   (Expression.Cast (_, 
						     Expression.FunctionCall(Expression.Lval (Expression.Symbol fname), 
									     arglist)))) ->
	     
	  begin
	    try
	     if (List.mem fname allocFuns) then
	       (* alloc functions are handled specially *)
	       begin
		   let qalloc = Hashtbl.find edge_to_mem_map edge in
		   let new_bdd_at_target = assign_alloc false bdd_at_cur_loc retval qalloc in
		   inherit_st target new_bdd_at_target worklist
	       end
	     else
	       begin
		 match (C_SD.lookup_formals fname) with
		   C_SD.Fixed formals ->
		     begin
		       let nameAssocList = 
			 try
			   List.combine 
			     (List.map (function s -> Expression.Symbol s) formals)
			     arglist
			 with _ -> failwith "Incorrect number of parameters in call"
		       in
		       let entry_loc = C_SD.lookup_entry_location fname in
		       let bdd_at_entry_loc = lookup_location entry_loc
		       in
		       let increase_bdd_at_entry_loc = 
			 List.fold_left (function bdd -> function (lv, e) -> do_assignment false bdd lv e) 
			   bdd_at_cur_loc nameAssocList 
		       in		
		       let new_bdd_at_entry_loc = CaddieBdd.bddOr increase_bdd_at_entry_loc bdd_at_entry_loc in 
		       inherit_st entry_loc new_bdd_at_entry_loc worklist;
		       (* get the assignment for retval *)
		       let returnlvalue = Expression.Lval(Expression.Symbol("__retres@"^fname)) in
		       let new_bdd_at_target = do_assignment false bdd_at_cur_loc retval returnlvalue in
		       inherit_st target new_bdd_at_target worklist
		     end
		 | _ -> failwith "Do not handle variable parameters"
	       end
		
		(* the following exception deals with functions that are not found *)
              with _ -> inherit_st target bdd_at_cur_loc worklist
	  end
	  | _ -> inherit_st target bdd_at_cur_loc worklist (* failwith ("Strange function call : "^(Expression.toString fexp)) *)					       
	end
    | _ -> (* other things are uninteresting: or the current bdd with the target *)
	inherit_st target bdd_at_cur_loc worklist
  in
  let i = ref 0 in
  while (worklist#has_next () ) do
    incr i ;
      (* do the points to analysis *)
    let cur_loc = worklist#next () in
      (* process location cur_loc  *)
    let out_edges = C_SD.get_outgoing_edges cur_loc in
    let bdd_at_cur_loc = lookup_location cur_loc in
      (* for each outgoing edge, process the target of the edge according to the bdd of this node *)
    List.map (process_edge cur_loc bdd_at_cur_loc) out_edges 
  done;
  Message.msg_string Message.Normal (Printf.sprintf "Iteration %d\n" !i) 
(* at this point, the map location_to_bdd_map has the points to information *)

(****************************************************************************)
(****************************************************************************)
(****************************************************************************)
(** A flow insensitive version of points to analysis *)
(****************************************************************************)
(****************************************************************************)

let bdd_pts_to_flow_insensitive () =
  let (global_list, local_list, fn_list) = get_all_lvals () in 
  let lval_list = global_list @ local_list @ fn_list in
(*
  Message.msg_string Message.Normal "Got all Lvals" ;
  List.iter (fun a -> print_string ("global " ^(Expression.lvalToString a)) ) global_list ;
  List.iter (fun a -> print_string ("local " ^(Expression.lvalToString a)) ) local_list ;
  print_string "funs\n" ;
  List.iter (fun a -> print_string (Expression.lvalToString a)) fn_list ;
*)
  (* how many bdd variables we need to encode all lvalues *)
  let num_bddvars_lval lval_list = int_of_float (ceil ((log (float_of_int (List.length lval_list))) /. (log 2.)))
  in
  let num_bddvars = num_bddvars_lval lval_list in
  Message.msg_string Message.Normal ("Number of BDD variables = "^(string_of_int num_bddvars)) ;
  encode_variables num_bddvars lval_list ;
    Hashtbl.iter (fun key b -> Message.msg_string Message.Normal (Expression.lvalToString key) ;    if (!debug_bdd_pts_to) then (CaddieBdd.bddPrint b)) lvalue_to_bdd_map ;
  num_bddVar := num_bddvars;
  let initialize_arrays lvals =
	(* for each array variable a, set the mapping a -> a#aloc *)
     let rec __loop l bdd = match l with [] -> bdd
     | a :: rest -> 
        match a with Expression.Symbol s ->
          if (not (Misc.is_suffix "#aloc" s) && (not (is_malloc_location a)) && (C_SD.is_array a)) then 
          begin
                let bdd_pcode = Hashtbl.find lvalue_to_bdd_map a in
                let bdd_qcode = Hashtbl.find lvalue_to_bdd_map (Expression.Symbol (s^"#aloc") ) in
                let qtarget =  CaddieBdd.replace bdd_qcode (CaddieBdd.toMap bddManager !stvarMap)
                in
                (* compute p->qalloc *)
                let pq = CaddieBdd.bddAnd bdd_pcode qtarget in
                let new_result = CaddieBdd.bddOr bdd pq in
                __loop rest new_result
          end
          else __loop rest bdd
        | _ -> Message.msg_string Message.Normal ("Strange lvalue in initialize_array"^(Expression.lvalToString a)) ; failwith "ooh" 
     in
     __loop lvals bddZero
  in
  let initialize_fns b = 
    (* for each function, put a self loop *)
    List.fold_left (fun bdd thisfn -> 
      let bdd_pcode = Hashtbl.find lvalue_to_bdd_map thisfn in
      let bdd_target = CaddieBdd.replace bdd_pcode (CaddieBdd.toMap bddManager !stvarMap) in
      CaddieBdd.bddOr bdd (CaddieBdd.bddAnd bdd_pcode bdd_target)
      ) b fn_list 
  in
  let initialize_arrays_and_fns lval_list =
    let b1 = initialize_arrays lval_list in
    initialize_fns b1 
  in
  let initial_bdd = initialize_arrays_and_fns lval_list in
  let ptstobdd = ref initial_bdd in
  let newptstobdd = ref initial_bdd in
  Message.msg_string Message.Normal "bdd_pts_to: Starting iteration.\n";
  
  (* do alias analysis on all branches until there is no change *)
  let changed = ref true in (* Ocaml does not have a do... while construct,
                               so I set it up like this *)
  while !changed do
    ptstobdd := !newptstobdd ;

    (* Process one edge *)
    let process_one_edge bdd_at_cur_loc edge  = 
      let bdd_at_target = bdd_at_cur_loc in
      let cmd = C_SD.get_command edge in
      match cmd.C_SD.Command.code with
      | C_SD.Command.Block l ->
	(* this is the interesting case -- *)
	  begin
             (* analyze_statement takes a points to bdd and a statement (assignment or return)
		and updates the points to set for the effect of the statement 
		*)
	    let analyze_statement bdd stmt = 
	      match stmt with
		C_SD.Command.Expr e ->
		  begin
		    match e with
		      Expression.Assignment(Expression.Assign, ptarget, qvalue) ->
			begin
			  do_assignment true bdd ptarget qvalue
			end
		    | _ -> failwith "Analyze_statement: this case not handled"
		  end
	      | C_SD.Command.Return _ -> 
		  bdd
	    in
	    List.fold_left (analyze_statement) bdd_at_cur_loc l
	      
	  end
      | C_SD.Command.FunctionCall fexp ->
	  begin
(*
		Message.msg_string Message.Normal "In FunctionCall" ;
		Message.msg_string Message.Normal (Expression.toString fexp) ; 
*)		
	    match fexp with
	      Expression.FunctionCall (Expression.Lval (Expression.Symbol fname), arglist) 
	    | Expression.Cast (_, Expression.FunctionCall(Expression.Lval (Expression.Symbol fname), arglist)) ->
		begin
		  try
		    match (C_SD.lookup_formals fname) with
		      C_SD.Fixed formals ->
			begin
			  let nameAssocList = 
			    try
			      List.combine 
				(List.map (function s -> Expression.Symbol s) formals)
				arglist
			    with _ -> failwith "Incorrect number of parameters in call"
			  in
			  let bdd_at_entry_loc = bdd_at_cur_loc
			  in
			  let increase_bdd_at_entry_loc = 
			    List.fold_left (function bdd -> function (lv, e) -> do_assignment true bdd lv e) 
			      bdd_at_cur_loc nameAssocList 
			  in
			  let new_bdd_at_entry_loc = CaddieBdd.bddOr increase_bdd_at_entry_loc bdd_at_entry_loc in
			  new_bdd_at_entry_loc
			end
		    | _ -> (* failwith "Do not handle variable parameters" *)
			bdd_at_cur_loc 
			  
		(* the following exception deals with functions that are not found *)
		  with _ -> bdd_at_cur_loc 
		end
	    | Expression.Assignment (Expression.Assign, retval, 
				     Expression.FunctionCall(Expression.Lval (Expression.Symbol fname), arglist)) 
	    | Expression.Assignment (Expression.Assign, 
				     retval, 
				   (Expression.Cast (_, Expression.FunctionCall(Expression.Lval (Expression.Symbol fname), arglist)))) ->
				     
				     begin
				       try
					 if (List.mem fname allocFuns) then
	                                   (* alloc functions are handled specially *)
					   begin
					     let qalloc = Hashtbl.find edge_to_mem_map edge in
					     let new_bdd_at_target = assign_alloc true bdd_at_cur_loc retval qalloc in
(*
		Message.msg_string Message.Normal "After FunctionCall, BDD is" ;
		CaddieBdd.bddPrint new_bdd_at_target ;
*)
						
					     new_bdd_at_target 
					   end
					 else
					   begin
					     match (C_SD.lookup_formals fname) with
					       C_SD.Fixed formals ->
						 begin
						   let nameAssocList = 
						     try
						       List.combine 
							 (List.map (function s -> Expression.Symbol s) formals)
							 arglist
						     with _ -> failwith "Incorrect number of parameters in call"
						   in
						   let bdd_at_entry_loc = bdd_at_cur_loc
						   in
						   let increase_bdd_at_entry_loc = 
						     List.fold_left (function bdd -> function (lv, e) -> do_assignment true bdd lv e) 
						       bdd_at_cur_loc nameAssocList 
						   in		
						   let new_bdd_at_entry_loc = CaddieBdd.bddOr increase_bdd_at_entry_loc bdd_at_entry_loc in 

						   
	                                            	       (* get the assignment for retval *)
 						   let returnlvalue = Expression.Lval(Expression.Symbol("__retres@"^fname)) in
						   let new_bdd_at_target = do_assignment true new_bdd_at_entry_loc retval returnlvalue in
						   new_bdd_at_target
						 end
					     | _ -> failwith "Do not handle variable parameters"
					   end
					     
				       with C_SD.NoSuchFunctionException f -> if !debug_bdd_pts_to then Message.msg_string Message.Normal ("Warning: Function "^f^" not found") ; bdd_at_cur_loc 
					| _ -> bdd_at_cur_loc (* THIS IS WRONG *)
				     end
	    | _ -> bdd_at_cur_loc  (* failwith ("Strange function call : "^(Expression.toString fexp)) *)					       
	  end
      | _ -> (* other things are uninteresting *)
	  bdd_at_cur_loc 
    in


    (* now do the points to analysis by trawling all edges and applying the
       appropriate rules
       *)
    newptstobdd := List.fold_left (fun b -> fun c -> CaddieBdd.bddOr b c) (!ptstobdd) (C_SD.map_edges (process_one_edge (!ptstobdd) )) ;
    
    changed := not (CaddieBdd.bddEqual (!ptstobdd) (!newptstobdd) ) 
  done ;
  (* at the end of the loop, ptstobdd contains (a reference to) the points to relation *)
  if (!debug_bdd_pts_to) then 
    begin
    CaddieBdd.bddPrint !ptstobdd ;
    Hashtbl.iter (fun key b -> Message.msg_string Message.Normal (Expression.lvalToString key) ; 
			CaddieBdd.bddPrint b) lvalue_to_bdd_map
   end ;

  
  !ptstobdd

(****************************************************************************************)
(* Lookup functions                                                                     *)
(* and query engine                                                                     *)
(*                                                                                      *)
(* Flow insensitive versions:
   checkAliasFI bdd lval1 lval2
   checkPointsToFI   bdd lval1 lval2
   getAllAliasesFI lval

   Flow sensitive versions:
   checkAliasFS loctbl loc lval1 lval2
   checkPointsToFS   loctbl loc lval1 lval2
   getAllAliasesFS lval
*)
(****************************************************************************************)
let rec get_pts_to bdd_loc (p : Expression.lval) =
  if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("In get_pts_to : p is "^(Expression.lvalToString p)) ;
  let addr_p =  (Expression.addressOf (Expression.Lval p)) in
  if (!debug_bdd_pts_to) then Message.msg_string  Message.Normal ("In get_pts_to : &p is "^(Expression.toString (addr_p))) ;
  match addr_p with
    Expression.Lval l -> 
      let bdd_pcode = get_bdd bdd_loc l in
		(* get y in p->y *)
      let y = CaddieBdd.bddCofactor bdd_loc bdd_pcode in
      y
  |	Expression.Unary (Expression.Address, ae) ->
      begin
	match ae with
	  Expression.Lval l ->
	    begin
	      let bdd_pcode = get_bdd bdd_loc l in
	      CaddieBdd.replace bdd_pcode (CaddieBdd.toMap bddManager !stvarMap)
	    end
	| _ -> failwith ("get_pts_to: Unary Address of strange expression "^(Expression.toString ae))
      end
  |	Expression.Binary (Expression.FieldOffset, e1, e2) ->
      begin
	match e1 with
	  Expression.Lval l1 ->
	    let bdd_l = get_bdd bdd_loc l1 in
	    CaddieBdd.bddCofactor bdd_loc bdd_l
	| Expression.Unary (Expression.Address, Expression.Lval l) ->
	    let bdd_pcode = get_bdd bdd_loc l in
	    CaddieBdd.replace bdd_pcode (CaddieBdd.toMap bddManager !stvarMap)
        | Expression.Binary (op, _ , _) when (op = Expression.Plus || op = Expression.Minus) ->
               begin
                 let l = collapse_expression e1 in 
                 match l with Expression.Lval ll ->
                   let bdd_pcode = get_bdd bdd_loc ll in
                   let y = CaddieBdd.bddCofactor bdd_loc bdd_pcode in
                   y
                 | _ -> Message.msg_string Message.Error ("Binary case inside foffset fails! "^(Expression.toString addr_p)) ; failwith "binary"
               end
              
	| _ -> get_pts_to bdd_loc (Expression.Dereference e1) 
      end
  | Expression.Binary (bop, e1, e2) when (bop = Expression.Plus || bop = Expression.Minus  ) ->
      begin
(*
  Message.msg_string Message.Normal "Case: binary" ;
  Message.msg_string Message.Normal ("addr_p : " ^ (Expression.toString addr_p)) ;
  *)
	let l = collapse_expression addr_p in
        match l with Expression.Lval ll ->
          let bdd_pcode = get_bdd bdd_loc ll in
          let y = CaddieBdd.bddCofactor bdd_loc bdd_pcode in
          y
        | Expression.Binary (op, _, _) when (op = Expression.Plus || op = Expression.Minus) ->
            begin
	      Message.msg_string Message.Normal ("Binary case fails with binop! " ^(Expression.toString l)) ; failwith "binary"
            end
        | _ -> Message.msg_string Message.Error ("Binary case fails! " ^(Expression.toString l)) ; failwith "binary"
	      
      end
  |	_ -> Message.msg_string Message.Normal ("get_pts_to : Strange address of expression: "^(Expression.lvalToString p)) ; failwith "bin"
	
	
(****************************************************************************************)

let checkAliasFI bdd_loc (p : Expression.lval) (q : Expression.lval) =
  try
    if (!debug_bdd_pts_to) then
      begin
	Message.msg_string Message.Normal ("checking Alias : "^(Expression.lvalToString p)^ " " ^(Expression.lvalToString q)) 
      end;
    let y = get_pts_to bdd_loc p and
		(* get x in q->x *)
	x = get_pts_to bdd_loc q 
    in
    if !debug_bdd_pts_to then 
      begin
	Message.msg_string  Message.Normal "x is " ;
	CaddieBdd.bddPrint x ;
	Message.msg_string  Message.Normal "y is " ;
	CaddieBdd.bddPrint y 
      end ;
    
		(* get intersection of x and y *)
    let xy = CaddieBdd.bddAnd y x in
    if (!debug_bdd_pts_to) then 
      begin
	Message.msg_string Message.Normal "x & y is " ;
	CaddieBdd.bddPrint xy ;
      end; 
    if (CaddieBdd.bddEqual xy (bddZero )) then
      false
    else
      true
  with
    excn -> Message.msg_string Message.Error ("Warning:Error raised in checkAliasFI!:"^(Printexc.to_string excn) ^ " for inputs " 
		^ (Expression.lvalToString p) ^ " and " ^ (Expression.lvalToString q) ) ; false
	
let checkPointsToFI bdd_loc (p : Expression.lval) (q : Expression.lval) = 
  begin
      try 
	let bdd_pcode = get_pts_to bdd_loc p in
	let bdd_qcode = Hashtbl.find lvalue_to_bdd_map q in
	let qtarget = CaddieBdd.replace bdd_qcode (CaddieBdd.toMap bddManager !stvarMap) in
    if !debug_bdd_pts_to then 
      begin
	Message.msg_string  Message.Normal "p is " ;
	CaddieBdd.bddPrint bdd_pcode ;
        Message.msg_string Message.Normal ("q: " ^ (Expression.lvalToString q));
	Message.msg_string  Message.Normal "q is " ;
	CaddieBdd.bddPrint bdd_qcode ; 
	Message.msg_string  Message.Normal "q (primed) is " ;
	CaddieBdd.bddPrint qtarget ; 
      end ;
        let xy = (CaddieBdd.bddAnd bdd_pcode qtarget) in 
    if (!debug_bdd_pts_to) then 
      begin
	Message.msg_string Message.Normal "x & y is " ;
	CaddieBdd.bddPrint xy ;
      end; 
	begin
	  if (CaddieBdd.bddEqual xy bddZero) then
	    false
	  else
	    true
	end
      with Not_found ->
	failwith ( "sth not found in checkPointsToFI")
    end

let getAllAliasesFI bdd_loc p =
  begin
    let bdd_pcode = Hashtbl.find lvalue_to_bdd_map p in
    let y = CaddieBdd.bddCofactor bdd_loc bdd_pcode in (* this is the set of lvals
							  that p points to 
							 *)
    let allaliases = CaddieBdd.bddCofactor bdd_loc y in
    (* now iterate over all lvalues. Is there a better way? *)
    [] (* TO DO *)
  end

(**********************************************************************)
(* Various querying and printing utilities                            *)
(**********************************************************************)

let alias_pointsto_flow_insensitive opfile1 opfile2 bdd_loc p q = 
  if (p = q) then
    ()
  else
    begin
      try 
	let bdd_pcode = Hashtbl.find lvalue_to_bdd_map p in
	let bdd_qcode = Hashtbl.find lvalue_to_bdd_map q in
	let qtarget = CaddieBdd.replace bdd_qcode (CaddieBdd.toMap bddManager !stvarMap) in
	let y = CaddieBdd.bddCofactor bdd_loc bdd_pcode in
	begin
	  if (CaddieBdd.bddEqual (CaddieBdd.bddAnd y qtarget) bddZero) then
	    ()
	  else
	    Printf.fprintf opfile1 "%s -> %s \n" 
	      (Expression.lvalToString p) 
	      (Expression.lvalToString q) 
	      ;
	  let x = CaddieBdd.bddCofactor bdd_loc bdd_qcode in
	  if (CaddieBdd.bddEqual (CaddieBdd.bddAnd x y) bddZero) then
	    ()
	  else
	    Printf.fprintf opfile2 "%s ==  %s ; \n" 
	      (Expression.lvalToString p) 
	      (Expression.lvalToString q) 
	      
	end
      with Not_found ->
	failwith ( "sth not found in print")
    end




(* is_alias : location -> Expression.Lval -> Expression.Lval -> bool *) 
let is_alias opfile loc p q =
  (* check if the set pointed to by p intersects the set pointed to by q at location loc*)
  try
    if (p = q) then true
    else
      let bdd_loc = lookup_location loc in
      let bdd_pcode = Hashtbl.find lvalue_to_bdd_map p in
      let bdd_qcode = Hashtbl.find lvalue_to_bdd_map q in 
		(* get y in p->y *)
      let y = CaddieBdd.bddCofactor bdd_loc bdd_pcode in
		(* get x in q->x *)
      let x = CaddieBdd.bddCofactor bdd_loc bdd_qcode in
		(* get intersection of x and y *)
      let xy = CaddieBdd.bddAnd y x in
      if (CaddieBdd.bddEqual xy (bddZero )) then
	false
      else
	begin
      (* print out p is alias of q *)
	  Printf.fprintf opfile "%s <-> %s %s\n" 
		(Expression.lvalToString (simplify_lval p)) 
		(Expression.lvalToString (simplify_lval q)) 
		(C_SD.location_to_string loc) ; 
	      true
	end
  with Not_found ->
    failwith ( " sth not found in is_alias") 

(* is_pointsto : location -> Expression.lval -> Expression.lval -> bool *)
let is_pointsto loc p q =
  if (p=q) then
    false
  else
    begin
  try 
    let bdd_loc = lookup_location loc in
    let bdd_pcode = Hashtbl.find lvalue_to_bdd_map p in
    let bdd_qcode = Hashtbl.find lvalue_to_bdd_map q in
               (* get q to target set of variables *)
    let qtarget = CaddieBdd.replace bdd_qcode (CaddieBdd.toMap bddManager !stvarMap) in
    let pq = CaddieBdd.bddAnd bdd_pcode qtarget in
    if (CaddieBdd.bddEqual (CaddieBdd.bddAnd bdd_loc pq) bddZero) then 
      false
    else
      (*
      begin
	Printf.fprintf opfile "%s -> %s at %s\n" 
	  (Expression.lvalToString p) 
	  (Expression.lvalToString q) 
	  (C_SD.location_to_string loc) ; 
	true
      end
	*)
      true
  with Not_found ->
    failwith ( " sth not found in is_pointsto")
    end

let alias_pointsto opfile1 opfile2 loc p q = 
  if (p = q) then
    ()
  else
    begin
      try 
	let bdd_loc = lookup_location loc in
	let bdd_pcode = Hashtbl.find lvalue_to_bdd_map p in
	let bdd_qcode = Hashtbl.find lvalue_to_bdd_map q in
	let qtarget = CaddieBdd.replace bdd_qcode (CaddieBdd.toMap bddManager !stvarMap) in
	let y = CaddieBdd.bddCofactor bdd_loc bdd_pcode in
	begin
	  if (CaddieBdd.bddEqual (CaddieBdd.bddAnd y qtarget) bddZero) then
	    ()
	  else
	    Printf.fprintf opfile1 "%s -> %s ; // at  %s\n" 
	      (Expression.lvalToString p) 
	      (Expression.lvalToString q) 
	      (C_SD.location_to_string loc) ;
	  let x = CaddieBdd.bddCofactor bdd_loc bdd_qcode in
	  if (CaddieBdd.bddEqual (CaddieBdd.bddAnd x y) bddZero) then
	    ()
	  else
	    Printf.fprintf opfile2 "%s == %s; // at  %s\n" 
	      (Expression.lvalToString p) 
	      (Expression.lvalToString q) 
	      (C_SD.location_to_string loc)
	end
      with Not_found ->
	failwith ( "sth not found in print")
    end


(* this function directly prints out all p->q at some location *)
let do_bdd_pointsto_flow_sensitive () = 

  Message.msg_string Message.Normal "Points to print start.\n" ;

  let printlist = new Search.bfsiterator in
  init_bdd_location printlist;

  Message.msg_string Message.Normal "get all locations\n" ;

  (* print out all the lvalues *)
  (*
  let printlval lval dummy_bdd = 
    Printf.fprintf opfile "%s\n" (Expression.lvalToString lval)
  in Hashtbl.iter printlval lvalue_to_bdd_map;
    *)
  (* print out all locations *)
  (*
  let printloc (a,b) dummy_bdd = 
    Printf.fprintf opfile "location (%d,%d) \n" a b 
  in
  Hashtbl.iter printloc location_to_bdd_map;
    *)
  let i = ref 0.0 in
  while (printlist#has_next () ) do
    let cur_loc = printlist#next () in
    let bdd_loc = lookup_location cur_loc in
    i := !i +. (CaddieBdd.bddCountMinterm bddManager bdd_loc (2*(!num_bddVar)))
  done;

  Message.msg_string Message.Normal (Printf.sprintf "number of edges %f\n" !i) 






(* End of various printing and querying utilities                                 *)
(**********************************************************************************)
(**********************************************************************************)
(* Main routines to call the analyses     *)
(**********************************************************************************)





type pointsToInfo =
    FlowSensitivePointsTo of ((int * int, CaddieBdd.bdd) Hashtbl.t)
  | FlowInsensitivePointsTo of CaddieBdd.bdd

(* Store the points to information *)
let store_FI_points_to_bdd bdd fname =
  begin
    CaddieBdd.bddStore bddManager bdd fname 
  end

(* Load a previously computed BDD from a file. *)
let load_FI_points_to_bdd fname =
  begin
    (* first set up the bdd manager and the var maps *)
  let lval_list = let (a,b, c) = get_all_lvals () in a @ b @ c in
  Message.msg_string Message.Normal "Got all Lvals" ;
  (* how many bdd variables we need to encode all lvalues *)
  let num_bddvars_lval lval_list = int_of_float (ceil ((log (float_of_int (List.length lval_list))) /. (log 2.)))
  in
  let num_bddvars = num_bddvars_lval lval_list in
  Message.msg_string Message.Normal ("Number of BDD variables = "^(string_of_int num_bddvars)) ;
  encode_variables num_bddvars lval_list ;
  num_bddVar := num_bddvars; 

  (* then invoke the bddLoad routine *)
  let bdd = try (CaddieBdd.bddLoad bddManager fname)
  with _ -> 
   begin
     Message.msg_string Message.Error "Loading BDD failed: Rebuilding BDD." ;
     let flow_insensitive_bdd =
	bdd_pts_to_flow_insensitive () 
     in 
     store_FI_points_to_bdd flow_insensitive_bdd ((Options.getValueOfString "mainsourcename")^".bdd") ;
     Message.msg_string Message.Normal (Printf.sprintf "Number of edges (flow insensitive) %f\n" 
					   (CaddieBdd.bddCountMinterm bddManager flow_insensitive_bdd (2*(!num_bddVar)))) ;
     flow_insensitive_bdd
   end
  in
    begin
	()
(*
    CaddieBdd.bddPrint bdd ;
    Hashtbl.iter (fun key b -> Message.msg_string Message.Normal (Expression.lvalToString key) ;
                        CaddieBdd.bddPrint b) lvalue_to_bdd_map
*)
   end ;

  FlowInsensitivePointsTo bdd 
  end


(* The main routine.
   flow_flag = true --> do a flow sensitive analysis.
             = false --> do a flow insensitive analysis.
*)
let do_bdd_alias flow_flag = 
  Message.msg_string Message.Normal "In do_bdd_alias\n" ;
(*
  let fname = (Options.getValueOfString "mainsourcename") in
  let aliasfname = fname^".alias" in
  let ptstofname = fname^".ptsto" in
*)


  (* call the analysis *)
  if (flow_flag) then
    begin
      bdd_pts_to_flow_sensitive () ;

      (* print out points to and alias information *)
      (*
      do_bdd_pointsto_flow_sensitive () 
      *)
      
      (* return a hashtable of loc -> bdds *)
      FlowSensitivePointsTo location_to_bdd_map

    end 
  else 
    begin
      let flow_insensitive_bdd =
	bdd_pts_to_flow_insensitive () 
      in 
      store_FI_points_to_bdd flow_insensitive_bdd ((Options.getValueOfString "mainsourcename")^".bdd") ;
      Message.msg_string Message.Normal (Printf.sprintf "Number of edges (flow insensitive) %f\n" 
					   (CaddieBdd.bddCountMinterm bddManager flow_insensitive_bdd (2*(!num_bddVar)))) ;
  
(* DEBUG 
      let aliasfname = "__aliases" and ptstofname = "__ptsto" in
      let opfile1 = open_out aliasfname and
	  opfile2 = open_out ptstofname in
      let enump lvalp p_bdd =
	let enumq lvalq q_bdd =
	  begin
	    ignore (  alias_pointsto_flow_insensitive opfile2 opfile1 flow_insensitive_bdd lvalp lvalq );
	  end
	in
	Hashtbl.iter enumq lvalue_to_bdd_map
      in
      Hashtbl.iter enump lvalue_to_bdd_map ;
  

      close_out opfile1 ;
      close_out opfile2 ;
      Message.msg_string Message.Normal "The Points to BDD is : " ;
      CaddieBdd.bddPrint flow_insensitive_bdd ;
 END DEBUG *)
    (* return the bdd *)
      FlowInsensitivePointsTo flow_insensitive_bdd

    end




let do_mod_analysis ptstobdd = 
	  (* construct the modified information for each function.
	     Basically build a BDD f(l1, l2), where the first
	     component is a function name, and the second component
	     is an lvalue whose transitive dereference may be modified by the function.
	     The second lvalue is assumed to be a formal parameter of the first,
	     or a global.
	     
	     This requires alias information.
	     *)
  (* we do a simple data flow analysis.
     Initially, the bdds are all false, and we populate the BDDs by going through
     the cfa until we get a fixpoint.

     At this point, the alias analysis has been run, and the lvalue to bdd maps have been set up.
     *)
  let (globallist, locallist, fnlist) = get_all_lvals () in											    
  let mod_bdd = ref bddZero in
  let is_modified lval func bdd =
    try
      let calleebdd = try Hashtbl.find lvalue_to_bdd_map (Expression.Symbol func) with Not_found -> failwith ("func "^func) in
      let lvaldd = try Hashtbl.find lvalue_to_bdd_map (Expression.fullsimp_lval lval) with Not_found -> 
	failwith ("func: "^func^" lval: "^(Expression.lvalToString lval)) in
      let primedlvaldd = (CaddieBdd.replace lvaldd (CaddieBdd.toMap bddManager !stvarMap)) in
      let conjunct = CaddieBdd.bddAnd calleebdd primedlvaldd in
      if (CaddieBdd.bddEqual (CaddieBdd.bddAnd bdd conjunct) bddZero) then
	false
      else
	true
    with _ -> false
  in

  let is_deref_of_global l fname =
    C_SD.is_global l (* or there exists a global to which this is aliased *) ||
    (List.exists (fun a -> checkAliasFI ptstobdd l a) globallist)
  in
  let is_deref_of_formal l fname =
    let formals = match C_SD.lookup_formals fname with C_SD.Fixed l | C_SD.Variable l -> l in
    let rec _formal_worker l fname = 
      match l with
	Expression.Symbol sy -> 
	  (List.mem sy formals) || (List.exists (fun s -> checkAliasFI ptstobdd l (Expression.Symbol s)) formals)
      | Expression.Dereference e 
      | Expression.Access(_, e, _) -> _formal_worker_e e fname
      | Expression.Indexed (e1, e2) -> _formal_worker_e e1 fname
    and _formal_worker_e e fname =
      match e with
	Expression.Lval l -> _formal_worker l fname
      |	_ -> failwith ("is_deref_of_formal : Strange expression " ^ (Expression.toString e))
    in
    match l with 
      Expression.Symbol _ -> false
    | _ -> _formal_worker l fname
  in

  (* Start the dataflow analysis *)
  Message.msg_string Message.Normal "Now doing the dataflow analysis " ;
  let changed = ref true in
  while !changed do
    changed := false ;
    let check_lval fname bdd lv = (* take an lval and a bdd and return a new bdd with updated
				     modifies information
				     *)
      try
      let bdd1 = 
	if is_deref_of_global lv fname then 
	  begin
	    let fnamebdd = Hashtbl.find lvalue_to_bdd_map (Expression.Symbol fname) in
	    let lvaldd = Hashtbl.find lvalue_to_bdd_map (Expression.fullsimp_lval lv) in
	    (*CaddieBdd.bddPrint fnamebdd ;
	      CaddieBdd.bddPrint lvaldd ;
	      *)
	    CaddieBdd.bddOr bdd (CaddieBdd.bddAnd fnamebdd (CaddieBdd.replace lvaldd (CaddieBdd.toMap bddManager !stvarMap)))
	  end
	else bdd
      in
(*      CaddieBdd.bddPrint bdd1 ; *)
      if is_deref_of_formal lv fname then  (* update bdd *)
	begin
	  let fnamebdd = Hashtbl.find lvalue_to_bdd_map (Expression.Symbol fname) in
	  let lvaldd = Hashtbl.find lvalue_to_bdd_map (Expression.fullsimp_lval lv) in
	  CaddieBdd.bddOr bdd (CaddieBdd.bddAnd fnamebdd (CaddieBdd.replace lvaldd (CaddieBdd.toMap bddManager !stvarMap)))
	end
      else bdd1
      with Not_found -> failwith "check_lval"  
    in
    (* update transformer for a function call.
       Mark as modified any actual that may be modified by the callee *)
    let check_funcall (caller : string) (callee : string) (arglist : Expression.expression list) bdd =
      (* for each argument, add to modifies list if callee modifies the argument *)
      let (formals, varflag) = match C_SD.lookup_formals callee with C_SD.Fixed l -> (l, false) | C_SD.Variable l -> (l,true) in
      try
      let callerbdd = Hashtbl.find lvalue_to_bdd_map (Expression.Symbol caller) in
      let rec _fold2 flag f l1 l2 bdd =
	match (l1, l2) with 
	  ([], [])  -> 
	    begin
	      bdd
	    end
	| ([], _) 
	| (_, []) ->
	    begin
	      if flag then bdd else failwith "_fold2: argument list mismatch for fixed argument function"
	    end
	| (a:: rest1, b :: rest2) ->
	    let newbdd = f a b bdd in
	    _fold2 flag f rest1 rest2 newbdd
      in
      _fold2 varflag 
	(fun fml -> fun actual -> fun bdd ->
	  if (is_modified fml callee bdd) 
	  then (* modify corresp arg in caller *)
	    begin  
	      let argvalbdd = 
		match actual with
		  Expression.Lval l 
		| Expression.Unary (Expression.Address, Expression.Lval l) -> 
		    Hashtbl.find lvalue_to_bdd_map (Expression.fullsimp_lval l)
		| Expression.Binary(Expression.FieldOffset, Expression.Unary(Expression.Address, Expression.Lval l), _) 
		| Expression.Binary(Expression.FieldOffset, Expression.Lval l, _) 
		| Expression.Binary(Expression.Plus, Expression.Lval l, _) 
		| Expression.Binary(Expression.Minus, Expression.Lval l, _) -> (* assume it's always ptr + int for arrays*)
		    Hashtbl.find lvalue_to_bdd_map (Expression.fullsimp_lval l)
		| _ -> failwith ("match failed in check_funcall for actual " ^ (Expression.toString actual))
	      in
	      let newbdd = CaddieBdd.bddOr bdd (CaddieBdd.bddAnd callerbdd
						  (CaddieBdd.replace argvalbdd (CaddieBdd.toMap bddManager !stvarMap)))
	      in
	      newbdd
	    end 
	  else bdd
	      ) 
	(List.map (fun s-> Expression.Symbol s) formals) arglist bdd
      with Not_found -> failwith "check_funcall"
    in
    let do_one_fun fname =  (* this is the workhorse that populates the BDDs *)
      Message.msg_string Message.Normal ("Doing function "^fname) ;
      if not (C_SD.is_defined fname) 
      then 
	() (* This function is not defined. So we ignore its effects *) 
      else
	begin
	  let fn_bdd = Hashtbl.find lvalue_to_bdd_map (Expression.Symbol fname)  in
	  let process_one_edge e =
	    (* for each edge, see if the variable written to is a global or a dereference
	       of a formal.
	       For a function call, add the ith actual to the modified list if
	       the callee modifies its ith formal.
	   *)
	    let ecmd = C_SD.get_command e in
	    match ecmd.C_SD.Command.code with
	    | C_SD.Command.Block l ->
		begin
		  let lvalswritten = ecmd.C_SD.Command.written in
                (*
		  List.iter (fun l -> Message.msg_string Message.Normal ("Lval " ^ (Expression.lvalToString l))) lvalswritten ;
		  *)
		  let newbdd = List.fold_left (check_lval fname) !mod_bdd lvalswritten 
		  in
	      (* check if bdd has changed. If it has, update changed to true. *)
		  changed := if !changed then !changed else not (CaddieBdd.bddEqual !mod_bdd newbdd) ;
		  mod_bdd := newbdd ;
		end
	    | C_SD.Command.FunctionCall fexp ->
		begin
		  let callee = 
		    match fexp with
		      Expression.FunctionCall (fncallee, _) 
		    | Expression.Assignment (Expression.Assign, _, Expression.FunctionCall(fncallee, _)) ->
			begin 
			  match fncallee with
			    Expression.Lval (Expression.Symbol s) -> s
			  | _ -> 
			      if (Options.getValueOfBool "nofp") then "#fptr" else failwith "function ptr call"
			end
		    | _ -> failwith "strange function call"
		  in
		  if callee = "#fptr" then () (* assume that function pointer calls do not modify
						 globals or formals 
						 *)
		  else
		    begin
		      begin
			match fexp with
			  Expression.FunctionCall (fncallee, arglist) ->
			    let bdd1 = check_funcall fname callee arglist !mod_bdd in
		  (* check if bdd has changed. If it has, update changed to true. *)
			    changed := if !changed then !changed else not (CaddieBdd.bddEqual !mod_bdd bdd1) ;
			    mod_bdd := bdd1
			|	Expression.Assignment (Expression.Assign, target, Expression.FunctionCall(fncallee, arglist)) ->
			    let bdd1 = check_lval fname !mod_bdd target in
			    let bdd2 = check_funcall fname callee arglist bdd1 in
		  (* check if bdd has changed. If it has, update changed to true. *)
			    changed := if !changed then !changed else not (CaddieBdd.bddEqual !mod_bdd bdd2) ;
			    mod_bdd := bdd2 
		      end ;
		      let callerbdd = try Hashtbl.find lvalue_to_bdd_map (Expression.Symbol fname) with _ -> failwith ("Aha : "^fname) in
	            (* for each global that the callee modifies, add modifies information for
		       the caller *)
		      let new_bdd = 
			List.fold_left (fun bdd -> fun g ->
			  if (is_modified g callee bdd) then
		          (* update bdd *)
			    begin
			      let gbdd = Hashtbl.find lvalue_to_bdd_map g in
			      CaddieBdd.bddOr bdd (CaddieBdd.bddAnd callerbdd 
						     (CaddieBdd.replace gbdd (CaddieBdd.toMap bddManager !stvarMap)))
			    end
			  else bdd
			      ) !mod_bdd globallist in
	      		  (* check if bdd has changed. If it has, update changed to true. *)
		      changed := if !changed then !changed else not (CaddieBdd.bddEqual !mod_bdd new_bdd) ;
		      mod_bdd := new_bdd 
			   
		  end
		end
		  
	    | C_SD.Command.Havoc l ->
		begin
		  let newbdd = check_lval fname !mod_bdd l in
	      (* check if bdd has changed. If it has, update changed to true. *)
		  changed := if !changed then !changed else not (CaddieBdd.bddEqual !mod_bdd newbdd) ;
		  mod_bdd := newbdd 
		end
	    | C_SD.Command.HavocAll -> failwith "modifies: HavocAll not implemented"
	    | _ -> () (* uninteresting -- no modification takes place *)
	  in
	  try
	    ignore (C_SD.map_edges_fname process_one_edge fname)
	      
	  with C_SD.NoSuchFunctionException _ ->
	(* if the function code is not available then assume that the function
	   modifies all pointer parameters but not global variables ?
	   *)
       (*
	 Note that the effect is that all functions whose bodies are not present
	 are assumed not to modify any local/global 
	 *)
	    () (* TO DO *)

	end
    in
    
    (* Iterate (and compute fixpoint) for all functions. 
       *)

  List.iter do_one_fun (C_SD.list_of_functions () ) ;
  done ;
  !mod_bdd

let is_modified bdd func lval =
  try
    let calleebdd = Hashtbl.find lvalue_to_bdd_map (Expression.Symbol func) in
    let lvaldd = 
      Hashtbl.find lvalue_to_bdd_map (Expression.fullsimp_lval lval) in
    let primedlvaldd = (CaddieBdd.replace lvaldd (CaddieBdd.toMap bddManager !stvarMap)) in
    let conjunct = CaddieBdd.bddAnd calleebdd primedlvaldd in
    if (CaddieBdd.bddEqual (CaddieBdd.bddAnd bdd conjunct) bddZero) then
      false
    else
      true
  with Not_found -> Message.msg_string Message.Normal ("Not found raised on " ^ func ^ (Expression.lvalToString lval)) ; false

let test_mod_analysis modbdd =
  let (globallist, locallist, funlist) = get_all_lvals () in
  (* for each function, print the globals and formals it touches *)

  List.iter 
    (fun fn ->
      let fname = match fn with Expression.Symbol s -> s in
      Message.msg_string Message.Normal ("Function "^fname);
      let fbdd = Hashtbl.find lvalue_to_bdd_map fn in
      let formals = match C_SD.lookup_formals fname with C_SD.Fixed l | C_SD.Variable l -> l in
      List.iter (fun g -> 
	if (is_modified modbdd fname g)
	then
	  Message.msg_string Message.Normal ("modifies "^(Expression.lvalToString g))
	else ()
      ) globallist ;
      List.iter (fun g -> 
	if (is_modified modbdd fname (Expression.Symbol g))
	then Message.msg_string Message.Normal ("modifies argument "^(g))
	else ()
      ) formals ;
      Message.msg_string Message.Normal "That's it"
	)

    funlist




