(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *)

module Predicate = Ast.Predicate
module Expression = Ast.Expression
module C_System_Descr = BlastCSystemDescr.C_System_Descr
module C_Command = BlastCSystemDescr.C_Command

module IntSet = Set.Make (struct type t = int let compare a b = b - a end )

module Stats = Bstats


open BlastArch.ModelCheckOutcome


(********************************************************************)
(*************** Code for query caching move to another file ********)
(********************************************************************)
module GL = Genlex
module ST = Stream

(* type qid_t = (int * int) list
 * type query_t = (qid_t,({-1..unsafe,0..??,+1..safe}, int)) *)

let all_edge_id = ref []

let get_qcfname () = 
  let _fn = Options.getValueOfString "qcfile" in
  let _fn' = if _fn = "" then (Options.getValueOfString "mainsourcename") else _fn in
  _fn'^".qc"

let query_cache_table = Hashtbl.create 101 



let get_query_tuples fname = 
  let raw_entries = try Misc.string_list_of_file fname with _ -> [] in
  let proc raw = 
    try
      let things = Misc.chop raw "," in
      match things with
      [s1;s2;s3] ->
        let ids_l = Misc.chop s1 ";" in
        let ids = 
          List.map 
          (fun s -> 
            match Misc.words_of_string s with [w1;w2] ->(int_of_string w1,int_of_string w2))
          ids_l
        in
        let res = int_of_string (List.hd (Misc.words_of_string s2)) in
        let time = int_of_string (List.hd (Misc.words_of_string s3)) in
        Some(ids,(res,time))
    with e -> 
      Message.msg_string Message.Error ("qc line: "^raw);
      Message.msg_string Message.Error ("Raises: "^(Printexc.to_string e));
      None
  in
  Misc.map_partial proc raw_entries
  
let load_query_cache () =
  let fname = get_qcfname () in
  let qlist = get_query_tuples fname in
  List.iter (fun (k,d) -> Hashtbl.add query_cache_table k d) qlist


let lookup_user_queries () = 
  let fname = Options.getValueOfString "qf" in
  List.map fst (get_query_tuples fname)

let query_id_to_string qid = 
  let qid' = List.map (fun (x,y) -> Printf.sprintf " %d %d " x y) qid in
  String.concat ";" qid'

let update_query_cache qid (res,time) = 
  let fname = get_qcfname () in 
  let _ = Hashtbl.add query_cache_table qid (res,time) in
  let qids = query_id_to_string qid in
  let out_string = Printf.sprintf "%s , %d , %d \n\n" qids res time in
  Misc.append_to_file fname out_string
 
let lookup_query_cache qid = 
  try Some (Hashtbl.find query_cache_table qid) with Not_found -> None

let query_id elist = 
  let r = List.map (fun e -> C_System_Descr.location_coords (C_System_Descr.get_source e))  elist
  in List.sort compare r
 

(*******************************************************************)










(*****************************************************************************************)

(* OBSOLETE: NOT SUPPORTED 
let inputCustomAbs () =
  let read_abs_info filename =
    let doParse inchan =
      let _ = VampyreErrormsg.startFile filename in
      let lexbuf = Lexing.from_channel inchan in
      let _ = VampyreErrormsg.theLexbuf := lexbuf in
      Absparse.main Absflex.token lexbuf 
    in
    ignore (Message.msg_string Message.Normal "Reading in abstraction ...");
    try
      let inchan = open_in filename
      in
      let a = doParse inchan
      in
      ignore (Message.msg_string Message.Normal " Abstraction read.\n") ;
      a
    with
      Sys_error _ -> (Message.msg_string Message.Error ("Cannot find abstraction info in "^filename^".\n"); ([],None,[]))
    | VampyreErrormsg.Error -> (Message.msg_string Message.Error ("Error raised in reading abstraction \n"); ([],None,[]))
  in
  let abs_file = Options.getValueOfString "custom" in
  if (abs_file = "") then
    ([],None,[]) (* no special types nothing at all ... *)
  else
    read_abs_info abs_file
*)

(*****************************************************************************************)


let constructInitialPredicates predFile =
  let readAndAddPredicates filename =
    let doParse inchan =
      let _ = VampyreErrormsg.startFile filename in
      let lexbuf = Lexing.from_channel inchan in
      let _ = VampyreErrormsg.theLexbuf := lexbuf in
      Inputparse.main Inputflex.token lexbuf 
    in
    ignore (Message.msg_string Message.Normal "Reading in seed predicates...");
    try
      let inchan = open_in filename
      in
      let a = doParse inchan
      in
	ignore (Message.msg_string Message.Normal "Seed predicates read.\n") ;	
	List.iter (fun a -> Message.msg_string Message.Normal (Predicate.toString a)) a ;
	Message.msg_string Message.Normal ("Read "^(string_of_int (List.length a))^" predicates");
	a
    with
      Sys_error _ -> (Message.msg_string Message.Error ("Cannot find predicate information in "^filename^".\n"); [])
    | VampyreErrormsg.Error -> (Message.msg_string Message.Error ("Error raised in reading seed predicates.\n"); [])
  in
  if (predFile = "") then
    (* Nothing to do *)
    []
  else
    readAndAddPredicates predFile

let constructInvariants () =
  let readAndAddInvariants filename =
    let doParse inchan =
      let _ = VampyreErrormsg.startFile filename in
      let lexbuf = Lexing.from_channel inchan in
      let _ = VampyreErrormsg.theLexbuf := lexbuf in
      Inputparse.main Inputflex.token lexbuf 
    in
    ignore (Message.msg_string Message.Normal "Reading in invariants...");
    try
      let inchan = open_in filename
      in
      let a = doParse inchan
      in
      ignore (Message.msg_string Message.Normal "Invariants read.\n") ;
      List.iter (fun a -> Message.msg_string Message.Normal (Predicate.toString a)) a ;
       a
    with
      Sys_error _ -> (Message.msg_string Message.Error ("Cannot find invariant information in "^filename^".\n"); [])
    | VampyreErrormsg.Error -> (Message.msg_string Message.Error ("Error raised in reading invariants.\n"); [])
  in
  let invFile = Options.getValueOfString "inv" in
  if (invFile = "") then
    (* Nothing to do *)
    []
  else
    readAndAddInvariants invFile

let constructEnvironmentAssumptions () = 
   let readAndAddAssumptions filename =
    let doParse inchan =
      let _ = VampyreErrormsg.startFile filename in
      let lexbuf = Lexing.from_channel inchan in
      let _ = VampyreErrormsg.theLexbuf := lexbuf in
      Inputparse.env_assumptions Inputflex.token lexbuf 
    in
    ignore (Message.msg_string Message.Normal "Reading in environment assumptions ...");
    try
      let inchan = open_in filename
      in
      let a = doParse inchan 
      in
      ignore (Message.msg_string Message.Normal "Assumptions read.\n") ;
      List.iter (fun (l,p) -> Message.msg_string Message.Normal ((Expression.lvalToString l )^" : "^(Predicate.toString p))) a ;
       a
    with
      Sys_error _ -> (Message.msg_string Message.Error ("Cannot find assumption information in "^filename^".\n"); [])
    | VampyreErrormsg.Error -> (Message.msg_string Message.Error ("Error raised in reading assumptions.\n"); [])
  in
  let asmFile = Options.getValueOfString "easm" in
  if (asmFile = "") then
    (* Nothing to do *)
    []
  else
    readAndAddAssumptions asmFile

let reachedLoc = ref None
let final_string = ref ""
		     
 


(* single query *)
let model_check o_err_locs entries seedPreds =
  let module This_C_Abstraction = Abstraction.Make_C_Abstraction(C_System_Descr) in
  let module This_Cf_C_Abstraction = CfAbstraction.Make_C_Abstraction(C_System_Descr) in
  let module This_LMC =
    LazyModelChecker.Make_Lazy_Model_Checker(This_C_Abstraction) in
  let module This_LMC_CF =
    CfLazyModelChecker.Make_Lazy_Model_Checker(This_Cf_C_Abstraction) in  
  let module This_LMC_CFB =
    CfbLazyModelChecker.Make_Lazy_Model_Checker(This_Cf_C_Abstraction) in 
  let module Region = This_C_Abstraction.Region in
  let _ = 
    This_C_Abstraction.check_races := 
    let cr = Options.getValueOfString "checkRace" in
      if ( cr = "tar" || cr = "extar") then
	begin
	  let rv = (Options.getValueOfString "racevar") in
	    Message.msg_string Message.Normal ("Checking for races on: "^rv);
	    Some (Ast.Expression.Symbol rv)
	end
      else None
  in
  This_C_Abstraction.assumed_invariants := Stats.time "read invariants" constructInvariants();  
  This_C_Abstraction.assumed_invariants_region := This_C_Abstraction.assumed_invariants_list_to_region();
  let _ = This_LMC.process_env_assumptions (constructEnvironmentAssumptions ()) in
  let initStateFile = Options.getValueOfString "init" in
  let initialState = Predicate.conjoinL (Stats.time "read initial state" constructInitialPredicates initStateFile) in
  let mainFunctions = Options.getValueOfStringList "main" in
    (* JHALA: Kill hacks *)
  let killLabel = Options.getValueOfString "killLoc" in
  let kill_location =
    match (Stats.time "Kill init" C_System_Descr.get_locations_at_label killLabel) with
        [l] -> Some(l)
      | []  ->
          if (killLabel = "") then ()
          else
            Message.msg_string Message.Normal ("Kill label '" ^ killLabel ^
                                            "' was not found in the source file(s)") ;
          None
      | _   ->
          Message.msg_string Message.Error ("Kill label '" ^ killLabel ^
                                            "' appears multiple times in the source file(s)") ;
          exit 1
  in  
  let _ = This_C_Abstraction.initialize_abstraction () in
  let _ = This_Cf_C_Abstraction.initialize_abstraction () in

  let _ = Message.msg_string Message.Normal "Done adding seed predicates" in

    Message.msg_string Message.Normal "\n\n\n********** Now running the model-checker **********\n\n" ;
    flush stdout;
    begin
      let indent_loc = "                                                " in
      let print_op indent_op op =
        let target_loc = C_System_Descr.get_target op
        and c_command = C_System_Descr.get_command op
        in
        Message.msg_string Message.Normal
          (indent_op ^ (C_Command.to_string c_command)) ;
        Message.msg_string Message.Normal
          (indent_loc ^ (C_System_Descr.location_to_string target_loc)) ;
        match c_command.C_Command.code with
          C_Command.FunctionCall _ ->
            indent_op ^ "   "
        | C_Command.Block l ->
            begin
              match (List.hd (List.rev l)) with
              | C_Command.Return _ ->
                  String.sub indent_op 0 ((String.length indent_op) - 3)
              | _ -> indent_op
            end
        | _ -> indent_op
      in
      let mc_return_val = 
	try
	  if (Options.getValueOfBool "cf")
	  then
	    begin
	      let init_region =
		let rec buildloclist lst loclist =
		  match lst with [] -> loclist
		  | a ::rest -> 
		      try
			buildloclist rest ((C_System_Descr.lookup_entry_location a) :: loclist) 
		      with 
			Not_found -> 
			  (Format.printf "Creating init_region: No function %s defined in program. Exiting\n" (a);
			   failwith ("No function "^(a)^" defined"))
		in
		let loclist = match entries with
		  None -> buildloclist mainFunctions []
		| _ -> failwith "Unimplemented CF multi-entry" in
		This_Cf_C_Abstraction.create_region loclist false seedPreds initialState
		  
	      and error_region =
		match o_err_locs with
		  [] -> None
		| _ ->
		    Some (This_Cf_C_Abstraction.create_region o_err_locs true Predicate.True Predicate.True)
		
	      in 
              let cfmc i e = match This_LMC_CF.model_check i e with No_error -> 1 | _ -> 0 in
              let cfbmc i (Some e) = 
                let start = Unix.times () in
                let rv = This_LMC_CFB.model_check i e in
                  let _ = 
                    if (Options.getValueOfBool "reachloc") then
                    let finish = Unix.times () in
                    let time = int_of_float (finish.Unix.tms_utime -. start.Unix.tms_utime) in
                    let qid_all = !all_edge_id in
                    let qid_safe = List.filter (fun id -> not (This_LMC_CFB.is_reachable id)) qid_all in
                
                    List.iter (fun id -> update_query_cache [id] (1,time)) qid_safe
                  in
                rv
              in
              let model_check = if (Options.getValueOfBool "cfb") then cfbmc else cfmc in
              let mc_return_val = Stats.time "Actual model check[cf]" (model_check init_region) error_region in
              let _ = final_string := "" in
              if mc_return_val = 1 then
                let _ = final_string := "\n\nNo error found.  The system is safe :-)\n" in
                Message.msg_string Message.Major !final_string
              else 
                let _ = final_string := "\n\nError found! The system is unsafe :-(\n\n" in
		      Message.msg_string Message.Major !final_string;
	    end
	  else
	    begin 
	      (* vanilla tree unroll starts here *)
	      let init_region =
		let rec buildloclist lst loclist =
		  match lst with [] -> loclist
		  | a ::rest -> 
		      try
			buildloclist rest ((C_System_Descr.lookup_entry_location a) :: loclist) 
		      with 
			Not_found -> 
			  (Format.printf "Creating init_region: No function %s defined in program. Exiting\n" (a);
			   failwith ("No function "^(a)^" defined"))
		in
		let loclist = match entries with
		  None -> buildloclist mainFunctions []
		| Some entries -> entries in
		This_C_Abstraction.create_region 
		  loclist false seedPreds initialState
		  
	      and error_region =
		match o_err_locs with
		  [] -> None
		| _ -> Some (This_C_Abstraction.create_region o_err_locs true Predicate.True Predicate.True)
	      in
	      
	      let mc_return_val =
		match (Options.getValueOfString "checkRace" ) with
		|  "tar" -> (This_LMC.model_check_tar init_region error_region) 
                |  "extar" -> (This_LMC.model_check_extar init_region error_region) 
		|  _ -> (This_LMC.model_check init_region error_region) in
	      match (mc_return_val) with
		No_error ->
		  if not (Options.getValueOfBool "interactive") then
		    let _ = final_string := "\n\nNo error found.  The system is safe :-)\n"
		    in
		      Message.msg_string Message.Major !final_string
			  
	      | Error_reached (reg_i, op_list, reg_e) ->
		  if Options.getValueOfBool "interactive" then begin
		    let coords = This_C_Abstraction.Region.get_location_id reg_e in
		    let loc = C_System_Descr.lookup_location_from_loc_coords coords in
		    reachedLoc := Some loc
		  end else begin
		    let _ = final_string := "\n\nError found! The system is unsafe :-(\n\n Error trace:\n" in
			Message.msg_string Message.Major !final_string;
		    (* this next line raises a seg fault for some reason...
		    Message.msg_printer Message.Normal This_C_Abstraction.Region.print reg_i ;
		    Message.msg_string Message.Normal "\n\n" ;
		     *)
		    Message.msg_string Message.Normal
		      (indent_loc ^ (C_System_Descr.location_to_string
				       (C_System_Descr.get_source (List.hd op_list)))) ;
		    let _ = List.fold_left print_op "" op_list in
		    Message.msg_string Message.Normal "\n\n" ;
		    Message.msg_printer Message.Normal This_C_Abstraction.Region.print reg_e;
		    (try 
		      Message.msg_string Message.Normal "Test case:" ;
                      let rec _prtest l = match l with [] -> () 
                      | (sym, v) :: t -> 
		  	Message.msg_string Message.Normal (Expression.toString sym);
		  	Message.msg_string Message.Normal (string_of_float v) ;
		  	_prtest t
                      in
                      _prtest (This_C_Abstraction.genTests reg_i) ;
                    with _ -> (Message.msg_string Message.Error "Test generation failed." ; ) ; (* end try *) );
		    (*if (Options.getValueOfBool "xml") then Absutil.dumpXml
                    * op_list *)
		  end
	      | Race_condition ((reg1,oplist1, reg1'), (reg2, oplist2, reg2'), lv1, lv2) ->
		  let _ = final_string := "\n\nRace condition found! The system is unsafe :-(\n\n\n"
		     ^(Printf.sprintf "\n On the lvalues %s, %s \n" (Ast.Expression.lvalToString lv1) 
			 (Ast.Expression.lvalToString lv2))
		  in
		  Message.msg_string Message.Major "\n\nRace condition found! The system is unsafe :-(\n\n\n";
		  Message.msg_string Message.Normal (Printf.sprintf "\n On the lvalues %s, %s \n" 
						       (Ast.Expression.lvalToString lv1) (Ast.Expression.lvalToString lv2));
		  Message.msg_string Message.Error (Printf.sprintf "\n On the lvalues %s, %s \n" 
						      (Ast.Expression.lvalToString lv1) (Ast.Expression.lvalToString lv2));
	          (* print out the first trace *)
		  Message.msg_string Message.Normal "Printing out the first trace.\n" ;
		  Message.msg_printer Message.Normal This_C_Abstraction.Region.print reg1 ;
		  Message.msg_string Message.Normal "\n\n" ;
	  
		  Message.msg_string Message.Normal
		    (indent_loc ^ (C_System_Descr.location_to_string
				     (C_System_Descr.get_source (List.hd oplist1)))) ;
		  let _ = List.fold_left print_op "" oplist1 in
		  Message.msg_string Message.Normal "\n\n" ;
		  Message.msg_printer Message.Normal This_C_Abstraction.Region.print reg1';
	          (* print out the second trace *)
		  Message.msg_string Message.Normal "\n\nPrinting out the second trace.\n" ;
		  Message.msg_printer Message.Normal This_C_Abstraction.Region.print reg2 ;
		  Message.msg_string Message.Normal "\n\n" ;
	  
		  Message.msg_string Message.Normal
		    (indent_loc ^ (C_System_Descr.location_to_string
				     (C_System_Descr.get_source (List.hd oplist2)))) ;
		  let _ = List.fold_left print_op "" oplist2 in
		  Message.msg_string Message.Normal "\n\n" ;
		  Message.msg_printer Message.Normal This_C_Abstraction.Region.print reg2';
	    end
	with e -> 
	  begin 
	    if (Options.getValueOfBool "cf") then This_Cf_C_Abstraction.dump_abs ()
	    else This_C_Abstraction.dump_abs (); 
            Message.msg_string Message.Error ("Exception raised :("^(Printexc.to_string e));
	    raise e
	  end
      in 
      ()
    end ;
    if (Options.getValueOfBool "cf") 
    then
      begin
        let print_stats = if Options.getValueOfBool "cfb" then
          This_LMC_CFB.print_stats else This_LMC_CF.print_stats
        in
	Message.msg_string Message.Normal "\n\nModel checker stats:\n" ;
	Message.msg_printer Message.Normal print_stats () ;
	Message.msg_string Message.Normal "\n\nAbstractor stats:\n" ;
	Message.msg_printer Message.Normal This_Cf_C_Abstraction.print_stats ();
      
        This_Cf_C_Abstraction.exit_abstraction ()
    end
    else
      begin
	Message.msg_string Message.Normal "\n\nPhi Regions:\n" ;
	Message.msg_string Message.Normal (This_LMC.print_phis false) ;
	Message.msg_string Message.Normal "\n\nModel checker stats:\n" ;
	Message.msg_printer Message.Normal This_LMC.print_stats () ;
	Message.msg_string Message.Normal "\n\nAbstractor stats:\n" ;
	Message.msg_printer Message.Normal This_C_Abstraction.print_stats () ; 
      
        This_Cf_C_Abstraction.exit_abstraction ()
      
      end
  (* end model_check () *)

        
let get_edges funlabel f = 
  (* given a function f, find out all edges in its CFA where it	  calls the function funlabel *)
   let edgelist = ref [] in
	let _worker e =
	  match (C_System_Descr.get_command e).C_Command.code with
	    C_Command.FunctionCall ex ->
	      begin
		match ex with
		  Expression.Assignment (_, _, Expression.FunctionCall (Expression.Lval (Expression.Symbol fl), _)) 
		| Expression.FunctionCall (Expression.Lval (Expression.Symbol fl), _) ->
		    if fl = funlabel then edgelist := e :: !edgelist 
		| _ -> ()
	      end
	  | _ -> ()
	in
	C_System_Descr.map_edges_fname _worker f ;
	!edgelist

 

let get_lvals_on_guards elist =
  List.fold_left (fun allofthem e ->
    let lvs_of_e = 
      let src = C_System_Descr.get_source e in
      let incoming = C_System_Descr.get_ingoing_edges src in
      List.fold_left (fun sofar inedge ->
        match (C_System_Descr.get_command inedge).C_Command.code with
        C_Command.Pred p ->
          begin
            let lv_list = Ast.Predicate.lvals_of_predicate p in
            Message.msg_string Message.Normal 
            ("The lvalues in this edge are: "^(Misc.strList (List.map Ast.Expression.lvalToString lv_list)));
            lv_list @ sofar
          end
        | _ -> sofar ) [] incoming 
    in
    lvs_of_e @ allofthem) [] elist


 let local_synt_partition edges = 
   let fname_edge_table = Hashtbl.create 11 in
   List.iter 
   (fun e -> let src = C_System_Descr.get_source e in
   let incoming = C_System_Descr.get_ingoing_edges src in 
   List.iter 
   (fun inedge ->
     match (C_System_Descr.get_command inedge).C_Command.code with
     C_Command.Pred p ->
       begin
         let lv = Ast.Predicate.varsOfPred p in
         Message.msg_string Message.Normal "The lvalues in this edge are: ";
         List.iter (Message.msg_string Message.Normal ) lv ;
         let thisfp = List.hd lv in
         try let old = Hashtbl.find fname_edge_table thisfp in
             Hashtbl.replace fname_edge_table thisfp (e :: old)
         with Not_found -> Hashtbl.add fname_edge_table thisfp ([ e ])
       end
     | _ -> failwith "Strange incoming edge!") 
   incoming) 
   edges ;
   let l = ref [] in
   Hashtbl.iter (fun key data -> l := data::!l) fname_edge_table ;
   !l
   

(* TBD:SPEED quadratic in elist *)
let aliased_lvalue_cluster elist =
  let cl_tbl = Hashtbl.create 101 in
  let cllv_tbl = Hashtbl.create 101 in
  let keys = ref [] in
  let _proc e = 
    let lv_l = get_lvals_on_guards [e] in
    let check idx = 
      let lvi_l = Hashtbl.find cllv_tbl idx in
      let hit = 
        Misc.list_exists_cross2 (fun l l' -> l=l' ||
        (*AliasAnalyzer.askAliasDatabase l l'*) AliasAnalyzer.queryAlias
        (Expression.Lval l) (Expression.Lval l'))
        lvi_l lv_l
      in
      let _ = 
        if hit then 
        let _ = Misc.hashtbl_check_update cl_tbl idx e in 
        Hashtbl.replace cllv_tbl idx (Misc.sort_and_compact (lv_l@lvi_l)) 
      in
      hit
    in
    try ignore(List.find check !keys) with Not_found ->
      let newkey = List.length !keys + 1 in
      Hashtbl.replace cllv_tbl newkey lv_l;
      Hashtbl.replace cl_tbl newkey [e];
      keys := newkey ::!keys;
  in 
  let output_cluster k = 
            let qids = 
              let qid = query_id (Hashtbl.find cl_tbl k) in
              (query_id_to_string qid) 
            in
            let lvs = 
              let ll = List.map Ast.Expression.lvalToString (Hashtbl.find cllv_tbl k) in
              String.concat "," ll
            in
            Message.msg_string Message.Normal 
            (Printf.sprintf "[QUERY] id: %s lvals: %s" qids lvs)
  in
    List.iter _proc elist;
    List.iter output_cluster !keys;
    List.map snd (Misc.hashtbl_to_list cl_tbl)



let get_user_query uqid_list all_call_edges = 
  let elist = List.flatten all_call_edges in
  let get_query_from_qid uqid = 
    let tbl = Hashtbl.create 31 in
    let rec proc uql = 
      match uql with [] -> ()
      | (x,y)::t -> Hashtbl.replace tbl (x,y) true;proc t 
    in 
    let ff e = let [id] = query_id [e] in Hashtbl.mem tbl id in
    proc uqid;
    List.filter ff elist
  in
  List.map get_query_from_qid uqid_list

    
let cluster_edges all_call_edges = 
  let funlabel = Options.getValueOfString "funlabel" in
  let funmode = Options.getValueOfInt "funmode" in
  let epl =
    if funlabel = "" then [List.flatten all_call_edges] 
    else if funmode = 0 then all_call_edges 
    else if funmode = 1 then List.map (fun e -> [e]) (List.flatten all_call_edges)
    else if funmode = 2 then List.flatten (List.map local_synt_partition all_call_edges)
    else if funmode = 3 then Stats.time "alias cluster" aliased_lvalue_cluster (List.flatten all_call_edges)
    else if funmode = 4 then get_user_query [Misc.group_pairwise (Options.getValueOfIntList "qid")] all_call_edges
    else if funmode = 5 then get_user_query (lookup_user_queries ()) all_call_edges
    else all_call_edges
  in
  let s1 = ("#Queries : "^(string_of_int (List.length epl))) in
  Message.msg_string Message.Major s1;
  Message.msg_string Message.Error s1;
  List.map (List.sort compare) epl


let filter_preds seedPreds elist =
  let lvt = Hashtbl.create 101 in
  let is_relevant p = 
    let plvs = Ast.Predicate.lvals_of_predicate p in
    let tg = Options.getValueOfBool "tg" in
    List.exists (fun lv -> Hashtbl.mem lvt lv || (tg &&
    (C_System_Descr.is_global lv))) plvs
  in
  let _ =  List.iter (fun lv -> Hashtbl.replace lvt lv true) (get_lvals_on_guards elist) in
  let atoms = Ast.Predicate.getAtoms seedPreds in
  let atoms' = List.filter is_relevant atoms in
  Predicate.conjoinL atoms'
  
  
  
(* calls model_check with a particular query *)              
let modelCheck o_err_locs entries seedPreds =
  let _ = if (Options.getValueOfBool "qc") then load_query_cache () in
   let funlabel = Options.getValueOfString "funlabel" in
   let replacefunlabel = "__BLAST_"^funlabel in
   let callers = Misc.sort_and_compact 
   (List.map (snd) (C_System_Descr.calls_made_into funlabel)) in    
   let d = ref 0 in
   let cumulative = ref 0 in
   let successes = ref 0 in
   let failures = ref 0 in
   let no_new = ref [] and 
       timeout = ref [] and
       bugs = ref [] 
   in
   let replace_edges label elist =
     Message.msg_string Message.Error ("Replace Edges with :"^label);
     let rep e =
        match (C_System_Descr.get_command e).C_Command.code with
       C_Command.FunctionCall (Expression.FunctionCall (_,args)) ->
         C_System_Descr.replace_code_in_cfa e 
         (Expression.FunctionCall (Expression.Lval (Expression.Symbol label),args))
        | _ -> Message.msg_string Message.Error ("Bad edge in replace_edges!")
     in
       List.iter rep elist                      
   in
   let run_one_check elist =
     if elist = [] then () else 
       begin
         let error_lvs = get_lvals_on_guards elist in
         let _ = C_System_Descr.set_error_lvals error_lvs in
         let edge_fname e =
           C_System_Descr.get_location_fname (C_System_Descr.get_source e) 
         in
         let edge_fnames = List.map edge_fname elist in
         let efn = Misc.strList edge_fnames in
         Options.set_errorseed  edge_fnames ;
         Message.msg_string Message.Major (Printf.sprintf "Call %d (cumulative %d) in fun:%s" !d !cumulative efn ) ;
         incr d ;
         cumulative := !cumulative + (List.length elist) ; 
         if (!d <= Options.getValueOfInt "skiptest" 
             || 
             (Options.getValueOfBool "selecttest") && not (List.mem !d (Options.getValueOfIntList "checktest"))) then
	    Message.msg_string Message.Normal ("Skipping query # "^(string_of_int !d)) 
         else
	  begin
          let qid = query_id elist in
          let qids = query_id_to_string qid in
          let ms1 = (Printf.sprintf "Trying error call # %d (cumulative %d) edges: %d, in functions %s, qid: %s" 
            !d !cumulative (List.length elist) efn qids) 
          in
            Message.msg_string Message.Error ms1;
	  Message.msg_string Message.Major ms1 ;
          Message.msg_string Message.Major ("Lvalues on edges in this call: ");
          Message.msg_string Message.Error ("Lvalues on edges in this call: ");
          List.iter (fun lv -> let ls = Ast.Expression.lvalToString lv in 
            Message.msg_string Message.Major ls ; Message.msg_string Message.Error ls ) 
            (get_lvals_on_guards elist) ;
          let start = Unix.times () in
          let (outcome,time) =
            match lookup_query_cache qid with
            Some(o,t) -> 
              begin
                Message.msg_string Message.Error  (Printf.sprintf "Query Cache: %d, time = %d" o t); 
                (o,t)
              end
            | None -> 
                  begin
                    let oref = ref 0 in
                    replace_edges replacefunlabel elist;
                    begin
                    try 
                      let seedPreds = filter_preds seedPreds elist in
                      model_check o_err_locs entries seedPreds  ;
                      Message.msg_string Message.Error !final_string;
                      if (Misc.is_substring !final_string ":-(" ) then 
                        if (Options.getValueOfBool "norefine") then 
                           (no_new := !d :: !no_new)
                        else
                        (bugs := !d :: !bugs;  oref := -1) 
                      else ( oref := 1) 
                    with 
                    Ast.TimeOutException ->
                      (Message.msg_string Message.Normal "This check times out!"; 
                      timeout:= !d :: !timeout)
                    | e -> 
                      (Message.msg_string Message.Error (Printexc.to_string e) ;
                      Message.msg_string Message.Error "Exception raised !" ; 
                      no_new := !d :: !no_new)
                    end;
                    replace_edges funlabel elist;
                    let time = 
                      let finish = Unix.times () in
                      int_of_float (finish.Unix.tms_utime -. start.Unix.tms_utime) 
                      in
                    update_query_cache qid (!oref,time);
                    (!oref,time)
                  end
          in
          let os = 
            if outcome = 1 then (incr successes; "SAFE") 
            else if outcome = -1 then (incr successes; "UNSAFE")
            else (no_new := !d::!no_new; "FAIL")
          in
          Message.msg_string Message.Major (Printf.sprintf "Query %d: %d s, %s" !d time os) ;
          Message.msg_string Message.Error (Printf.sprintf "Query %d: %d s, %s" !d time os);
         end 
    end
  in (* end run_one_check *)
 
  let run_reachloc_pass edge_partition_l =
    let proc i elist = 
      let issafe id = match (lookup_query_cache [id]) with Some(1,_) -> true | _ -> false in
      let qid = query_id elist in
      let (res,rss) = 
        if List.for_all issafe qid then (incr successes; (1,"SAFE"))
        else (no_new:=i::!no_new;(0,"FAIL")) 
      in
      Message.msg_string Message.Major (Printf.sprintf "Query %d: %d s, %s" i (-2) rss);
      Message.msg_string Message.Error (Printf.sprintf "Query %d: %d s, %s" i (-2) rss);
      update_query_cache qid (res,-2);
      (i+1)
    in
    if not (Options.getValueOfBool "cfb") then failwith "reachloc: only with CFB"
    else
      begin 
        let elist = List.flatten edge_partition_l in
        all_edge_id := query_id elist;
        (*replace_edges replacefunlabel elist;*)
        Options.set_errorseed [funlabel];
        let seedPreds = filter_preds seedPreds elist in 
        model_check o_err_locs entries seedPreds;
        (*replace_edges funlabel elist;*)
        ignore(List.fold_left proc 0 edge_partition_l)
      end
  in
  
  if funlabel = "" then model_check o_err_locs entries seedPreds
  else
    begin
      Message.msg_string Message.Error 
       (Printf.sprintf "Number of modeled functions: %d" (List.length (C_System_Descr.list_of_functions ())));
       let all_call_edges = (List.map (get_edges funlabel) callers) in
       let numcalls = List.length all_call_edges in
       let numinstr = List.fold_left (fun sofar thisone -> sofar + (List.length thisone)) 0 all_call_edges in
       Message.msg_string Message.Major ("#Functions calling error: "^(string_of_int (numcalls))) ;
       Message.msg_string Message.Error ("#Functions calling error: "^(string_of_int (numcalls))) ;
       Message.msg_string Message.Major ("#Callsites calling error: "^(string_of_int (numinstr))) ;
       Message.msg_string Message.Error ("#Callsites calling error: "^(string_of_int (numinstr))) ;
       let qids = query_id_to_string (query_id (List.flatten all_call_edges)) in 

       Message.msg_string Message.Major ("All query edges: "^qids);
       Message.msg_string Message.Error ("All query edges: "^qids);
       let edge_partition_l = cluster_edges all_call_edges in
       if (Options.getValueOfBool "cfb" && Options.getValueOfBool "reachloc")
       then (run_reachloc_pass edge_partition_l)
       else (List.iter run_one_check edge_partition_l);
             (* print stats *)
      Message.msg_string Message.Normal ("Number of safe uses is "^(string_of_int (!successes))) ;
      Message.msg_string Message.Error ("Number of safe uses is "^(string_of_int (!successes))) ;
      Message.msg_string Message.Normal ("Number of bugs is "^(string_of_int (!failures))) ;
      Message.msg_string Message.Error ("Number of bugs is "^(string_of_int (!failures))) ;
	List.iter (fun d -> Printf.fprintf Pervasives.stdout "Bug (%d) " d ;
		Printf.fprintf Pervasives.stderr "Bug (%d) " d ;
           ) !bugs ;
      Printf.fprintf Pervasives.stdout "\n" ;
      Printf.fprintf Pervasives.stderr "\n" ;
      Message.msg_string Message.Normal ("Failed : (Blast exception, including no new predicates )") ;
      Message.msg_string Message.Error ("Failed : (Blast exception, including no new predicates)") ;
	List.iter (fun d -> Printf.fprintf Pervasives.stdout "%d " d ;
		Printf.fprintf Pervasives.stderr "%d " d ;
           ) !no_new ;
      Message.msg_string Message.Normal ("Timed out : ") ;
      Message.msg_string Message.Error ("Timed out : ") ;
	List.iter (fun d -> Printf.fprintf Pervasives.stdout "%d " d ;
		Printf.fprintf Pervasives.stderr "%d " d ;
           ) !timeout ;
	Printf.fprintf Pervasives.stdout "\n\n-selecttest " ;
	List.iter (fun d -> Printf.fprintf Pervasives.stdout "-checktest %d " d ;
           ) !timeout ;
	Printf.fprintf Pervasives.stdout "\n\n" ;
	Printf.fprintf Pervasives.stderr "\n\n" ;
    end

    
    
(* main hook into TI_model_checker *)
let ti_modelcheck seedPreds =
   let module This_C_Abstraction = 
      Abstraction.Make_C_Abstraction(C_System_Descr) in
  let module Region = This_C_Abstraction.Region in
  let module This_LMC = LazyModelChecker.Make_Lazy_Model_Checker(This_C_Abstraction) in
  This_C_Abstraction.assumed_invariants := Stats.time "read invariants" constructInvariants();  
    This_C_Abstraction.assumed_invariants_region := This_C_Abstraction.assumed_invariants_list_to_region();
    let initStateFile = Options.getValueOfString "init" in
  let initialState = Predicate.conjoinL (Stats.time "read initial state" constructInitialPredicates initStateFile) in
  let mainFunctions = Options.getValueOfStringList "main" in
  let _ = This_C_Abstraction.initialize_abstraction () in
  let init_region =
      let rec buildloclist lst loclist =
	match lst with [] -> loclist
	  | a ::rest -> 
	      try
		buildloclist rest ((C_System_Descr.lookup_entry_location a) :: loclist) 
	      with 
		  Not_found -> 
		    (Format.printf "Creating init_region: No function %s defined in program. Exiting\n" (a);
		     failwith ("No function "^(a)^" defined"))
      in
      let loclist = buildloclist mainFunctions [] in
	This_C_Abstraction.ac_create_region false loclist false seedPreds  initialState
  in
    Message.msg_string Message.Normal "\n\n\n********** Now running the model-checker **********\n\n" ;
    flush stdout;
   try
        This_LMC.compute_transition_invariant () 
   with e -> 
	begin 
	  Message.msg_string Message.Error ("Exception raised :("^(Printexc.to_string e));
	  This_C_Abstraction.dump_abs ();
          Message.msg_printer Message.Normal This_C_Abstraction.print_stats ();
          exit 0
        end


let cfbmodelcheck seedPreds o_err_loc = 
  let module This_Cf_C_Abstraction =
    CfAbstraction.Make_C_Abstraction(C_System_Descr)
  in
  let module This_LMC_CFB =
    CfbLazyModelChecker.Make_Lazy_Model_Checker(This_Cf_C_Abstraction) in
  let module Region = This_Cf_C_Abstraction.Region in 
  This_Cf_C_Abstraction.assumed_invariants := Stats.time "read invariants" constructInvariants ();
  This_Cf_C_Abstraction.assumed_invariants_region := This_Cf_C_Abstraction.assumed_invariants_list_to_region();
  let initStateFile = Options.getValueOfString "init" in
  let initialState = Predicate.conjoinL (Stats.time "read initial state" constructInitialPredicates initStateFile) in
  let mainFunctions = Options.getValueOfStringList "main" in
 let _ = This_Cf_C_Abstraction.initialize_abstraction () in
    Message.msg_string Message.Normal "\n\n\n********** Now running the cfb-model-checker **********\n\n" ;
    flush stdout;
     let init_region =
      let rec buildloclist lst loclist =
	match lst with [] -> loclist
	  | a ::rest -> 
	      try
		buildloclist rest ((C_System_Descr.lookup_entry_location a) :: loclist) 
	      with 
		  Not_found -> 
		    (Format.printf "Creating init_region: No function %s defined in program. Exiting\n" (a);
		     failwith ("No function "^(a)^" defined"))
      in
      let loclist = buildloclist mainFunctions [] in
	This_Cf_C_Abstraction.create_region loclist false seedPreds  initialState
    and error_region =
      match o_err_loc with
	| Some err_loc ->
	    (This_Cf_C_Abstraction.create_region [err_loc] true Predicate.True Predicate.True)
	| None -> failwith "FMC: No error location!"
    in
    try
	This_LMC_CFB.model_check init_region error_region
    with e -> 
      begin 
        Message.msg_string Message.Error ("Exception raised :("^(Printexc.to_string e));
        raise e
      end





  
(* the main hook into the focimodelchecker *)
let focimodelcheck conc_flag seedPreds o_err_loc = 
  let module This_C_Abstraction = 
      Abstraction.Make_C_Abstraction(C_System_Descr) 
  in
  let module This_FMC = FociModelChecker.Make_Foci_Model_Checker(This_C_Abstraction)  in
  let module This_FMCC = FociModelChecker_concrete.Make_Foci_Model_Checker(This_C_Abstraction) in
  let module Region = This_C_Abstraction.Region in
    This_C_Abstraction.assumed_invariants := Stats.time "read invariants" constructInvariants();  
    This_C_Abstraction.assumed_invariants_region := This_C_Abstraction.assumed_invariants_list_to_region();
  let initStateFile = Options.getValueOfString "init" in
  let initialState = Predicate.conjoinL (Stats.time "read initial state" constructInitialPredicates initStateFile) in
  let mainFunctions = Options.getValueOfStringList "main" in
  let _ = This_C_Abstraction.initialize_abstraction () in
    Message.msg_string Message.Normal "\n\n\n********** Now running the model-checker **********\n\n" ;
    flush stdout;
     let init_region =
      let rec buildloclist lst loclist =
	match lst with [] -> loclist
	  | a ::rest -> 
	      try
		buildloclist rest ((C_System_Descr.lookup_entry_location a) :: loclist) 
	      with 
		  Not_found -> 
		    (Format.printf "Creating init_region: No function %s defined in program. Exiting\n" (a);
		     failwith ("No function "^(a)^" defined"))
      in
      let loclist = buildloclist mainFunctions [] in
	This_C_Abstraction.ac_create_region (not conc_flag) 
	  loclist false seedPreds  initialState
    and error_region =
      match o_err_loc with
	| Some err_loc ->
	    (This_C_Abstraction.create_region [err_loc] true Predicate.True Predicate.True)
	| None -> failwith "FMC: No error location!"
    in
    let model_checker = if conc_flag then This_FMCC.model_check else This_FMC.model_check in
      (* the main call into the focimodelchecker ... *)
      try
	model_checker init_region error_region
      with e -> 
	begin 
	  Message.msg_string Message.Error ("Exception raised :("^(Printexc.to_string e));
          This_C_Abstraction.dump_abs ();
          Message.msg_printer Message.Normal This_C_Abstraction.print_stats ();
	  raise e
	end


    
(*  TEST MODEL CHECKER NO LONGER SUPPORTED  
let testModelCheck i seedPreds =
  let initStateFile = Options.getValueOfString "init" in
  let initialState = Predicate.conjoinL (Stats.time "read initial state" constructInitialPredicates initStateFile) in
  let errorPredFile = Options.getValueOfString "error" in
  let errorPred = 
    let plist = (Stats.time "read error state" constructInitialPredicates errorPredFile) in
    if plist = [] then Predicate.True else List.hd plist
  in
  if i = 0 then
    TestModelChecker.test_model_check initialState errorPred seedPreds
  else  
    TestModelChecker.ccured_model_check initialState errorPred seedPreds 
 *)   

let runquery queryfile seedPreds =
  (*let inchan = open_in queryfile in
  let lexbuf = Qlflex.init inchan in
  Message.msg_string Message.Normal "In runquery -- About to parse" ;
  let stmts = Qlparse.main Qlflex.token lexbuf in
  Message.msg_string Message.Normal "In runquery -- Parsed input, now running statements" ;
  Ql.runStmts stmts*)
  failwith "No queries implemented"


(* INTERACTIVE MODE NO LONGER SUPPORTED 
let interactive () =
  (* Shell mode. Read inputs from stdin and execute them. *)

  Message.set_verbosity Message.Quiet;

  let reachLabel = (Options.getValueOfString "L")
  and source_files = (Options.getSourceFiles ()) in
  (* No reachLabel has been specified on the command line *)
  if (source_files = [])
  then
    begin
      Message.msg_string Message.Error ("Error: no input source file specified") ;
      exit 1
    end ;
  Message.msg_string Message.Normal ("Parsing files: " ^ (Misc.strList source_files)) ;
  
  let cil_files = Stats.time "Parse:" C_System_Descr.initialize_cil source_files in
  (* First do CIL specific optimizations, e.g., slicing. *)
  
  (* To do the slicing, we have to first read in the instrumentation predicates. *)
  (* Add seed predicates, if any to system *)
  let seedPredsFile = Options.getValueOfString "pred" in
  let seedPreds = Predicate.conjoinL (Stats.time "read seeds" constructInitialPredicates seedPredsFile)
  in
(*
   if (Options.getValueOfInt "O" > 1) then Panalysis.spec_slice cil_files seedPreds ;
 *)
  (* Then build CFA *)
  Stats.time "Build CFA:" C_System_Descr.initialize_blast cil_files;
(* OBSOLETE:
  List.iter SourceStats.printLines source_files;
  List.iter SourceStats.printStats cil_files;
  Format.printf ".\n";
*)
  Format.print_flush ();
  
  Message.msg_string Message.Debug "\n\n\n\n\nPrinting System Description\n\n" ;
  Message.msg_printer Message.Debug C_System_Descr.print_system () ;

  (* Do program analysis optimizations at the CFA level. *)
  Stats.time "Alias analysis" AliasAnalyzer.constructAliasDatabase () ; 
  AliasAnalyzer.constructMustAliasDatabase () ; 

  (* NOW -- we can build the modifies information, and THEN, add the call hooks. *)
  
  Stats.time "Modifies Database"
    (C_System_Descr.post_alias_analysis AliasAnalyzer.get_lval_aliases_scope) ();
  
  Message.msg_string Message.Normal "back from post_alias_analysis";  
    (*
     (* Construct call graph and check for recursion *)
     Stats.time "Call graph construction" C_System_Descr.compute_callgraph ();
     let callgraph_name = Options.getValueOfString "callgraph" in
     if callgraph_name = "" then () else 
     C_System_Descr.output_callgraph callgraph_name ;
     C_System_Descr.output_call_paths ();
   *)
  
    
  (* JHALA: what does this do ? 
 Rupak: The old modifies info computation. Retired. 
  if (Options.getValueOfString "alias" = "bdd") then AliasAnalyzer.constructModifiesInfo () ; 
*)

  let prompt () =
    Format.printf "> @?" in
  let parse_error () = Format.eprintf "Syntax Error\n@?" in

  let lexer = GL.make_lexer ["QUIT"; "SOURCE"; "TARGET"; "REACH"] in

  let entry_locs = ref [] in
  let err_locs = ref [] in

  let do_one st =
    match ST.next st with
      GL.Kwd "QUIT" -> raise End_of_file
    | GL.Kwd "SOURCE" ->
	(match ST.next st with
	  GL.String fname ->
	    (match ST.next st with
	      GL.Int line ->
		let pos = (fname, line) in
		let locs = C_System_Descr.get_position_sources pos in
		entry_locs := locs @ !entry_locs
	    | _ -> parse_error ())
	| _ -> parse_error ())
    | GL.Kwd "TARGET" ->
	(match ST.next st with
	  GL.String fname ->
	    (match ST.next st with
	      GL.Int line ->
		let pos = (fname, line) in
		let locs = C_System_Descr.get_position_sources pos in
		err_locs := locs @ !err_locs
	    | _ -> parse_error ())
	| _ -> parse_error ())
    | GL.Kwd "REACH" ->
	let reachFrom entry_loc =
	  let srcFname, srcLine =
	    match C_System_Descr.get_source_position entry_loc with
	      None -> failwith "Bad REACH source!"
	    | Some (fname, line, _) -> fname, line in
	  let rec reaches err_loc =
	    (*match err_locs with
	      [] -> ()
	    | _ ->*)
		modelCheck [err_loc] (Some [entry_loc]) seedPreds;
		match !reachedLoc with
		  None -> ()
		| Some loc ->
		    let fname, line =
		      match C_System_Descr.get_source_position loc with
			None -> failwith "Bad REACH target!"
		      | Some (fname, line, _) ->
			  Format.printf "REACH \"%s\" %d \"%s\" %d\n" srcFname srcLine fname line;
			  fname, line in
		    reachedLoc := None;
		    
		    (*reaches (List.filter
			       (fun x -> match C_System_Descr.get_source_position x with
				 Some (fname', line', _) -> fname <> fname' || line <> line'
			       | None -> true) err_locs)*)
	  in (*reaches !err_locs*)
	  List.iter reaches !err_locs
	in List.iter reachFrom !entry_locs;
	entry_locs := [];
	err_locs := [];
	Format.printf ".\n"
    | _ -> parse_error () in

  try
    while true do
      (*prompt ();*)
      (try
	do_one (lexer (Stream.of_string (read_line ())))
      with ST.Failure -> ());
      Format.print_flush ()
    done
  with
  | End_of_file    -> Format.printf "\n@?"; exit 0
*)

(* subgoal manager *)


let print_copyright () =
  (Message.msg_string Message.Normal "Blast 2.0. Copyright  2005  Regents of the University of California at Berkeley";
   Message.msg_string Message.Normal "This program contains Foci, Copyright  2003 Cadence Berkeley Laboratories," ;
   Message.msg_string Message.Normal "Cadence Design Systems. All rights reserved.")
    

(* main function: the glue *)
let main () =
(* try *)
  Sys.command "rm -f .newpred";
  Message.set_verbosity Message.Default;
  Stats.reset ();
  Options.buildOptionsDatabase ();
  print_copyright () ;

  if (Options.getValueOfBool "msvc") then Cil.msvcMode := true;
  Cil.initCIL ();

 
(* 
  if (Options.getValueOfBool "interactive" = true) then
    (* Shell mode. Read inputs from stdin and execute them. *)
    interactive ()
  else 
*)
  begin
    let reachLabel = (Options.getValueOfString "L")
    and source_files = (Options.getSourceFiles ()) in
    (* No reachLabel has been specified on the command line *)
    if(reachLabel = "") then
      begin
        Printf.printf "No error label specified!\n" ;
        exit 1
       end
    else
       () ;
     if (source_files == [])
     then
      begin
         Message.msg_string Message.Error ("Error: no input source file specified") ;
         exit 1
      end ;
     Message.msg_string Message.Normal ("Begin Parsing files: " ^ (Misc.strList source_files)) ;
     
     let cil_files = Stats.time "Parse:" C_System_Descr.initialize_cil source_files in
    (* First do CIL specific optimizations, e.g., slicing. *)
       Message.msg_string Message.Normal ("Finished Parsing");
     (* To do the slicing, we have to first read in the instrumentation predicates. *)
    (* Add seed predicates, if any to system *)
     let seedPredsFile = Options.getValueOfString "pred" in
     let seedPreds = Predicate.conjoinL (Stats.time "read seeds" constructInitialPredicates seedPredsFile) in
     let aliasPreds = 
       let aliasPredsFile = Options.getValueOfString "aliasfile" in
       Predicate.conjoinL (Stats.time "read alias pairs" constructInitialPredicates aliasPredsFile)
     in
(*
    if (Options.getValueOfInt "O" > 1) then Panalysis.spec_slice cil_files seedPreds ;
*)
    (* Then build CFA *)
       Message.msg_string Message.Normal ("Begin Building CFA");
       Stats.time "Build CFA:" C_System_Descr.initialize_blast cil_files;
       Message.msg_string Message.Normal ("Finished Building CFA");
       
    Message.msg_string Message.Debug "\n\n\n\n\nPrinting System Description\n\n" ;
    Message.msg_printer Message.Debug C_System_Descr.print_system () ;

    (* Do program analysis optimizations at the CFA level. *)
    Stats.time "Alias analysis" AliasAnalyzer.constructAliasDatabase () ; 
       Message.msg_string Message.Normal ("Finished Building Aliases");
    AliasAnalyzer.constructMustAliasDatabase () ; 
       Message.msg_string Message.Normal ("Finished Building Must Aliases");
    Stats.time "Modifies Database"
      (C_System_Descr.post_alias_analysis AliasAnalyzer.get_lval_aliases_scope) ();
    Stats.time "Add alias override"  (AliasAnalyzer.add_alias_override) aliasPreds;
  Message.msg_string Message.Normal "back from post_alias_analysis";  
 
   (* Construct call graph and check for recursion *)
    Stats.time "Call graph construction" C_System_Descr.compute_callgraph ();
    let callgraph_name = Options.getValueOfString "callgraph" in
    if callgraph_name = "" then () else 
      C_System_Descr.output_callgraph callgraph_name ;
    C_System_Descr.output_call_paths ();

(* TEST MODEL CHECKER NO LONGER SUPPORTED -- RUPAK
    if (Options.getValueOfBool "test") then
      begin
	testModelCheck 0  seedPreds; 
	exit 0 
      end 
    else () ;
   if (Options.getValueOfBool "ccured") then
     begin
	(* here is where the ccured function gets called *)
	testModelCheck 1  seedPreds; 
        exit 0
     end
   else ();
*)
 
    let queryfile = Options.getValueOfString "query" in
    if queryfile <> "" then
      begin
	runquery queryfile seedPreds
      end
    else 
      begin
    (* Add the sym_var_hooks for CF reachability *)
	let error_location =
	  match (C_System_Descr.get_locations_at_label reachLabel) with
	    [l] -> Some l
	  | [] when Options.getValueOfString "checkRace" <> "" -> None
	  | []  ->
              Message.msg_string Message.Error ("Error: label '" ^ reachLabel ^
						"' was not found in the source file(s)") ;
              exit 1
	  | _   ->
              Message.msg_string Message.Error ("Error: label '" ^ reachLabel ^
						"' appears multiple times in the source file(s)") ;
              exit 1
	in

(* NOT SUPPORTED
   if (Options.getValueOfString "interface" <> "") then
     begin
	(* here is where the interface synthesis function gets called *)
	TestModelChecker.synthesize_interface seedPreds error_location ; 
        exit 0
     end
   else ();
*)

   (*if (Options.getValueOfBool "cfb") then
     begin
       let _ = 
         try 
           Stats.time "cfbmodelcheck" (cfbmodelcheck seedPreds) error_location
         with e -> Message.msg_string Message.Error ("Error: "^(Printexc.to_string e));
         in
         This_C_Abstraction.dump_abs ();
         Message.msg_printer Message.Normal This_C_Abstraction.print_stats ();
     end
   else*)	  
     if (Options.getValueOfBool "fmc") then
       begin
           try 
             Stats.time "focimodelcheck" (focimodelcheck false seedPreds ) error_location
          with e -> Message.msg_string Message.Error ("Error: "^(Printexc.to_string e));
       end
   else
     if (Options.getValueOfBool "fmcc") then
       begin
           try Stats.time "focimodelcheck" (focimodelcheck true seedPreds) error_location
           with _ -> ()
       end
     else
       if (Options.getValueOfBool "ti") then
         begin
           let _ = 
             try Stats.time "ti_modelcheck" ti_modelcheck seedPreds
             with _ -> ()
         in
                  exit 1
       end
     else
       Stats.time "modelCheck"
       (modelCheck (Misc.option_to_list error_location) None) seedPreds;
      end (* ends if queryfile <> "" *)
  end;
  let memstats = Gc.stat () in
  Message.msg_string Message.Normal ("Minor Words : "^(string_of_float memstats.Gc.minor_words)) ;
      Message.msg_string Message.Normal ("Major Words : "^(string_of_float memstats.Gc.major_words)) ;
      Message.msg_string Message.Normal ("Total size of heap in words : "^(string_of_int memstats.Gc.heap_words)) ;
      if (Options.getValueOfBool "quiet") then () else 
      Stats.print stderr "Timings:\n"
(* with a -> flush stdout; flush stderr; raise a*)

;;


(* [Printexc.catch main ()] is deprecated as of OCaml 3.01: run the bytecode
   with OCAMLRUNPARAM=b to get a diagnostic with backtrace *)
let () =
  try
    main ();
    Message.msg_string Message.Normal !final_string;
    Message.msg_string Message.Error !final_string;
    
  with e -> 
    begin
      let flag = 
	match e with
	    Failure s ->
	      (not (Misc.is_substring s "o new pre")) 
	  | _ -> true
      in 
	if true then
	  begin
            let s = Printexc.to_string e in
	    Message.msg_string Message.Error ("Ack! The gremlins again!: "^s);
	    Message.msg_string Message.Normal ("Ack! The gremlins again!: "^s);
	  end;
	raise e;
    end


